@@ -115,18 +115,18 @@
  *      (until we add flags to _db_stack_frame_, increasing it by 4 bytes)
  */
 
-#define DEBUG_ON        (1 <<  1)  /* Debug enabled */
-#define FILE_ON         (1 <<  2)  /* File name print enabled */
-#define LINE_ON         (1 <<  3)  /* Line number print enabled */
-#define DEPTH_ON        (1 <<  4)  /* Function nest level print enabled */
-#define PROCESS_ON      (1 <<  5)  /* Process name print enabled */
-#define NUMBER_ON       (1 <<  6)  /* Number each line of output */
-#define PID_ON          (1 <<  8)  /* Identify each line with process id */
-#define TIMESTAMP_ON    (1 <<  9)  /* timestamp every line of output */
-#define FLUSH_ON_WRITE  (1 << 10)  /* Flush on every write */
-#define OPEN_APPEND     (1 << 11)  /* Open for append      */
-#define SANITY_CHECK_ON (1 << 12)  /* Check memory on every DBUG_ENTER/RETURN */
-#define TRACE_ON        ((uint)1 << 31)  /* Trace enabled. MUST be the highest bit!*/
+#define DEBUG_ON        (1U <<  1) /* Debug enabled */
+#define FILE_ON         (1U <<  2) /* File name print enabled */
+#define LINE_ON         (1U <<  3) /* Line number print enabled */
+#define DEPTH_ON        (1U <<  4) /* Function nest level print enabled */
+#define PROCESS_ON      (1U <<  5) /* Process name print enabled */
+#define NUMBER_ON       (1U <<  6) /* Number each line of output */
+#define PID_ON          (1U <<  8) /* Identify each line with process id */
+#define TIMESTAMP_ON    (1U <<  9) /* timestamp every line of output */
+#define FLUSH_ON_WRITE  (1U << 10) /* Flush on every write */
+#define OPEN_APPEND     (1U << 11) /* Open for append      */
+#define SANITY_CHECK_ON (1U << 12) /* Check memory on every DBUG_ENTER/RETURN */
+#define TRACE_ON        (1U << 31) /* Trace enabled. MUST be the highest bit!*/
 
 #define sf_sanity() (0)
 #define TRACING (cs->stack->flags & TRACE_ON)
@@ -226,8 +226,11 @@ typedef struct _db_code_state_ {
   const char *file;             /* Name of current user file                */
   struct _db_stack_frame_ *framep; /* Pointer to current frame              */
   struct settings *stack;       /* debugging settings                       */
+  const char *jmpfunc;          /* Remember current function for setjmp     */
+  const char *jmpfile;          /* Remember current file for setjmp         */
   int lineno;                   /* Current debugger output line number      */
   uint level;                   /* Current function nesting level           */
+  int jmplevel;                 /* Remember nesting level at setjmp()       */
 
 /*
  *      The following variables are used to hold the state information
@@ -320,33 +323,8 @@ static void DbugVfprintf(FILE *stream, const char* format, va_list args);
 */
 
 #include <my_pthread.h>
-/*
-** Protects writing to all file descriptors, init_settings.keywords
-** pointer and it's pointee - a linked list with keywords.
-*/
 static pthread_mutex_t THR_LOCK_dbug;
 
-static void LockMutex(CODE_STATE *cs)
-{
-  if (!cs->locked)
-    pthread_mutex_lock(&THR_LOCK_dbug);
-}
-static void UnlockMutex(CODE_STATE *cs)
-{
-  if (!cs->locked)
-    pthread_mutex_unlock(&THR_LOCK_dbug);
-}
-static void LockIfInitSettings(CODE_STATE *cs)
-{
-  if (cs->stack == &init_settings)
-    LockMutex(cs);
-}
-static void UnlockIfInitSettings(CODE_STATE *cs)
-{
-  if (cs->stack == &init_settings)
-    UnlockMutex(cs);
-}
-
 static CODE_STATE *code_state(void)
 {
   CODE_STATE *cs, **cs_ptr;
@@ -474,9 +452,16 @@ static int DbugParse(CODE_STATE *cs, const char *control)
   const char *end;
   int rel, f_used=0;
   struct settings *stack;
+  int org_cs_locked;
 
   stack= cs->stack;
 
+  if (!(org_cs_locked= cs->locked))
+  {
+    cs->locked= 1;
+    pthread_mutex_lock(&THR_LOCK_dbug);
+  }
+
   if (control[0] == '-' && control[1] == '#')
     control+=2;
 
@@ -490,9 +475,7 @@ static int DbugParse(CODE_STATE *cs, const char *control)
     stack->sub_level= 0;
     stack->out_file= sstderr;
     stack->functions= NULL;
-    LockIfInitSettings(cs);
     stack->keywords= NULL;
-    UnlockIfInitSettings(cs);
     stack->processes= NULL;
   }
   else if (!stack->out_file)
@@ -508,9 +491,7 @@ static int DbugParse(CODE_STATE *cs, const char *control)
     {
       /* never share with the global parent - it can change under your feet */
       stack->functions= ListCopy(init_settings.functions);
-      LockIfInitSettings(cs);
       stack->keywords= ListCopy(init_settings.keywords);
-      UnlockIfInitSettings(cs);
       stack->processes= ListCopy(init_settings.processes);
     }
     else
@@ -534,31 +515,21 @@ static int DbugParse(CODE_STATE *cs, const char *control)
     case 'd':
       if (sign < 0 && control == end)
       {
-        LockIfInitSettings(cs);
         if (!is_shared(stack, keywords))
           FreeList(stack->keywords);
         stack->keywords=NULL;
-        UnlockIfInitSettings(cs);
         stack->flags &= ~DEBUG_ON;
         break;
       }
-      LockIfInitSettings(cs);
       if (rel && is_shared(stack, keywords))
         stack->keywords= ListCopy(stack->keywords);
-      UnlockIfInitSettings(cs);
       if (sign < 0)
       {
         if (DEBUGGING)
-        {
-          LockIfInitSettings(cs);
           stack->keywords= ListDel(stack->keywords, control, end);
-          UnlockIfInitSettings(cs);
-        }
         break;
       }
-      LockIfInitSettings(cs);
       stack->keywords= ListAdd(stack->keywords, control, end);
-      UnlockIfInitSettings(cs);
       stack->flags |= DEBUG_ON;
       break;
     case 'D':
@@ -693,6 +664,11 @@ static int DbugParse(CODE_STATE *cs, const char *control)
     control=end+1;
     end= DbugStrTok(control);
   }
+  if (!org_cs_locked)
+  {
+    pthread_mutex_unlock(&THR_LOCK_dbug);
+    cs->locked= 0;
+  }
   return !rel || f_used;
 }
   
@@ -1025,9 +1001,7 @@ int _db_explain_ (CODE_STATE *cs, char *buf, size_t len)
 
   get_code_state_if_not_set_or_return *buf=0;
 
-  LockIfInitSettings(cs);
   op_list_to_buf('d', cs->stack->keywords, DEBUGGING);
-  UnlockIfInitSettings(cs);
   op_int_to_buf ('D', cs->stack->delay, 0);
   op_list_to_buf('f', cs->stack->functions, cs->stack->functions);
   op_bool_to_buf('F', cs->stack->flags & FILE_ON);
@@ -1122,6 +1096,7 @@ int _db_explain_init_(char *buf, size_t len)
 void _db_enter_(const char *_func_, const char *_file_,
                 uint _line_, struct _db_stack_frame_ *_stack_frame_)
 {
+  int save_errno;
   CODE_STATE *cs;
   if (!((cs=code_state())))
   {
@@ -1129,6 +1104,7 @@ void _db_enter_(const char *_func_, const char *_file_,
     _stack_frame_->prev= 0;
     return;
   }
+  save_errno= errno;
 
   _stack_frame_->line= -1;
   _stack_frame_->func= cs->func;
@@ -1149,14 +1125,12 @@ void _db_enter_(const char *_func_, const char *_file_,
       cs->stack->flags &= ~SANITY_CHECK_ON;
     if (TRACING)
     {
-      int save_errno= errno;
-      LockMutex(cs);
+      if (!cs->locked)
+        pthread_mutex_lock(&THR_LOCK_dbug);
       DoPrefix(cs, _line_);
       Indent(cs, cs->level);
       (void) fprintf(cs->stack->out_file->file, ">%s\n", cs->func);
-      UnlockMutex(cs);
-      DbugFlush(cs);
-      errno=save_errno;
+      DbugFlush(cs);                       /* This does a unlock */
     }
     break;
   case DISABLE_TRACE:
@@ -1165,6 +1139,7 @@ void _db_enter_(const char *_func_, const char *_file_,
   case DONT_TRACE:
     break;
   }
+  errno=save_errno;
 }
 
 /*
@@ -1189,6 +1164,7 @@ void _db_enter_(const char *_func_, const char *_file_,
 
 void _db_return_(struct _db_stack_frame_ *_stack_frame_)
 {
+  int save_errno=errno;
   uint _slevel_= _stack_frame_->level & ~TRACE_ON;
   CODE_STATE *cs;
   get_code_state_or_return;
@@ -1209,14 +1185,12 @@ void _db_return_(struct _db_stack_frame_ *_stack_frame_)
       cs->stack->flags &= ~SANITY_CHECK_ON;
     if (TRACING)
     {
-      int save_errno=errno;
-      LockMutex(cs);
+      if (!cs->locked)
+        pthread_mutex_lock(&THR_LOCK_dbug);
       DoPrefix(cs, _stack_frame_->line);
       Indent(cs, cs->level);
       (void) fprintf(cs->stack->out_file->file, "<%s\n", cs->func);
-      UnlockMutex(cs);
       DbugFlush(cs);
-      errno=save_errno;
     }
   }
   /*
@@ -1228,6 +1202,7 @@ void _db_return_(struct _db_stack_frame_ *_stack_frame_)
   cs->file= _stack_frame_->file;
   if (cs->framep != NULL)
     cs->framep= cs->framep->prev;
+  errno=save_errno;
 }
 
 
@@ -1238,7 +1213,7 @@ void _db_return_(struct _db_stack_frame_ *_stack_frame_)
  *
  *  SYNOPSIS
  *
- *      VOID _db_pargs_(_line_, keyword)
+ *      int _db_pargs_(_line_, keyword)
  *      int _line_;
  *      char *keyword;
  *
@@ -1251,12 +1226,14 @@ void _db_return_(struct _db_stack_frame_ *_stack_frame_)
  *
  */
 
-void _db_pargs_(uint _line_, const char *keyword)
+int _db_pargs_(uint _line_, const char *keyword)
 {
   CODE_STATE *cs;
-  get_code_state_or_return;
+  get_code_state_or_return 0;
   cs->u_line= _line_;
   cs->u_keyword= keyword;
+
+  return DEBUGGING && _db_keyword_(cs, cs->u_keyword, 0);
 }
 
 
@@ -1290,26 +1267,26 @@ void _db_doprnt_(const char *format,...)
 {
   va_list args;
   CODE_STATE *cs;
+  int save_errno;
+
   get_code_state_or_return;
 
+  va_start(args,format);
 
-  if (_db_keyword_(cs, cs->u_keyword, 0))
-  {
-    int save_errno=errno;
-    LockMutex(cs);
-    DoPrefix(cs, cs->u_line);
-    if (TRACING)
-      Indent(cs, cs->level + 1);
-    else
-      (void) fprintf(cs->stack->out_file->file, "%s: ", cs->func);
-    (void) fprintf(cs->stack->out_file->file, "%s: ", cs->u_keyword);
-    va_start(args,format);
-    DbugVfprintf(cs->stack->out_file->file, format, args);
-    UnlockMutex(cs);
-    va_end(args);
-    DbugFlush(cs);
-    errno=save_errno;
-  }
+  if (!cs->locked)
+    pthread_mutex_lock(&THR_LOCK_dbug);
+  save_errno=errno;
+  DoPrefix(cs, cs->u_line);
+  if (TRACING)
+    Indent(cs, cs->level + 1);
+  else
+    (void) fprintf(cs->stack->out_file->file, "%s: ", cs->func);
+  (void) fprintf(cs->stack->out_file->file, "%s: ", cs->u_keyword);
+  DbugVfprintf(cs->stack->out_file->file, format, args);
+  DbugFlush(cs);
+  errno=save_errno;
+
+  va_end(args);
 }
 
 /*
@@ -1350,9 +1327,10 @@ void _db_dump_(uint _line_, const char *keyword,
   CODE_STATE *cs;
   get_code_state_or_return;
 
+  if (!cs->locked)
+    pthread_mutex_lock(&THR_LOCK_dbug);
   if (_db_keyword_(cs, keyword, 0))
   {
-    LockMutex(cs);
     DoPrefix(cs, _line_);
     if (TRACING)
     {
@@ -1380,9 +1358,10 @@ void _db_dump_(uint _line_, const char *keyword,
       fputc(' ',cs->stack->out_file->file);
     }
     (void) fputc('\n',cs->stack->out_file->file);
-    UnlockMutex(cs);
     DbugFlush(cs);
   }
+  else if (!cs->locked)
+    pthread_mutex_unlock(&THR_LOCK_dbug);
 }
 
 
@@ -1608,10 +1587,8 @@ static void PushState(CODE_STATE *cs)
 static void FreeState(CODE_STATE *cs, int free_state)
 {
   struct settings *state= cs->stack;
-  LockIfInitSettings(cs);
   if (!is_shared(state, keywords))
     FreeList(state->keywords);
-  UnlockIfInitSettings(cs);
   if (!is_shared(state, functions))
     FreeList(state->functions);
   if (!is_shared(state, processes))
@@ -1734,16 +1711,10 @@ FILE *_db_fp_(void)
 BOOLEAN _db_keyword_(CODE_STATE *cs, const char *keyword, int strict)
 {
   int match= strict ? INCLUDE : INCLUDE|MATCHED;
-  BOOLEAN result = FALSE;
   get_code_state_if_not_set_or_return FALSE;
 
-  if (!DEBUGGING || !(DoTrace(cs) & DO_TRACE))
-    return FALSE;
-
-  LockIfInitSettings(cs);
-  result= (InList(cs->stack->keywords, keyword, strict) & match) ? TRUE : FALSE;
-  UnlockIfInitSettings(cs);
-  return result;
+  return (DEBUGGING && DoTrace(cs) & DO_TRACE &&
+          InList(cs->stack->keywords, keyword, strict) & match);
 }
 
 /*
@@ -1969,16 +1940,16 @@ static void DBUGCloseFile(CODE_STATE *cs, sFILE *new_value)
   sFILE *fp;
   if (!cs || !cs->stack || !cs->stack->out_file)
     return;
+  if (!cs->locked)
+    pthread_mutex_lock(&THR_LOCK_dbug);
 
   fp= cs->stack->out_file;
   if (--fp->used == 0)
   {
     if (fclose(fp->file) == EOF)
     {
-      LockMutex(cs);
       (void) fprintf(stderr, ERR_CLOSE, cs->process);
       perror("");
-      UnlockMutex(cs);
     }
     else
     {
@@ -1986,6 +1957,8 @@ static void DBUGCloseFile(CODE_STATE *cs, sFILE *new_value)
     }
   }
   cs->stack->out_file= new_value;
+  if (!cs->locked)
+    pthread_mutex_unlock(&THR_LOCK_dbug);
 }
 
 
@@ -2157,6 +2130,8 @@ static void DbugFlush(CODE_STATE *cs)
     if (cs->stack->delay)
       (void) Delay(cs->stack->delay);
   }
+  if (!cs->locked)
+    pthread_mutex_unlock(&THR_LOCK_dbug);
 } /* DbugFlush */
 
 
