@@ -12,7 +12,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
 
 #include <my_global.h>
 #include <my_stacktrace.h>
@@ -34,19 +34,19 @@
 #include <execinfo.h>
 #endif
 
-#ifdef __linux__
 #define PTR_SANE(p) ((p) && (char*)(p) >= heap_start && (char*)(p) <= heap_end)
+
 static char *heap_start;
-char *__bss_start;
-#else
-#define PTR_SANE(p) (p)
-#endif /* __linux */
+
+#if(defined HAVE_BSS_START) && !(defined __linux__)
+extern char *__bss_start;
+#endif
 
 void my_init_stacktrace()
 {
-#ifdef __linux__
+#if(defined HAVE_BSS_START) && !(defined __linux__)
   heap_start = (char*) &__bss_start;
-#endif /* __linux__ */
+#endif
 }
 
 #ifdef __linux__
@@ -149,15 +149,15 @@ static int safe_print_str(const char *addr, int max_len)
 
 int my_safe_print_str(const char* val, int max_len)
 {
-#ifdef __linux__
   char *heap_end;
 
+#ifdef __linux__
   // Try and make use of /proc filesystem to safely print memory contents.
   if (!safe_print_str(val, max_len))
     return 0;
+#endif
 
   heap_end= (char*) sbrk(0);
-#endif
 
   if (!PTR_SANE(val))
   {
@@ -483,7 +483,18 @@ void my_write_core(int sig)
 
 #else /* __WIN__*/
 
+#ifdef _MSC_VER
+/* Silence warning in OS header dbghelp.h */
+#pragma warning(push)
+#pragma warning(disable : 4091)
+#endif
+
 #include <dbghelp.h>
+
+#ifdef _MSC_VER
+#pragma warning(pop)
+#endif
+
 #include <tlhelp32.h>
 #include <my_sys.h>
 #if _MSC_VER
@@ -765,7 +776,7 @@ int my_safe_print_str(const char *val, int len)
 
 size_t my_write_stderr(const void *buf, size_t count)
 {
-  return (size_t) write(fileno(stderr), buf, count);
+  return (size_t) write(STDERR_FILENO, buf, count);
 }
 
 
