@@ -11,7 +11,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111-1301 USA */
 
 #include "feedback.h"
 
@@ -428,8 +428,8 @@ int calculate_server_uid(char *dest)
 
   my_sha1((uint8*) shabuf, (char*) rawbuf, sizeof(rawbuf));
 
-  assert(base64_needed_encoded_length(sizeof(shabuf)) <= SERVER_UID_SIZE);
-  base64_encode(shabuf, sizeof(shabuf), dest);
+  assert(my_base64_needed_encoded_length(sizeof(shabuf)) <= SERVER_UID_SIZE);
+  my_base64_encode(shabuf, sizeof(shabuf), dest);
 
   return 0;
 }
