@@ -65,6 +65,7 @@ public:
   double val_real();
   my_decimal *val_decimal(my_decimal *);
   enum Item_result result_type () const { return STRING_RESULT; }
+  enum_field_types field_type() const { return string_field_type(); }
   void left_right_max_length();
   bool fix_fields(THD *thd, Item **ref);
   void update_null_value()
@@ -141,6 +142,7 @@ public:
 
 class Item_func_md5 :public Item_str_ascii_checksum_func
 {
+  String tmp_value;
 public:
   Item_func_md5(THD *thd, Item *a): Item_str_ascii_checksum_func(thd, a) {}
   String *val_str_ascii(String *);
@@ -149,6 +151,8 @@ public:
     fix_length_and_charset(32, default_charset());
   }
   const char *func_name() const { return "md5"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_md5>(thd, mem_root, this); }
 };
 
 
@@ -158,7 +162,9 @@ public:
   Item_func_sha(THD *thd, Item *a): Item_str_ascii_checksum_func(thd, a) {}
   String *val_str_ascii(String *);    
   void fix_length_and_dec();      
-  const char *func_name() const { return "sha"; }	
+  const char *func_name() const { return "sha"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_sha>(thd, mem_root, this); }
 };
 
 class Item_func_sha2 :public Item_str_ascii_checksum_func
@@ -169,6 +175,8 @@ public:
   String *val_str_ascii(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "sha2"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_sha2>(thd, mem_root, this); }
 };
 
 class Item_func_to_base64 :public Item_str_ascii_checksum_func
@@ -180,6 +188,8 @@ public:
   String *val_str_ascii(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "to_base64"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_to_base64>(thd, mem_root, this); }
 };
 
 class Item_func_from_base64 :public Item_str_binary_checksum_func
@@ -191,6 +201,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "from_base64"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_from_base64>(thd, mem_root, this); }
 };
 
 #include <my_crypt.h>
@@ -202,7 +214,6 @@ class Item_aes_crypt :public Item_str_binary_checksum_func
 
 protected:
   int what;
-  String tmp_value;
 public:
   Item_aes_crypt(THD *thd, Item *a, Item *b)
    :Item_str_binary_checksum_func(thd, a, b) {}
@@ -212,10 +223,12 @@ public:
 class Item_func_aes_encrypt :public Item_aes_crypt
 {
 public:
-  Item_func_aes_encrypt(THD *thd, Item *a, Item *b)
-   :Item_aes_crypt(thd, a, b) {}
+  Item_func_aes_encrypt(THD *thd, Item *a, Item *b):
+    Item_aes_crypt(thd, a, b) {}
   void fix_length_and_dec();
   const char *func_name() const { return "aes_encrypt"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_aes_encrypt>(thd, mem_root, this); }
 };
 
 class Item_func_aes_decrypt :public Item_aes_crypt
@@ -225,23 +238,27 @@ public:
     Item_aes_crypt(thd, a, b) {}
   void fix_length_and_dec();
   const char *func_name() const { return "aes_decrypt"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_aes_decrypt>(thd, mem_root, this); }
 };
 
 
 class Item_func_concat :public Item_str_func
 {
   String tmp_value;
-  bool realloc_result(String *str, uint length) const;
 public:
   Item_func_concat(THD *thd, List<Item> &list): Item_str_func(thd, list) {}
   Item_func_concat(THD *thd, Item *a, Item *b): Item_str_func(thd, a, b) {}
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "concat"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_concat>(thd, mem_root, this); }
 };
 
 class Item_func_decode_histogram :public Item_str_func
 {
+  String tmp_value;
 public:
   Item_func_decode_histogram(THD *thd, Item *a, Item *b):
     Item_str_func(thd, a, b) {}
@@ -253,6 +270,8 @@ public:
     maybe_null= 1;
   }
   const char *func_name() const { return "decode_histogram"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_decode_histogram>(thd, mem_root, this); }
 };
 
 class Item_func_concat_ws :public Item_str_func
@@ -264,6 +283,8 @@ public:
   void fix_length_and_dec();
   const char *func_name() const { return "concat_ws"; }
   table_map not_null_tables() const { return 0; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_concat_ws>(thd, mem_root, this); }
 };
 
 class Item_func_reverse :public Item_str_func
@@ -274,6 +295,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "reverse"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_reverse>(thd, mem_root, this); }
 };
 
 
@@ -286,6 +309,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "replace"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_replace>(thd, mem_root, this); }
 };
 
 
@@ -307,9 +332,10 @@ public:
     DBUG_VOID_RETURN;
   }
   String *val_str(String *str);
-  bool fix_fields(THD *thd, Item **ref);
   void fix_length_and_dec();
   const char *func_name() const { return "regexp_replace"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_regexp_replace>(thd, mem_root, this); }
 };
 
 
@@ -328,9 +354,10 @@ public:
     DBUG_VOID_RETURN;
   }
   String *val_str(String *str);
-  bool fix_fields(THD *thd, Item **ref);
   void fix_length_and_dec();
   const char *func_name() const { return "regexp_substr"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_regexp_substr>(thd, mem_root, this); }
 };
 
 
@@ -344,6 +371,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "insert"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_insert>(thd, mem_root, this); }
 };
 
 
@@ -365,6 +394,8 @@ public:
   Item_func_lcase(THD *thd, Item *item): Item_str_conv(thd, item) {}
   const char *func_name() const { return "lcase"; }
   void fix_length_and_dec();
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_lcase>(thd, mem_root, this); }
 };
 
 class Item_func_ucase :public Item_str_conv
@@ -373,6 +404,8 @@ public:
   Item_func_ucase(THD *thd, Item *item): Item_str_conv(thd, item) {}
   const char *func_name() const { return "ucase"; }
   void fix_length_and_dec();
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_ucase>(thd, mem_root, this); }
 };
 
 
@@ -384,6 +417,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "left"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_left>(thd, mem_root, this); }
 };
 
 
@@ -395,6 +430,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "right"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_right>(thd, mem_root, this); }
 };
 
 
@@ -407,6 +444,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "substr"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_substr>(thd, mem_root, this); }
 };
 
 
@@ -419,6 +458,9 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "substring_index"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_substr_index>(thd, mem_root, this); }
+
 };
 
 
@@ -450,6 +492,8 @@ public:
   const char *func_name() const { return "trim"; }
   virtual void print(String *str, enum_query_type query_type);
   virtual const char *mode_name() const { return "both"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_trim>(thd, mem_root, this); }
 };
 
 
@@ -461,6 +505,8 @@ public:
   String *val_str(String *);
   const char *func_name() const { return "ltrim"; }
   const char *mode_name() const { return "leading"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_ltrim>(thd, mem_root, this); }
 };
 
 
@@ -472,6 +518,8 @@ public:
   String *val_str(String *);
   const char *func_name() const { return "rtrim"; }
   const char *mode_name() const { return "trailing"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_rtrim>(thd, mem_root, this); }
 };
 
 
@@ -509,6 +557,8 @@ public:
                                           "password" : "old_password"); }
   static char *alloc(THD *thd, const char *password, size_t pass_len,
                      enum PW_Alg al);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_password>(thd, mem_root, this); }
 };
 
 
@@ -529,6 +579,8 @@ public:
     max_length = args[0]->max_length + 9;
   }
   const char *func_name() const { return "des_encrypt"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_des_encrypt>(thd, mem_root, this); }
 };
 
 class Item_func_des_decrypt :public Item_str_binary_checksum_func
@@ -549,6 +601,8 @@ public:
       max_length-= 9U;
   }
   const char *func_name() const { return "des_decrypt"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_des_decrypt>(thd, mem_root, this); }
 };
 
 
@@ -579,10 +633,12 @@ public:
   String *val_str(String *);
   void fix_length_and_dec() { maybe_null=1; max_length = 13; }
   const char *func_name() const { return "encrypt"; }
-  bool check_vcol_func_processor(uchar *int_arg) 
+  bool check_vcol_func_processor(void *arg)
   {
-    return trace_unsupported_by_check_vcol_func_processor(func_name());
+    return FALSE;
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_encrypt>(thd, mem_root, this); }
 };
 
 #include "sql_crypt.h"
@@ -601,6 +657,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "encode"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_encode>(thd, mem_root, this); }
 protected:
   virtual void crypto_transform(String *);
 private:
@@ -614,6 +672,8 @@ class Item_func_decode :public Item_func_encode
 public:
   Item_func_decode(THD *thd, Item *a, Item *seed_arg): Item_func_encode(thd, a, seed_arg) {}
   const char *func_name() const { return "decode"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_decode>(thd, mem_root, this); }
 protected:
   void crypto_transform(String *);
 };
@@ -624,19 +684,21 @@ class Item_func_sysconst :public Item_str_func
 public:
   Item_func_sysconst(THD *thd): Item_str_func(thd)
   { collation.set(system_charset_info,DERIVATION_SYSCONST); }
-  Item *safe_charset_converter(THD *thd, CHARSET_INFO *tocs);
+  Item *safe_charset_converter(THD *thd, CHARSET_INFO *tocs)
+  {
+    return const_charset_converter(thd, tocs, true, fully_qualified_func_name());
+  }
   /*
     Used to create correct Item name in new converted item in
     safe_charset_converter, return string representation of this function
     call
   */
   virtual const char *fully_qualified_func_name() const = 0;
-  bool check_vcol_func_processor(uchar *int_arg) 
+  bool check_vcol_func_processor(void *arg)
   {
-    return trace_unsupported_by_check_vcol_func_processor(
-                                           fully_qualified_func_name());
+    return mark_unsupported_function(fully_qualified_func_name(), arg,
+                                     VCOL_SESSION_FUNC);
   }
-  bool const_item() const;
 };
 
 
@@ -652,6 +714,8 @@ public:
   }
   const char *func_name() const { return "database"; }
   const char *fully_qualified_func_name() const { return "database()"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_database>(thd, mem_root, this); }
 };
 
 
@@ -682,6 +746,8 @@ public:
   {
     return save_str_value_in_field(field, &str_value);
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_user>(thd, mem_root, this); }
 };
 
 
@@ -695,6 +761,12 @@ public:
   bool fix_fields(THD *thd, Item **ref);
   const char *func_name() const { return "current_user"; }
   const char *fully_qualified_func_name() const { return "current_user()"; }
+  bool check_vcol_func_processor(void *arg)
+  {
+    context= 0;
+    return mark_unsupported_function(fully_qualified_func_name(), arg,
+                                     VCOL_SESSION_FUNC);
+  }
 };
 
 
@@ -715,8 +787,17 @@ public:
   String *val_str(String *)
   {
     DBUG_ASSERT(fixed == 1);
-    return null_value ? NULL : &str_value;
+    return (null_value ? 0 : &str_value);
   }
+  bool check_vcol_func_processor(void *arg)
+  {
+
+    context= 0;
+    return mark_unsupported_function(fully_qualified_func_name(), arg,
+                                     VCOL_SESSION_FUNC);
+  }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_current_role>(thd, mem_root, this); }
 };
 
 
@@ -728,6 +809,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "soundex"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_soundex>(thd, mem_root, this); }
 };
 
 
@@ -740,6 +823,8 @@ public:
   String *val_str(String *str);
   void fix_length_and_dec();
   const char *func_name() const { return "elt"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_elt>(thd, mem_root, this); }
 };
 
 
@@ -752,11 +837,14 @@ public:
   String *val_str(String *str);
   void fix_length_and_dec();
   const char *func_name() const { return "make_set"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_make_set>(thd, mem_root, this); }
 };
 
 
 class Item_func_format :public Item_str_ascii_func
 {
+  String tmp_str;
   MY_LOCALE *locale;
 public:
   Item_func_format(THD *thd, Item *org, Item *dec):
@@ -769,6 +857,8 @@ public:
   void fix_length_and_dec();
   const char *func_name() const { return "format"; }
   virtual void print(String *str, enum_query_type query_type);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_format>(thd, mem_root, this); }
 };
 
 
@@ -781,12 +871,13 @@ public:
     Item_str_func(thd, list)
   { collation.set(cs); }
   String *val_str(String *);
-  void fix_length_and_dec() 
+  void fix_length_and_dec()
   {
     max_length= arg_count * 4;
   }
   const char *func_name() const { return "char"; }
-  void print(String *str, enum_query_type query_type);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_char>(thd, mem_root, this); }
 };
 
 
@@ -799,6 +890,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "repeat"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_repeat>(thd, mem_root, this); }
 };
 
 
@@ -809,48 +902,55 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "space"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_space>(thd, mem_root, this); }
 };
 
 
 class Item_func_binlog_gtid_pos :public Item_str_func
 {
+  String tmp_value;
 public:
   Item_func_binlog_gtid_pos(THD *thd, Item *arg1, Item *arg2):
     Item_str_func(thd, arg1, arg2) {}
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "binlog_gtid_pos"; }
+  bool check_vcol_func_processor(void *arg)
+  {
+
+    return mark_unsupported_function(func_name(), "()", arg, VCOL_IMPOSSIBLE);
+  }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_binlog_gtid_pos>(thd, mem_root, this); }
 };
 
 
-class Item_func_pad: public Item_str_func
-{
-protected:
-  String tmp_value, pad_str;
-public:
-  Item_func_pad(THD *thd, Item *arg1, Item *arg2, Item *arg3):
-    Item_str_func(thd, arg1, arg2, arg3) {}
-  void fix_length_and_dec();
-};
-
-
-class Item_func_rpad :public Item_func_pad
+class Item_func_rpad :public Item_str_func
 {
+  String tmp_value, rpad_str;
 public:
   Item_func_rpad(THD *thd, Item *arg1, Item *arg2, Item *arg3):
-    Item_func_pad(thd, arg1, arg2, arg3) {}
+    Item_str_func(thd, arg1, arg2, arg3) {}
   String *val_str(String *);
+  void fix_length_and_dec();
   const char *func_name() const { return "rpad"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_rpad>(thd, mem_root, this); }
 };
 
 
-class Item_func_lpad :public Item_func_pad
+class Item_func_lpad :public Item_str_func
 {
+  String tmp_value, lpad_str;
 public:
   Item_func_lpad(THD *thd, Item *arg1, Item *arg2, Item *arg3):
-    Item_func_pad(thd, arg1, arg2, arg3) {}
+    Item_str_func(thd, arg1, arg2, arg3) {}
   String *val_str(String *);
+  void fix_length_and_dec();
   const char *func_name() const { return "lpad"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_lpad>(thd, mem_root, this); }
 };
 
 
@@ -867,6 +967,8 @@ public:
     max_length=64;
     maybe_null= 1;
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_conv>(thd, mem_root, this); }
 };
 
 
@@ -884,6 +986,8 @@ public:
     decimals=0;
     fix_char_length(args[0]->max_length * 2);
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_hex>(thd, mem_root, this); }
 };
 
 class Item_func_unhex :public Item_str_func
@@ -903,6 +1007,8 @@ public:
     decimals=0;
     max_length=(1+args[0]->max_length)/2;
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_unhex>(thd, mem_root, this); }
 };
 
 
@@ -933,6 +1039,8 @@ public:
   Item_func_like_range_min(THD *thd, Item *a, Item *b):
     Item_func_like_range(thd, a, b, true) { }
   const char *func_name() const { return "like_range_min"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_like_range_min>(thd, mem_root, this); }
 };
 
 
@@ -942,6 +1050,8 @@ public:
   Item_func_like_range_max(THD *thd, Item *a, Item *b):
     Item_func_like_range(thd, a, b, false) { }
   const char *func_name() const { return "like_range_max"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_like_range_max>(thd, mem_root, this); }
 };
 #endif
 
@@ -966,6 +1076,9 @@ public:
   }
   virtual void print(String *str, enum_query_type query_type);
   const char *func_name() const { return "cast_as_binary"; }
+  bool need_parentheses_in_default() { return true; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_binary>(thd, mem_root, this); }
 };
 
 
@@ -982,10 +1095,12 @@ public:
     maybe_null=1;
     max_length=MAX_BLOB_WIDTH;
   }
-  bool check_vcol_func_processor(uchar *int_arg) 
+  bool check_vcol_func_processor(void *arg)
   {
-    return trace_unsupported_by_check_vcol_func_processor(func_name());
+    return mark_unsupported_function(func_name(), "()", arg, VCOL_IMPOSSIBLE);
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_load_file>(thd, mem_root, this); }
 };
 
 
@@ -1001,6 +1116,8 @@ class Item_func_export_set: public Item_str_func
   String  *val_str(String *str);
   void fix_length_and_dec();
   const char *func_name() const { return "export_set"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_export_set>(thd, mem_root, this); }
 };
 
 
@@ -1018,6 +1135,8 @@ public:
                                   2 * collation.collation->mbmaxlen;
     max_length= (uint32) MY_MIN(max_result_length, MAX_BLOB_WIDTH);
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_quote>(thd, mem_root, this); }
 };
 
 class Item_func_conv_charset :public Item_str_func
@@ -1100,6 +1219,8 @@ public:
   void fix_length_and_dec();
   const char *func_name() const { return "convert"; }
   virtual void print(String *str, enum_query_type query_type);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_conv_charset>(thd, mem_root, this); }
 };
 
 class Item_func_set_collation :public Item_str_func
@@ -1111,6 +1232,7 @@ public:
   void fix_length_and_dec();
   bool eq(const Item *item, bool binary_cmp) const;
   const char *func_name() const { return "collate"; }
+  enum precedence precedence() const { return COLLATE_PRECEDENCE; }
   enum Functype functype() const { return COLLATE_FUNC; }
   virtual void print(String *str, enum_query_type query_type);
   Item_field *field_for_view_update()
@@ -1118,6 +1240,9 @@ public:
     /* this function is transparent for view updating */
     return args[0]->field_for_view_update();
   }
+  bool need_parentheses_in_default() { return true; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_set_collation>(thd, mem_root, this); }
 };
 
 
@@ -1145,6 +1270,8 @@ public:
     :Item_func_expr_str_metadata(thd, a) { }
   String *val_str(String *);
   const char *func_name() const { return "charset"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_charset>(thd, mem_root, this); }
 };
 
 
@@ -1155,6 +1282,8 @@ public:
     :Item_func_expr_str_metadata(thd, a) {}
   String *val_str(String *);
   const char *func_name() const { return "collation"; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_collation>(thd, mem_root, this); }
 };
 
 
@@ -1187,6 +1316,9 @@ public:
   }
   Item* propagate_equal_fields(THD *thd, const Context &ctx, COND_EQUAL *cond)
   { return this; }
+  void print(String *str, enum_query_type query_type);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_weight_string>(thd, mem_root, this); }
 };
 
 class Item_func_crc32 :public Item_int_func
@@ -1198,6 +1330,8 @@ public:
   const char *func_name() const { return "crc32"; }
   void fix_length_and_dec() { max_length=10; }
   longlong val_int();
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_crc32>(thd, mem_root, this); }
 };
 
 class Item_func_uncompressed_length : public Item_int_func
@@ -1208,6 +1342,8 @@ public:
   const char *func_name() const{return "uncompressed_length";}
   void fix_length_and_dec() { max_length=10; maybe_null= true; }
   longlong val_int();
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_uncompressed_length>(thd, mem_root, this); }
 };
 
 #ifdef HAVE_COMPRESS
@@ -1218,24 +1354,28 @@ public:
 
 class Item_func_compress: public Item_str_binary_checksum_func
 {
-  String tmp_value;
+  String buffer;
 public:
   Item_func_compress(THD *thd, Item *a)
    :Item_str_binary_checksum_func(thd, a) {}
   void fix_length_and_dec(){max_length= (args[0]->max_length*120)/100+12;}
   const char *func_name() const{return "compress";}
   String *val_str(String *) ZLIB_DEPENDED_FUNCTION
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_compress>(thd, mem_root, this); }
 };
 
 class Item_func_uncompress: public Item_str_binary_checksum_func
 {
-  String tmp_value;
+  String buffer;
 public:
   Item_func_uncompress(THD *thd, Item *a)
    :Item_str_binary_checksum_func(thd, a) {}
   void fix_length_and_dec(){ maybe_null= 1; max_length= MAX_BLOB_WIDTH; }
   const char *func_name() const{return "uncompress";}
   String *val_str(String *) ZLIB_DEPENDED_FUNCTION
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_uncompress>(thd, mem_root, this); }
 };
 
 
@@ -1249,14 +1389,14 @@ public:
                   DERIVATION_COERCIBLE, MY_REPERTOIRE_ASCII);
     fix_char_length(MY_UUID_STRING_LENGTH);
   }
-  bool const_item() const { return false; }
-  table_map used_tables() const { return RAND_TABLE_BIT; }
   const char *func_name() const{ return "uuid"; }
   String *val_str(String *);
-  bool check_vcol_func_processor(uchar *int_arg) 
+  bool check_vcol_func_processor(void *arg)
   {
-    return trace_unsupported_by_check_vcol_func_processor(func_name());
+    return mark_unsupported_function(func_name(), "()", arg, VCOL_NON_DETERMINISTIC);
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_uuid>(thd, mem_root, this); }
 };
 
 
@@ -1268,7 +1408,7 @@ protected:
   uint *keys_num;
   LEX_STRING *keys_str;
   bool names, force_names;
-  bool prepare_arguments(bool force_names);
+  bool prepare_arguments(THD *thd, bool force_names);
   void print_arguments(String *str, enum_query_type query_type);
 public:
   Item_func_dyncol_create(THD *thd, List<Item> &args, DYNCALL_CREATE_DEF *dfs);
@@ -1278,6 +1418,8 @@ public:
   String *val_str(String *);
   virtual void print(String *str, enum_query_type query_type);
   virtual enum Functype functype() const   { return DYNCOL_FUNC; }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_dyncol_create>(thd, mem_root, this); }
 };
 
 
@@ -1290,21 +1432,25 @@ public:
   const char *func_name() const{ return "column_add"; }
   String *val_str(String *);
   virtual void print(String *str, enum_query_type query_type);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_dyncol_add>(thd, mem_root, this); }
 };
 
 class Item_func_dyncol_json: public Item_str_func
 {
 public:
-  Item_func_dyncol_json(THD *thd, Item *str): Item_str_func(thd, str)
-    {collation.set(DYNCOL_UTF);}
+  Item_func_dyncol_json(THD *thd, Item *str): Item_str_func(thd, str) {}
   const char *func_name() const{ return "column_json"; }
   String *val_str(String *);
   void fix_length_and_dec()
   {
     max_length= MAX_BLOB_WIDTH;
     maybe_null= 1;
+    collation.set(&my_charset_bin);
     decimals= 0;
   }
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_dyncol_json>(thd, mem_root, this); }
 };
 
 /*
@@ -1324,22 +1470,41 @@ public:
   const char *func_name() const { return "column_get"; }
   String *val_str(String *);
   longlong val_int();
+  longlong val_int_signed_typecast()
+  {
+    unsigned_flag= false;   // Mark that we want to have a signed value
+    longlong value= val_int(); // val_int() can change unsigned_flag
+    if (!null_value && unsigned_flag && value < 0)
+      push_note_converted_to_negative_complement(current_thd);
+    return value;
+  }
+  longlong val_int_unsigned_typecast()
+  {
+    unsigned_flag= true; // Mark that we want to have an unsigned value
+    longlong value= val_int(); // val_int() can change unsigned_flag
+    if (!null_value && unsigned_flag == 0 && value < 0)
+      push_note_converted_to_positive_complement(current_thd);
+    return value;
+  }
   double val_real();
   my_decimal *val_decimal(my_decimal *);
-  bool get_dyn_value(DYNAMIC_COLUMN_VALUE *val, String *tmp);
+  bool get_dyn_value(THD *thd, DYNAMIC_COLUMN_VALUE *val, String *tmp);
   bool get_date(MYSQL_TIME *ltime, ulonglong fuzzydate);
   void print(String *str, enum_query_type query_type);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_dyncol_get>(thd, mem_root, this); }
 };
 
 
 class Item_func_dyncol_list: public Item_str_func
 {
 public:
-  Item_func_dyncol_list(THD *thd, Item *str): Item_str_func(thd, str)
-    {collation.set(DYNCOL_UTF);}
+  Item_func_dyncol_list(THD *thd, Item *str): Item_str_func(thd, str) {};
   void fix_length_and_dec() { maybe_null= 1; max_length= MAX_BLOB_WIDTH; };
   const char *func_name() const{ return "column_list"; }
   String *val_str(String *);
+  Item *get_copy(THD *thd, MEM_ROOT *mem_root)
+  { return get_item_copy<Item_func_dyncol_list>(thd, mem_root, this); }
 };
 
 #endif /* ITEM_STRFUNC_INCLUDED */
