@@ -1,5 +1,5 @@
 /* Copyright (c) 2000, 2011, Oracle and/or its affiliates.
-   Copyright (c) 2008, 2017, MariaDB Corporation.
+   Copyright (c) 2008-2011 Monty Program Ab
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,7 +12,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
 
 
 /**
@@ -254,8 +254,6 @@ int opt_sum_query(THD *thd,
   int error= 0;
   DBUG_ENTER("opt_sum_query");
 
-  thd->lex->current_select->min_max_opt_list.empty();
-
   if (conds)
     where_tables= conds->used_tables();
 
@@ -408,7 +406,7 @@ int opt_sum_query(THD *thd,
 	  if (!error && reckey_in_range(is_max, &ref, item_field->field, 
 			                conds, range_fl, prefix_len))
 	    error= HA_ERR_KEY_NOT_FOUND;
-          table->disable_keyread();
+          table->file->ha_end_keyread();
           table->file->ha_index_end();
           if (error)
 	  {
@@ -449,14 +447,7 @@ int opt_sum_query(THD *thd,
           item_sum->aggregator_clear();
         }
         else
-        {
           item_sum->reset_and_add();
-          /*
-            Save a reference to the item for possible rollback
-            of the min/max optimizations for this select
-          */
-	  thd->lex->current_select->min_max_opt_list.push_back(item_sum);
-        }
         item_sum->make_const();
         recalc_const_item= 1;
         break;
@@ -977,7 +968,7 @@ static bool find_key_for_maxmin(bool max_fl, TABLE_REF *ref,
             converted (for example to upper case)
           */
           if (field->part_of_key.is_set(idx))
-            table->enable_keyread();
+            table->file->ha_start_keyread(idx);
           DBUG_RETURN(TRUE);
         }
       }
@@ -1051,7 +1042,6 @@ static int maxmin_in_range(bool max_fl, Field* field, COND *cond)
   case Item_func::LT_FUNC:
   case Item_func::LE_FUNC:
     less_fl= 1;
-    /* fall through */
   case Item_func::GT_FUNC:
   case Item_func::GE_FUNC:
   {
