@@ -18,7 +18,7 @@
 #include "sql_priv.h"
 #include <my_dir.h>
 #include "rpl_mi.h"
-#include "slave.h"                              // SLAVE_MAX_HEARTBEAT_PERIOD
+#include "slave.h"
 #include "strfunc.h"
 #include "sql_repl.h"
 
@@ -664,14 +664,14 @@ file '%s')", fname);
     mi->connect_retry= (uint) connect_retry;
     mi->ssl= (my_bool) ssl;
     mi->ssl_verify_server_cert= ssl_verify_server_cert;
-    mi->heartbeat_period= MY_MIN(SLAVE_MAX_HEARTBEAT_PERIOD, master_heartbeat_period);
+    mi->heartbeat_period= master_heartbeat_period;
   }
   DBUG_PRINT("master_info",("log_file_name: %s  position: %ld",
                             mi->master_log_name,
                             (ulong) mi->master_log_pos));
 
   mi->rli.mi= mi;
-  if (init_relay_log_info(&mi->rli, slave_info_fname))
+  if (mi->rli.init(slave_info_fname))
     goto err;
 
   mi->inited = 1;
@@ -799,8 +799,8 @@ int flush_master_info(Master_info* mi,
      contents of file). But because of number of lines in the first line
      of file we don't care about this garbage.
   */
-  char heartbeat_buf[FLOATING_POINT_BUFFER];
-  my_fcvt(mi->heartbeat_period, 3, heartbeat_buf, NULL);
+  char heartbeat_buf[sizeof(mi->heartbeat_period) * 4]; // buffer to suffice always
+  sprintf(heartbeat_buf, "%.3f", mi->heartbeat_period);
   my_b_seek(file, 0L);
   my_b_printf(file,
               "%u\n%s\n%s\n%s\n%s\n%s\n%d\n%d\n%d\n%s\n%s\n%s\n%s\n%s\n%d\n%s\n%s\n%s\n%s\n%d\n%s\n%s\n"
@@ -1116,7 +1116,7 @@ bool Master_info_index::init_all_master_info()
     DBUG_RETURN(1);
   }
 
-  thd= new THD;  /* Needed by start_slave_threads */
+  thd= new THD(next_thread_id());  /* Needed by start_slave_threads */
   thd->thread_stack= (char*) &thd;
   thd->store_globals();
 
@@ -1239,7 +1239,7 @@ bool Master_info_index::init_all_master_info()
   if (succ_num) // Have some Error and some Success
   {
     sql_print_warning("Reading of some Master_info entries failed");
-    DBUG_RETURN(2);
+    DBUG_RETURN(1);
   }
 
   sql_print_error("Reading of all Master_info entries failed!");
