@@ -187,6 +187,10 @@ class Master_info : public Slave_reporting_capability
     return opt_slave_parallel_threads > 0 &&
       parallel_mode > SLAVE_PARALLEL_NONE;
   }
+  void release();
+  void wait_until_free();
+  void lock_slave_threads();
+  void unlock_slave_threads();
 
   /* the variables below are needed because we can change masters on the fly */
   char master_log_name[FN_REFLEN+6]; /* Room for multi-*/
@@ -205,7 +209,7 @@ class Master_info : public Slave_reporting_capability
   File fd; // we keep the file open, so we need to remember the file pointer
   IO_CACHE file;
 
-  mysql_mutex_t data_lock, run_lock, sleep_lock;
+  mysql_mutex_t data_lock, run_lock, sleep_lock, start_stop_lock;
   mysql_cond_t data_cond, start_cond, stop_cond, sleep_cond;
   THD *io_thd;
   MYSQL* mysql;
@@ -297,6 +301,9 @@ class Master_info : public Slave_reporting_capability
   uint64 gtid_reconnect_event_skip_count;
   /* gtid_event_seen is false until we receive first GTID event from master. */
   bool gtid_event_seen;
+  bool in_start_all_slaves, in_stop_all_slaves;
+  uint users;                                   /* Active user for object */
+  uint killed;
 
   /* domain-id based filter */
   Domain_id_filter domain_id_filter;
@@ -341,13 +348,12 @@ public:
   bool check_duplicate_master_info(LEX_STRING *connection_name,
                                    const char *host, uint port);
   bool add_master_info(Master_info *mi, bool write_to_file);
-  bool remove_master_info(LEX_STRING *connection_name);
+  bool remove_master_info(Master_info *mi);
   Master_info *get_master_info(const LEX_STRING *connection_name,
                                Sql_condition::enum_warning_level warning);
-  bool give_error_if_slave_running();
-  uint any_slave_sql_running();
   bool start_all_slaves(THD *thd);
   bool stop_all_slaves(THD *thd);
+  void free_connections();
 };
 
 
@@ -360,6 +366,8 @@ public:
 };
 
 
+Master_info *get_master_info(const LEX_STRING *connection_name,
+                             Sql_condition::enum_warning_level warning);
 bool check_master_connection_name(LEX_STRING *name);
 void create_logfile_name_with_suffix(char *res_file_name, size_t length,
                              const char *info_file, 
@@ -369,5 +377,8 @@ void create_logfile_name_with_suffix(char *res_file_name, size_t length,
 uchar *get_key_master_info(Master_info *mi, size_t *length,
                            my_bool not_used __attribute__((unused)));
 void free_key_master_info(Master_info *mi);
+uint any_slave_sql_running();
+bool give_error_if_slave_running(bool already_lock);
+
 #endif /* HAVE_REPLICATION */
 #endif /* RPL_MI_H */
