@@ -1,4 +1,5 @@
 /* Copyright (c) 2000, 2013, Oracle and/or its affiliates.
+   Copyright (c) 2016, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -136,7 +137,7 @@ bool String::set_real(double num,uint decimals, CHARSET_INFO *cs)
   size_t len;
 
   str_charset=cs;
-  if (decimals >= NOT_FIXED_DEC)
+  if (decimals >= FLOATING_POINT_DECIMALS)
   {
     len= my_gcvt(num, MY_GCVT_ARG_DOUBLE, sizeof(buff) - 1, buff, NULL);
     return copy(buff, len, &my_charset_latin1, cs, &dummy_errors);
@@ -183,16 +184,7 @@ bool String::copy(const char *str,uint32 arg_length, CHARSET_INFO *cs)
 {
   if (alloc(arg_length))
     return TRUE;
-  if (Ptr == str && arg_length == str_length)
-  {
-    /*
-      This can happen in some cases. This code is here mainly to avoid
-      warnings from valgrind, but can also be an indication of error.
-    */
-    DBUG_PRINT("warning", ("Copying string on itself: %p  %u",
-                           str, arg_length));
-  }
-  else if ((str_length=arg_length))
+  if ((str_length=arg_length))
     memcpy(Ptr,str,arg_length);
   Ptr[arg_length]=0;
   str_charset=cs;
@@ -741,7 +733,7 @@ void String::qs_append(int i)
 void String::qs_append(ulonglong i)
 {
   char *buff= Ptr + str_length;
-  char *end= longlong10_to_str(i, buff,10);
+  char *end= longlong10_to_str(i, buff, 10);
   str_length+= (int) (end-buff);
 }
 
@@ -768,7 +760,7 @@ int sortcmp(const String *s,const String *t, CHARSET_INFO *cs)
 {
  return cs->coll->strnncollsp(cs,
                               (uchar *) s->ptr(),s->length(),
-                              (uchar *) t->ptr(),t->length(), 0);
+                              (uchar *) t->ptr(),t->length());
 }
 
 
@@ -909,27 +901,6 @@ String *copy_if_not_alloced(String *to,String *from,uint32 from_length)
     (void) from->realloc(from_length);
     return from;
   }
-  if (from->uses_buffer_owned_by(to))
-  {
-    DBUG_ASSERT(!from->alloced);
-    DBUG_ASSERT(to->alloced);
-    /*
-      "from" is a constant string pointing to a fragment of alloced string "to":
-        to=  xxxFFFyyy
-      - FFF is the part of "to" pointed by "from"
-      - xxx is the part of "to" before "from"
-      - yyy is the part of "to" after "from"
-    */
-    uint32 xxx_length= (uint32) (from->ptr() - to->ptr());
-    uint32 yyy_length= (uint32) (to->end() - from->end());
-    DBUG_ASSERT(to->length() >= yyy_length);
-    to->length(to->length() - yyy_length); // Remove the "yyy" part
-    DBUG_ASSERT(to->length() >= xxx_length);
-    to->replace(0, xxx_length, "", 0);     // Remove the "xxx" part
-    to->realloc(from_length);
-    to->str_charset= from->str_charset;
-    return to;
-  }
   if (to->realloc(from_length))
     return from;				// Actually an error
   if ((to->str_length=MY_MIN(from->str_length,from_length)))
@@ -1047,10 +1018,10 @@ String_copier::well_formed_copy(CHARSET_INFO *to_cs,
   {
     m_cannot_convert_error_pos= NULL;
     return to_cs->cset->copy_fix(to_cs, to, to_length, from, from_length,
-                                 nchars, &m_native_copy_status);
+                                 nchars, this);
   }
   return my_convert_fix(to_cs, to, to_length, from_cs, from, from_length,
-                        nchars, this);
+                        nchars, this, this);
 }
 
 
