@@ -318,39 +318,20 @@ int wsrep_write_cache(wsrep_t*  const wsrep,
 
 void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
 {
-  int len= snprintf(NULL, 0, "%s/GRA_%ld_%lld.log",
-                    wsrep_data_home_dir, thd->thread_id,
-                    (long long)wsrep_thd_trx_seqno(thd));
-  if (len < 0)
+  char filename[PATH_MAX]= {0};
+  int len= snprintf(filename, PATH_MAX, "%s/GRA_%lld_%lld.log",
+                    wsrep_data_home_dir, (longlong) thd->thread_id,
+                    (longlong) wsrep_thd_trx_seqno(thd));
+  if (len >= PATH_MAX)
   {
-    WSREP_ERROR("snprintf error: %d, skipping dump.", len);
-    return;
-  }
-  /*
-    len doesn't count the \0 end-of-string. Use len+1 below
-    to alloc and pass as an argument to snprintf.
-  */
-
-  char *filename= (char *)malloc(len+1);
-  int len1= snprintf(filename, len+1, "%s/GRA_%ld_%lld.log",
-                    wsrep_data_home_dir, thd->thread_id,
-                    (long long)wsrep_thd_trx_seqno(thd));
-
-  if (len > len1)
-  {
-    WSREP_ERROR("RBR dump path truncated: %d, skipping dump.", len);
-    free(filename);
+    WSREP_ERROR("RBR dump path too long: %d, skipping dump.", len);
     return;
   }
 
   FILE *of= fopen(filename, "wb");
-
   if (of)
   {
-    if (fwrite(rbr_buf, buf_len, 1, of) == 0)
-       WSREP_ERROR("Failed to write buffer of length %llu to '%s'",
-                   (unsigned long long)buf_len, filename);
-
+    fwrite (rbr_buf, buf_len, 1, of);
     fclose(of);
   }
   else
@@ -358,7 +339,6 @@ void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
     WSREP_ERROR("Failed to open file '%s': %d (%s)",
                 filename, errno, strerror(errno));
   }
-  free(filename);
 }
 
 /*
@@ -394,9 +374,9 @@ int wsrep_binlog_savepoint_rollback(THD *thd, void *sv)
 void wsrep_dump_rbr_direct(THD* thd, IO_CACHE* cache)
 {
   char filename[PATH_MAX]= {0};
-  int len= snprintf(filename, PATH_MAX, "%s/GRA_%ld_%lld.log",
-                    wsrep_data_home_dir, thd->thread_id,
-                    (long long)wsrep_thd_trx_seqno(thd));
+  int len= snprintf(filename, PATH_MAX, "%s/GRA_%lld_%lld.log",
+                    wsrep_data_home_dir, (longlong) thd->thread_id,
+                    (longlong) wsrep_thd_trx_seqno(thd));
   size_t bytes_in_cache = 0;
   // check path
   if (len >= PATH_MAX)
@@ -464,35 +444,19 @@ void wsrep_dump_rbr_buf_with_header(THD *thd, const void *rbr_buf,
 {
   DBUG_ENTER("wsrep_dump_rbr_buf_with_header");
 
+  char filename[PATH_MAX]= {0};
   File file;
   IO_CACHE cache;
   Log_event_writer writer(&cache);
-  Format_description_log_event *ev=NULL;
+  Format_description_log_event *ev;
 
-  longlong thd_trx_seqno= (long long)wsrep_thd_trx_seqno(thd);
+  int len= my_snprintf(filename, PATH_MAX, "%s/GRA_%lld_%lld_v2.log",
+                       wsrep_data_home_dir, (longlong) thd->thread_id,
+                       (long long) wsrep_thd_trx_seqno(thd));
 
-  int len= snprintf(NULL, 0, "%s/GRA_%ld_%lld_v2.log",
-                    wsrep_data_home_dir, thd->thread_id,
-                    thd_trx_seqno);
-  /*
-    len doesn't count the \0 end-of-string. Use len+1 below
-    to alloc and pass as an argument to snprintf.
-  */
-  char *filename;
-  if (len < 0 || !(filename= (char*)malloc(len+1)))
+  if (len >= PATH_MAX)
   {
-    WSREP_ERROR("snprintf error: %d, skipping dump.", len);
-    DBUG_VOID_RETURN;
-  }
-
-  int len1= snprintf(filename, len+1, "%s/GRA_%ld_%lld_v2.log",
-                     wsrep_data_home_dir, thd->thread_id,
-                     thd_trx_seqno);
-
-  if (len > len1)
-  {
-    WSREP_ERROR("RBR dump path truncated: %d, skipping dump.", len);
-    free(filename);
+    WSREP_ERROR("RBR dump path too long: %d, skipping dump.", len);
     DBUG_VOID_RETURN;
   }
 
@@ -506,6 +470,7 @@ void wsrep_dump_rbr_buf_with_header(THD *thd, const void *rbr_buf,
 
   if (init_io_cache(&cache, file, 0, WRITE_CACHE, 0, 0, MYF(MY_WME | MY_NABP)))
   {
+    mysql_file_close(file, MYF(MY_WME));
     goto cleanup2;
   }
 
@@ -532,7 +497,6 @@ cleanup2:
   end_io_cache(&cache);
 
 cleanup1:
-  free(filename);
   mysql_file_close(file, MYF(MY_WME));
 
   if (!thd->wsrep_applier) delete ev;
