@@ -1,7 +1,7 @@
 /*********** File AM Zip C++ Program Source Code File (.CPP) ***********/
 /* PROGRAM NAME: FILAMZIP                                              */
 /* -------------                                                       */
-/*  Version 1.3                                                        */
+/*  Version 1.1                                                        */
 /*                                                                     */
 /* COPYRIGHT:                                                          */
 /* ----------                                                          */
@@ -45,12 +45,12 @@
 
 #define WRITEBUFFERSIZE (16384)
 
-bool ZipLoadFile(PGLOBAL g, PCSZ zfn, PCSZ fn, PCSZ entry, bool append, bool mul);
+bool ZipLoadFile(PGLOBAL g, char *zfn, char *fn, char *entry, bool append, bool mul);
 
 /***********************************************************************/
 /*  Compress a file in zip when creating a table.                      */
 /***********************************************************************/
-static bool ZipFile(PGLOBAL g, ZIPUTIL *zutp, PCSZ fn, PCSZ entry, char *buf)
+static bool ZipFile(PGLOBAL g, ZIPUTIL *zutp, char *fn, char *entry, char *buf)
 {
 	int   rc = RC_OK, size_read, size_buf = WRITEBUFFERSIZE;
 	FILE *fin;
@@ -88,7 +88,7 @@ static bool ZipFile(PGLOBAL g, ZIPUTIL *zutp, PCSZ fn, PCSZ entry, char *buf)
 /***********************************************************************/
 /*  Find and Compress several files in zip when creating a table.      */
 /***********************************************************************/
-static bool ZipFiles(PGLOBAL g, ZIPUTIL *zutp, PCSZ pat, char *buf)
+static bool ZipFiles(PGLOBAL g, ZIPUTIL *zutp, char *pat, char *buf)
 {
 	char filename[_MAX_PATH];
 	int  rc;
@@ -203,7 +203,7 @@ static bool ZipFiles(PGLOBAL g, ZIPUTIL *zutp, PCSZ pat, char *buf)
 /***********************************************************************/
 /*  Load and Compress a file in zip when creating a table.             */
 /***********************************************************************/
-bool ZipLoadFile(PGLOBAL g, PCSZ zfn, PCSZ fn, PCSZ entry, bool append, bool mul)
+bool ZipLoadFile(PGLOBAL g, char *zfn, char *fn, char *entry, bool append, bool mul)
 {
 	char    *buf;
 	bool     err;
@@ -228,7 +228,7 @@ bool ZipLoadFile(PGLOBAL g, PCSZ zfn, PCSZ fn, PCSZ entry, bool append, bool mul
 /***********************************************************************/
 /*  Constructors.                                                      */
 /***********************************************************************/
-ZIPUTIL::ZIPUTIL(PCSZ tgt)
+ZIPUTIL::ZIPUTIL(PSZ tgt)
 {
 	zipfile = NULL;
 	target = tgt;
@@ -269,7 +269,7 @@ void ZIPUTIL::getTime(tm_zip& tmZip)
 /*	append:		set true to append the zip file													 */
 /*  return:	true if open, false otherwise.														 */
 /***********************************************************************/
-bool ZIPUTIL::open(PGLOBAL g, PCSZ filename, bool append)
+bool ZIPUTIL::open(PGLOBAL g, char *filename, bool append)
 {
 	if (!zipfile && !(zipfile = zipOpen64(filename,
 		                               append ? APPEND_STATUS_ADDINZIP
@@ -290,15 +290,12 @@ void ZIPUTIL::close()
 		zipfile = NULL;
 	}	// endif zipfile
 
-	if (fp)
-		fp->Count = 0;
-
 }	// end of close
 
 /***********************************************************************/
 /*  OpenTableFile: Open a DOS/UNIX table file from a ZIP file.         */
 /***********************************************************************/
-bool ZIPUTIL::OpenTable(PGLOBAL g, MODE mode, PCSZ fn, bool append)
+bool ZIPUTIL::OpenTable(PGLOBAL g, MODE mode, char *fn, bool append)
 {
 	/*********************************************************************/
 	/*  The file will be compressed.                                     */
@@ -341,10 +338,10 @@ bool ZIPUTIL::OpenTable(PGLOBAL g, MODE mode, PCSZ fn, bool append)
 /***********************************************************************/
 /*  Add target in zip file.				   		      												 */
 /***********************************************************************/
-bool ZIPUTIL::addEntry(PGLOBAL g, PCSZ entry)
+bool ZIPUTIL::addEntry(PGLOBAL g, char *entry)
 {
 	//?? we dont need the stinking time
-	zip_fileinfo zi = { {0, 0, 0, 0, 0, 0}, 0, 0, 0 };
+	zip_fileinfo zi = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 
 	getTime(zi.tmz_date);
 	target = entry;
@@ -385,11 +382,10 @@ void ZIPUTIL::closeEntry()
 /***********************************************************************/
 /*  Constructors.                                                      */
 /***********************************************************************/
-UNZIPUTL::UNZIPUTL(PCSZ tgt, bool mul)
+UNZIPUTL::UNZIPUTL(PSZ tgt, bool mul)
 {
 	zipfile = NULL;
 	target = tgt;
-	pwd = NULL;
 	fp = NULL;
 	memory = NULL;
 	size = 0;
@@ -405,26 +401,6 @@ UNZIPUTL::UNZIPUTL(PCSZ tgt, bool mul)
 #endif
 } // end of UNZIPUTL standard constructor
 
-UNZIPUTL::UNZIPUTL(PDOSDEF tdp)
-{
-	zipfile = NULL;
-	target = tdp->GetEntry();
-	pwd = tdp->Pwd;
-	fp = NULL;
-	memory = NULL;
-	size = 0;
-	entryopen = false;
-	multiple = tdp->GetMul();
-	memset(fn, 0, sizeof(fn));
-
-	// Init the case mapping table.
-#if defined(__WIN__)
-	for (int i = 0; i < 256; ++i) mapCaseTable[i] = toupper(i);
-#else
-	for (int i = 0; i < 256; ++i) mapCaseTable[i] = i;
-#endif
-} // end of UNZIPUTL standard constructor
-
 #if 0
 UNZIPUTL::UNZIPUTL(PZIPUTIL zutp)
 {
@@ -442,8 +418,8 @@ UNZIPUTL::UNZIPUTL(PZIPUTIL zutp)
 /* This code is the copyright property of Alessandro Felice Cantatore. */
 /* http://xoomer.virgilio.it/acantato/dev/wildcard/wildmatch.html			 */
 /***********************************************************************/
-bool UNZIPUTL::WildMatch(PCSZ pat, PCSZ str) {
-	PCSZ s, p;
+bool UNZIPUTL::WildMatch(PSZ pat, PSZ str) {
+	PSZ  s, p;
 	bool star = FALSE;
 
 loopStart:
@@ -458,7 +434,7 @@ loopStart:
 			if (!*++pat) return TRUE;
 			goto loopStart;
 		default:
-			if (mapCaseTable[(uint)*s] != mapCaseTable[(uint)*p])
+			if (mapCaseTable[(uchar)*s] != mapCaseTable[(uchar)*p])
 				goto starCheck;
 			break;
 		} /* endswitch */
@@ -477,7 +453,7 @@ starCheck:
 /*  param: filename	path and the filename of the zip file to open.		 */
 /*  return:	true if open, false otherwise.														 */
 /***********************************************************************/
-bool UNZIPUTL::open(PGLOBAL g, PCSZ filename)
+bool UNZIPUTL::open(PGLOBAL g, char *filename)
 {
 	if (!zipfile && !(zipfile = unzOpen64(filename)))
 		sprintf(g->Message, "Zipfile open error on %s", filename);
@@ -496,9 +472,6 @@ void UNZIPUTL::close()
 		zipfile = NULL;
 	}	// endif zipfile
 
-	if (fp)
-		fp->Count = 0;
-
 }	// end of close
 
 /***********************************************************************/
@@ -570,7 +543,7 @@ int UNZIPUTL::nextEntry(PGLOBAL g)
 /***********************************************************************/
 /*  OpenTableFile: Open a DOS/UNIX table file from a ZIP file.         */
 /***********************************************************************/
-bool UNZIPUTL::OpenTable(PGLOBAL g, MODE mode, PCSZ fn)
+bool UNZIPUTL::OpenTable(PGLOBAL g, MODE mode, char *fn)
 {
 	/*********************************************************************/
 	/*  The file will be decompressed into virtual memory.               */
@@ -608,7 +581,7 @@ bool UNZIPUTL::OpenTable(PGLOBAL g, MODE mode, PCSZ fn)
 			if (openEntry(g))
 				return true;
 
-			if (size > 0) {
+			if (size > 0)	{
 				/*******************************************************************/
 				/*  Link a Fblock. This make possible to automatically close it    */
 				/*  in case of error g->jump.                                      */
@@ -639,28 +612,6 @@ bool UNZIPUTL::OpenTable(PGLOBAL g, MODE mode, PCSZ fn)
 	return false;
 } // end of OpenTableFile
 
-/***********************************************************************/
-/*  Insert only if the entry does not exist.   												 */
-/***********************************************************************/
-bool UNZIPUTL::IsInsertOk(PGLOBAL g, PCSZ fn)
-{
-	bool ok = true, b = open(g, fn);
-
-	if (!b) {
-		if (!target || *target == 0) {
-			unz_global_info64 ginfo;
-			int err = unzGetGlobalInfo64(zipfile, &ginfo);
-
-			ok = !(err == UNZ_OK && ginfo.number_entry > 0);
-		} else						// Check if the target exist
-			ok = (unzLocateFile(zipfile, target, 0) != UNZ_OK);
-
-		unzClose(zipfile);
-	} // endif b
-
-	return ok;
-} // end of IsInsertOk
-
 /***********************************************************************/
 /*  Open target in zip file.						      												 */
 /***********************************************************************/
@@ -674,24 +625,18 @@ bool UNZIPUTL::openEntry(PGLOBAL g)
 	if (rc != UNZ_OK) {
 		sprintf(g->Message, "unzGetCurrentFileInfo64 rc=%d", rc);
 		return true;
-	} else if ((rc = unzOpenCurrentFilePassword(zipfile, pwd)) != UNZ_OK) {
+	} else if ((rc = unzOpenCurrentFile(zipfile)) != UNZ_OK) {
 		sprintf(g->Message, "unzOpen fn=%s rc=%d", fn, rc);
 		return true;
 	}	// endif rc
 
 	size = finfo.uncompressed_size;
-
-	try {
-		memory = new char[size + 1];
-	} catch (...) {
-		strcpy(g->Message, "Out of memory");
-		return true;
-	} // end try/catch
+	memory = new char[size + 1];
 
 	if ((rc = unzReadCurrentFile(zipfile, memory, size)) < 0) {
 		sprintf(g->Message, "unzReadCurrentFile rc = %d", rc);
 		unzCloseCurrentFile(zipfile);
-		delete[] memory;
+		free(memory);
 		memory = NULL;
 		entryopen = false;
 	} else {
@@ -716,7 +661,7 @@ void UNZIPUTL::closeEntry()
 	}	// endif entryopen
 
 	if (memory) {
-		delete[] memory;
+		free(memory);
 		memory = NULL;
 	}	// endif memory
 
@@ -730,17 +675,15 @@ void UNZIPUTL::closeEntry()
 UNZFAM::UNZFAM(PDOSDEF tdp) : MAPFAM(tdp)
 {
 	zutp = NULL;
-	tdfp = tdp;
-  //target = tdp->GetEntry();
-	//mul = tdp->GetMul();
+  target = tdp->GetEntry();
+	mul = tdp->GetMul();
 } // end of UNZFAM standard constructor
 
 UNZFAM::UNZFAM(PUNZFAM txfp) : MAPFAM(txfp)
 {
 	zutp = txfp->zutp;
-	tdfp = txfp->tdfp;
-	//target = txfp->target;
-	//mul = txfp->mul;
+	target = txfp->target;
+	mul = txfp->mul;
 } // end of UNZFAM copy constructor
 
 /***********************************************************************/
@@ -768,13 +711,7 @@ int UNZFAM::Cardinality(PGLOBAL g)
 	int card = -1;
 	int len = GetFileLength(g);
 
-	if (len) {
-		// Estimated ???
-		card = (len / (int)Lrecl) * 2;
-		card = card ? card : 10;				// Lrecl can be too big 
-	} else
-		card = 0;
-
+	card = (len / (int)Lrecl) * 2;           // Estimated ???
 	return card;
 } // end of Cardinality
 
@@ -789,7 +726,7 @@ bool UNZFAM::OpenTableFile(PGLOBAL g)
 	/*********************************************************************/
 	/*  Allocate the ZIP utility class.                                  */
 	/*********************************************************************/
-	zutp = new(g) UNZIPUTL(tdfp);
+	zutp = new(g) UNZIPUTL(target, mul);
 
 	//  We used the file name relative to recorded datapath
 	PlugSetPath(filename, To_File, Tdbp->GetPath());
@@ -904,19 +841,17 @@ void UNZFAM::CloseTableFile(PGLOBAL g, bool)
 UZXFAM::UZXFAM(PDOSDEF tdp) : MPXFAM(tdp)
 {
 	zutp = NULL;
-	tdfp = tdp;
-	//target = tdp->GetEntry();
-	//mul = tdp->GetMul();
+	target = tdp->GetEntry();
+	mul = tdp->GetMul();
 	//Lrecl = tdp->GetLrecl();
 } // end of UZXFAM standard constructor
 
 UZXFAM::UZXFAM(PUZXFAM txfp) : MPXFAM(txfp)
 {
 	zutp = txfp->zutp;
-	tdfp = txfp->tdfp;
-	//target = txfp->target;
-	//mul = txfp->mul;
-	//Lrecl = txfp->Lrecl;
+	target = txfp->target;
+	mul = txfp->mul;
+//Lrecl = txfp->Lrecl;
 } // end of UZXFAM copy constructor
 
 /***********************************************************************/
@@ -972,7 +907,7 @@ bool UZXFAM::OpenTableFile(PGLOBAL g)
 		/*  Allocate the ZIP utility class.                                  */
 		/*********************************************************************/
 		if (!zutp)
-			zutp = new(g)UNZIPUTL(tdfp);
+			zutp = new(g)UNZIPUTL(target, mul);
 
 		//  We used the file name relative to recorded datapath
 		PlugSetPath(filename, To_File, Tdbp->GetPath());
@@ -1034,25 +969,6 @@ bool ZIPFAM::OpenTableFile(PGLOBAL g)
 {
 	char filename[_MAX_PATH];
 	MODE mode = Tdbp->GetMode();
-	int  len = TXTFAM::GetFileLength(g);
-
-	//  We used the file name relative to recorded datapath
-	PlugSetPath(filename, To_File, Tdbp->GetPath());
-
-	if (len < 0)
-		return true;
-	else if (!append && len > 0) {
-		strcpy(g->Message, "No insert into existing zip file");
-		return true;
-	} else if (append && len > 0) {
-		UNZIPUTL *zutp = new(g) UNZIPUTL(target, false);
-
-		if (!zutp->IsInsertOk(g, filename)) {
-			strcpy(g->Message, "No insert into existing entry");
-			return true;
-		}	// endif Ok
-
-	} // endif's
 
 	/*********************************************************************/
 	/*  Allocate the ZIP utility class.                                  */
@@ -1112,31 +1028,15 @@ ZPXFAM::ZPXFAM(PDOSDEF tdp) : FIXFAM(tdp)
 	target = tdp->GetEntry();
 	append = tdp->GetAppend();
 	//Lrecl = tdp->GetLrecl();
-} // end of ZPXFAM standard constructor
+} // end of UZXFAM standard constructor
 
 /***********************************************************************/
 /*  OpenTableFile: Open a DOS/UNIX table file from a ZIP file.         */
 /***********************************************************************/
 bool ZPXFAM::OpenTableFile(PGLOBAL g)
 {
-	char filename[_MAX_PATH];
-	MODE mode = Tdbp->GetMode();
-	int  len = TXTFAM::GetFileLength(g);
-
-	if (len < 0)
-		return true;
-	else if (!append && len > 0) {
-		strcpy(g->Message, "No insert into existing zip file");
-		return true;
-	} else if (append && len > 0) {
-		UNZIPUTL *zutp = new(g) UNZIPUTL(target, false);
-
-		if (!zutp->IsInsertOk(g, filename)) {
-			strcpy(g->Message, "No insert into existing entry");
-			return true;
-		}	// endif Ok
-
-	} // endif's
+	char    filename[_MAX_PATH];
+	MODE    mode = Tdbp->GetMode();
 
 	/*********************************************************************/
 	/*  Allocate the ZIP utility class.                                  */
