@@ -1,11 +1,11 @@
 /************* TabDos C++ Program Source Code File (.CPP) **************/
 /* PROGRAM NAME: TABDOS                                                */
 /* -------------                                                       */
-/*  Version 4.9.4                                                      */
+/*  Version 4.9.2                                                      */
 /*                                                                     */
 /* COPYRIGHT:                                                          */
 /* ----------                                                          */
-/*  (C) Copyright to the author Olivier BERTRAND          1998-2019    */
+/*  (C) Copyright to the author Olivier BERTRAND          1998-2016    */
 /*                                                                     */
 /* WHAT THIS PROGRAM DOES:                                             */
 /* -----------------------                                             */
@@ -98,7 +98,6 @@ DOSDEF::DOSDEF(void)
   Ofn = NULL;
 	Entry = NULL;
   To_Indx = NULL;
-	Pwd = NULL;
   Recfm = RECFM_VAR;
   Mapped = false;
 	Zipped = false;
@@ -132,8 +131,7 @@ bool DOSDEF::DefineAM(PGLOBAL g, LPCSTR am, int)
   bool   map = (am && (*am == 'M' || *am == 'm'));
   LPCSTR dfm = (am && (*am == 'F' || *am == 'f')) ? "F"
              : (am && (*am == 'B' || *am == 'b')) ? "B"
-		         : (am && (*am == 'X' || *am == 'x')) ? "X"
-		         : (am && !stricmp(am, "DBF"))        ? "D" : "V";
+             : (am && !stricmp(am, "DBF"))        ? "D" : "V";
 
 	if ((Zipped = GetBoolCatInfo("Zipped", false))) {
 		Entry = GetStringCatInfo(g, "Entry", NULL);
@@ -141,16 +139,14 @@ bool DOSDEF::DefineAM(PGLOBAL g, LPCSTR am, int)
 		                               : false;
 		Mulentries = GetBoolCatInfo("Mulentries", Mulentries);
 		Append = GetBoolCatInfo("Append", false);
-		Pwd = GetStringCatInfo(g, "Password", NULL);
-	}	// endif Zipped
+	}
 
   Desc = Fn = GetStringCatInfo(g, "Filename", NULL);
   Ofn = GetStringCatInfo(g, "Optname", Fn);
   GetCharCatInfo("Recfm", (PSZ)dfm, buf, sizeof(buf));
   Recfm = (toupper(*buf) == 'F') ? RECFM_FIX :
           (toupper(*buf) == 'B') ? RECFM_BIN :
-		      (toupper(*buf) == 'X') ? RECFM_NAF : // MGO
-		      (toupper(*buf) == 'D') ? RECFM_DBF : RECFM_VAR;
+          (toupper(*buf) == 'D') ? RECFM_DBF : RECFM_VAR;
   Lrecl = GetIntCatInfo("Lrecl", 0);
 
   if (Recfm != RECFM_DBF)
@@ -184,7 +180,7 @@ bool DOSDEF::DefineAM(PGLOBAL g, LPCSTR am, int)
 /***********************************************************************/
 bool DOSDEF::GetOptFileName(PGLOBAL g, char *filename)
   {
-  PCSZ ftype;
+  char   *ftype;
 
   switch (Recfm) {
     case RECFM_VAR: ftype = ".dop"; break;
@@ -241,9 +237,9 @@ void DOSDEF::RemoveOptValues(PGLOBAL g)
 /***********************************************************************/
 bool DOSDEF::DeleteIndexFile(PGLOBAL g, PIXDEF pxdf)
   {
-  PCSZ ftype;
-  char filename[_MAX_PATH];
-  bool sep, rc = false;
+  char   *ftype;
+  char    filename[_MAX_PATH];
+  bool    sep, rc = false;
 
   if (!To_Indx)
     return false;           // No index
@@ -355,7 +351,7 @@ PTDB DOSDEF::GetTable(PGLOBAL g, MODE mode)
 	if (Zipped) {
 #if defined(ZIP_SUPPORT)
 		if (Recfm == RECFM_VAR) {
-			if (mode == MODE_READ || mode == MODE_ANY || mode == MODE_ALTER) {
+			if (mode == MODE_READ || mode == MODE_ANY) {
 				txfp = new(g) UNZFAM(this);
 			} else if (mode == MODE_INSERT) {
 				txfp = new(g) ZIPFAM(this);
@@ -366,7 +362,7 @@ PTDB DOSDEF::GetTable(PGLOBAL g, MODE mode)
 
 			tdbp = new(g) TDBDOS(this, txfp);
 		} else {
-			if (mode == MODE_READ || mode == MODE_ANY || mode == MODE_ALTER) {
+			if (mode == MODE_READ || mode == MODE_ANY) {
 				txfp = new(g) UZXFAM(this);
 			} else if (mode == MODE_INSERT) {
 				txfp = new(g) ZPXFAM(this);
@@ -704,7 +700,7 @@ int TDBDOS::MakeBlockValues(PGLOBAL g)
 //      savmin = cdp->GetBmap();
 //      cdp->SetBmap(PlugSubAlloc(g, NULL, block * sizeof(int)));
 
-        if (trace(1))
+        if (trace)
           htrc("Dval(%p) Bmap(%p) col(%d) %s Block=%d lg=%d\n",
               cdp->GetDval(), cdp->GetBmap(), i, cdp->GetName(), block, lg);
 
@@ -729,7 +725,7 @@ int TDBDOS::MakeBlockValues(PGLOBAL g)
           memset(cdp->GetMax(), 0, block * lg);
           } // endif Type
 
-        if (trace(1))
+        if (trace)
           htrc("min(%p) max(%p) col(%d) %s Block=%d lg=%d\n",
               cdp->GetMin(), cdp->GetMax(), i, cdp->GetName(), block, lg);
 
@@ -901,7 +897,7 @@ bool TDBDOS::SaveBlockValues(PGLOBAL g)
             "wb", (int)errno, filename);
     strcat(strcat(g->Message, ": "), strerror(errno));
 
-    if (trace(1))
+    if (trace)
       htrc("%s\n", g->Message);
 
     return true;
@@ -1311,7 +1307,6 @@ PBF TDBDOS::InitBlockFilter(PGLOBAL g, PFIL filp)
         } // endif !opm
 
       // if opm, pass thru
-      // fall through
     case OP_IN:
       if (filp->GetArgType(0) == TYPE_COLBLK &&
           filp->GetArgType(1) == TYPE_ARRAY) {
@@ -1514,8 +1509,8 @@ PBF TDBDOS::CheckBlockFilari(PGLOBAL g, PXOB *arg, int op, bool *cnv)
     if (n == 8 && ctype != TYPE_LIST) {
       // Should never happen
       strcpy(g->Message, "Block opt: bad constant");
-			throw 99;
-		} // endif Conv
+      longjmp(g->jumper[g->jump_level], 99);
+      } // endif Conv
 
     if (type[0] == 1) {
       // Make it always as Column-op-Value
@@ -1634,7 +1629,7 @@ int TDBDOS::TestBlock(PGLOBAL g)
         To_Filter = NULL; // So remove filter
       } // endswitch Beval
 
-    if (trace(1))
+    if (trace)
       htrc("BF Eval Beval=%d\n", Beval);
 
     } // endif To_BlkFil
@@ -1647,8 +1642,8 @@ int TDBDOS::TestBlock(PGLOBAL g)
 /***********************************************************************/
 int TDBDOS::MakeIndex(PGLOBAL g, PIXDEF pxdf, bool add)
   {
-	int     k, n, rc = RC_OK;
-	bool    fixed, doit, sep, b = (pxdf != NULL);
+  int     k, n;
+  bool    fixed, doit, sep, b = (pxdf != NULL);
   PCOL   *keycols, colp;
   PIXDEF  xdp, sxp = NULL;
   PKPDEF  kdp;
@@ -1692,105 +1687,95 @@ int TDBDOS::MakeIndex(PGLOBAL g, PIXDEF pxdf, bool add)
   } else if (!(pxdf = dfp->GetIndx()))
     return RC_INFO;              // No index to make
 
-	try {
-		// Allocate all columns that will be used by indexes.
-	// This must be done before opening the table so specific
-	// column initialization can be done (in particular by TDBVCT)
-		for (n = 0, xdp = pxdf; xdp; xdp = xdp->GetNext())
-			for (kdp = xdp->GetToKeyParts(); kdp; kdp = kdp->GetNext()) {
-				if (!(colp = ColDB(g, kdp->GetName(), 0))) {
-					sprintf(g->Message, MSG(INDX_COL_NOTIN), kdp->GetName(), Name);
-					goto err;
-				} else if (colp->GetResultType() == TYPE_DECIM) {
-					sprintf(g->Message, "Decimal columns are not indexable yet");
-					goto err;
-				} // endif Type
+  // Allocate all columns that will be used by indexes.
+  // This must be done before opening the table so specific
+  // column initialization can be done (in particular by TDBVCT)
+  for (n = 0, xdp = pxdf; xdp; xdp = xdp->GetNext())
+    for (kdp = xdp->GetToKeyParts(); kdp; kdp = kdp->GetNext()) {
+      if (!(colp = ColDB(g, kdp->GetName(), 0))) {
+        sprintf(g->Message, MSG(INDX_COL_NOTIN), kdp->GetName(), Name);
+        goto err;
+      } else if (colp->GetResultType() == TYPE_DECIM) {
+        sprintf(g->Message, "Decimal columns are not indexable yet");
+        goto err;
+      } // endif Type
 
-				colp->InitValue(g);
-				n = MY_MAX(n, xdp->GetNparts());
-			} // endfor kdp
+      colp->InitValue(g);
+      n = MY_MAX(n, xdp->GetNparts());
+      } // endfor kdp
 
-		keycols = (PCOL*)PlugSubAlloc(g, NULL, n * sizeof(PCOL));
-		sep = dfp->GetBoolCatInfo("SepIndex", false);
+  keycols = (PCOL*)PlugSubAlloc(g, NULL, n * sizeof(PCOL));
+  sep = dfp->GetBoolCatInfo("SepIndex", false);
 
-		/*********************************************************************/
-		/*  Construct and save the defined indexes.                          */
-		/*********************************************************************/
-		for (xdp = pxdf; xdp; xdp = xdp->GetNext())
-			if (!OpenDB(g)) {
-				if (xdp->IsAuto() && fixed)
-					// Auto increment key and fixed file: use an XXROW index
-					continue;      // XXROW index doesn't need to be made
+  /*********************************************************************/
+  /*  Construct and save the defined indexes.                          */
+  /*********************************************************************/
+  for (xdp = pxdf; xdp; xdp = xdp->GetNext())
+    if (!OpenDB(g)) {
+      if (xdp->IsAuto() && fixed)
+        // Auto increment key and fixed file: use an XXROW index
+        continue;      // XXROW index doesn't need to be made
 
-				// On Update, redo only indexes that are modified
-				doit = !To_SetCols;
-				n = 0;
+      // On Update, redo only indexes that are modified
+      doit = !To_SetCols;
+      n = 0;
 
-				if (sxp)
-					xdp->SetID(sxp->GetID() + 1);
+      if (sxp)
+        xdp->SetID(sxp->GetID() + 1);
 
-				for (kdp = xdp->GetToKeyParts(); kdp; kdp = kdp->GetNext()) {
-					// Check whether this column was updated
-					for (colp = To_SetCols; !doit && colp; colp = colp->GetNext())
-						if (!stricmp(kdp->GetName(), colp->GetName()))
-							doit = true;
+      for (kdp = xdp->GetToKeyParts(); kdp; kdp = kdp->GetNext()) {
+        // Check whether this column was updated
+        for (colp = To_SetCols; !doit && colp; colp = colp->GetNext())
+          if (!stricmp(kdp->GetName(), colp->GetName()))
+            doit = true;
 
-					keycols[n++] = ColDB(g, kdp->GetName(), 0);
-				} // endfor kdp
+        keycols[n++] = ColDB(g, kdp->GetName(), 0);
+        } // endfor kdp
 
-			// If no indexed columns were updated, don't remake the index
-			// if indexes are in separate files.
-				if (!doit && sep)
-					continue;
+      // If no indexed columns were updated, don't remake the index
+      // if indexes are in separate files.
+      if (!doit && sep)
+        continue;
 
-				k = xdp->GetNparts();
+      k = xdp->GetNparts();
 
-				// Make the index and save it
-				if (dfp->Huge)
-					pxp = new(g) XHUGE;
-				else
-					pxp = new(g) XFILE;
+      // Make the index and save it
+      if (dfp->Huge)
+        pxp = new(g) XHUGE;
+      else
+        pxp = new(g) XFILE;
 
-				if (k == 1)            // Simple index
-					x = new(g) XINDXS(this, xdp, pxp, keycols);
-				else                   // Multi-Column index
-					x = new(g) XINDEX(this, xdp, pxp, keycols);
+      if (k == 1)            // Simple index
+        x = new(g) XINDXS(this, xdp, pxp, keycols);
+      else                   // Multi-Column index
+        x = new(g) XINDEX(this, xdp, pxp, keycols);
 
-				if (!x->Make(g, sxp)) {
-					// Retreive define values from the index
-					xdp->SetMaxSame(x->GetMaxSame());
-					//      xdp->SetSize(x->GetSize());
+      if (!x->Make(g, sxp)) {
+        // Retreive define values from the index
+        xdp->SetMaxSame(x->GetMaxSame());
+//      xdp->SetSize(x->GetSize());
 
-									// store KXYCOL Mxs in KPARTDEF Mxsame
-					xdp->SetMxsame(x);
+        // store KXYCOL Mxs in KPARTDEF Mxsame
+        xdp->SetMxsame(x);
 
 #if defined(TRACE)
-					printf("Make done...\n");
+        printf("Make done...\n");
 #endif   // TRACE
 
-					//      if (x->GetSize() > 0)
-					sxp = xdp;
+//      if (x->GetSize() > 0)
+          sxp = xdp;
 
-					xdp->SetInvalid(false);
-				} else
-					goto err;
+        xdp->SetInvalid(false);
+      } else
+        goto err;
 
-			} else
-				return RC_INFO;     // Error or Physical table does not exist
+    } else
+      return RC_INFO;     // Error or Physical table does not exist
 
-	} catch (int n) {
-		if (trace(1))
-			htrc("Exception %d: %s\n", n, g->Message);
-		rc = RC_FX;
-	} catch (const char *msg) {
-		strcpy(g->Message, msg);
-		rc = RC_FX;
-	} // end catch
-
-	if (Use == USE_OPEN)
+  if (Use == USE_OPEN)
     CloseDB(g);
 
-  return rc;
+  return RC_OK;
 
 err:
   if (sxp)
@@ -1805,8 +1790,8 @@ err:
 /*  Make a dynamic index.                                              */
 /***********************************************************************/
 bool TDBDOS::InitialyzeIndex(PGLOBAL g, volatile PIXDEF xdp, bool sorted)
-{
-  int     k;
+  {
+  int     k, rc;
   volatile bool dynamic;
   bool    brc;
   PCOL    colp;
@@ -1876,7 +1861,13 @@ bool TDBDOS::InitialyzeIndex(PGLOBAL g, volatile PIXDEF xdp, bool sorted)
   } else                      // Column contains same values as ROWID
     kxp = new(g) XXROW(this);
 
-	try {
+  //  Prepare error return
+  if (g->jump_level == MAX_JUMP) {
+    strcpy(g->Message, MSG(TOO_MANY_JUMPS));
+    return true;
+    } // endif
+
+  if (!(rc = setjmp(g->jumper[++g->jump_level])) != 0) {
     if (dynamic) {
       ResetBlockFilter(g);
       kxp->SetDynamic(dynamic);
@@ -1901,17 +1892,12 @@ bool TDBDOS::InitialyzeIndex(PGLOBAL g, volatile PIXDEF xdp, bool sorted)
 
       } // endif brc
 
-	} catch (int n) {
-		if (trace(1))
-			htrc("Exception %d: %s\n", n, g->Message);
-		brc = true;
-	} catch (const char *msg) {
-		strcpy(g->Message, msg);
-		brc = true;
-	} // end catch
+  } else
+    brc = true;
 
-	return brc;
-} // end of InitialyzeIndex
+  g->jump_level--;
+  return brc;
+  } // end of InitialyzeIndex
 
 /***********************************************************************/
 /*  DOS GetProgMax: get the max value for progress information.        */
@@ -2001,7 +1987,7 @@ int TDBDOS::Cardinality(PGLOBAL g)
         if (len >= 0) {
           int rec;
 
-          if (trace(1))
+          if (trace)
             htrc("Estimating lines len=%d ending=%d/n",
                   len, ((PDOSDEF)To_Def)->Ending);
 
@@ -2018,7 +2004,7 @@ int TDBDOS::Cardinality(PGLOBAL g)
 
           Cardinal = (len + rec - 1) / rec;
 
-          if (trace(1))
+          if (trace)
             htrc("avglen=%d MaxSize%d\n", rec, Cardinal);
 
           } // endif len
@@ -2048,7 +2034,7 @@ int TDBDOS::GetMaxSize(PGLOBAL g)
     if (len >= 0) {
       int rec;
 
-      if (trace(1))
+      if (trace)
         htrc("Estimating lines len=%d ending=%d/n",
               len, ((PDOSDEF)To_Def)->Ending);
 
@@ -2059,7 +2045,7 @@ int TDBDOS::GetMaxSize(PGLOBAL g)
       rec = EstimatedLength() + ((PDOSDEF)To_Def)->Ending;
       MaxSize = (len + rec - 1) / rec;
 
-      if (trace(1))
+      if (trace)
         htrc("avglen=%d MaxSize%d\n", rec, MaxSize);
 
       } // endif len
@@ -2108,7 +2094,7 @@ bool TDBDOS::IsUsingTemp(PGLOBAL)
 /***********************************************************************/
 bool TDBDOS::OpenDB(PGLOBAL g)
   {
-  if (trace(1))
+  if (trace)
     htrc("DOS OpenDB: tdbp=%p tdb=R%d use=%d mode=%d\n",
           this, Tdb_No, Use, Mode);
 
@@ -2132,8 +2118,7 @@ bool TDBDOS::OpenDB(PGLOBAL g)
     return false;
     } // endif use
 
-  if (Mode == MODE_DELETE && !Next && Txfp->GetAmType() != TYPE_AM_DOS
-		                               && Txfp->GetAmType() != TYPE_AM_MGO) {
+  if (Mode == MODE_DELETE && !Next && Txfp->GetAmType() != TYPE_AM_DOS) {
     // Delete all lines. Not handled in MAP or block mode
     Txfp = new(g) DOSFAM((PDOSDEF)To_Def);
     Txfp->SetTdbp(this);
@@ -2171,20 +2156,18 @@ bool TDBDOS::OpenDB(PGLOBAL g)
   To_BlkFil = InitBlockFilter(g, To_Filter);
 
   /*********************************************************************/
-	/*  Lrecl does not include line ending															 */
-	/*********************************************************************/
-	size_t linelen = Lrecl + ((PDOSDEF)To_Def)->Ending + 1;
-
-  To_Line = (char*)PlugSubAlloc(g, NULL, linelen);
+  /*  Allocate the line buffer plus a null character.                  */
+  /*********************************************************************/
+  To_Line = (char*)PlugSubAlloc(g, NULL, Lrecl + 1);
 
   if (Mode == MODE_INSERT) {
     // Spaces between fields must be filled with blanks
     memset(To_Line, ' ', Lrecl);
     To_Line[Lrecl] = '\0';
   } else
-    memset(To_Line, 0, linelen);
+    memset(To_Line, 0, Lrecl + 1);
 
-  if (trace(1))
+  if (trace)
     htrc("OpenDos: R%hd mode=%d To_Line=%p\n", Tdb_No, Mode, To_Line);
 
   if (SkipHeader(g))         // When called from CSV/FMT files
@@ -2202,7 +2185,7 @@ bool TDBDOS::OpenDB(PGLOBAL g)
 /***********************************************************************/
 int TDBDOS::ReadDB(PGLOBAL g)
   {
-  if (trace(2))
+  if (trace > 1)
     htrc("DOS ReadDB: R%d Mode=%d key=%p link=%p Kindex=%p To_Line=%p\n",
           GetTdb_No(), Mode, To_Key_Col, To_Link, To_Kindex, To_Line);
 
@@ -2227,7 +2210,7 @@ int TDBDOS::ReadDB(PGLOBAL g)
         if (SetRecpos(g, recpos))
           return RC_FX;
 
-        if (trace(2))
+        if (trace > 1)
           htrc("File position is now %d\n", GetRecpos());
 
         if (Mode == MODE_READ)
@@ -2243,7 +2226,7 @@ int TDBDOS::ReadDB(PGLOBAL g)
 
     } // endif To_Kindex
 
-  if (trace(2))
+  if (trace > 1)
     htrc(" ReadDB: this=%p To_Line=%p\n", this, To_Line);
 
   /*********************************************************************/
@@ -2279,14 +2262,14 @@ bool TDBDOS::PrepareWriting(PGLOBAL)
 /***********************************************************************/
 int TDBDOS::WriteDB(PGLOBAL g)
   {
-  if (trace(2))
+  if (trace > 1)
     htrc("DOS WriteDB: R%d Mode=%d \n", Tdb_No, Mode);
 
   // Make the line to write
   if (PrepareWriting(g))
     return RC_FX;
 
-  if (trace(2))
+  if (trace > 1)
     htrc("Write: line is='%s'\n", To_Line);
 
   // Now start the writing process
@@ -2321,8 +2304,8 @@ void TDBDOS::CloseDB(PGLOBAL g)
 /***********************************************************************/
 /*  DOSCOL public constructor (also called by MAPCOL).                 */
 /***********************************************************************/
-DOSCOL::DOSCOL(PGLOBAL g, PCOLDEF cdp, PTDB tp, PCOL cp, int i, PCSZ am)
-      : COLBLK(cdp, tp, i)
+DOSCOL::DOSCOL(PGLOBAL g, PCOLDEF cdp, PTDB tp, PCOL cp, int i, PSZ am)
+  : COLBLK(cdp, tp, i)
   {
   char *p;
   int   prec = Format.Prec;
@@ -2352,7 +2335,7 @@ DOSCOL::DOSCOL(PGLOBAL g, PCOLDEF cdp, PTDB tp, PCOL cp, int i, PCSZ am)
   Dval = NULL;
   Buf = NULL;
 
-  if (txfp && txfp->Blocked && Opt && (cdp->GetMin() || cdp->GetDval())) {
+  if (txfp->Blocked && Opt && (cdp->GetMin() || cdp->GetDval())) {
     int nblk = txfp->GetBlock();
 
     Clustered = (cdp->GetXdb2()) ? 2 : 1;
@@ -2403,7 +2386,7 @@ DOSCOL::DOSCOL(PGLOBAL g, PCOLDEF cdp, PTDB tp, PCOL cp, int i, PCSZ am)
     Dcm = (*p) ? atoi(p) : GetScale();
     } // endif fmt
 
-  if (trace(1))
+  if (trace)
     htrc(" making new %sCOL C%d %s at %p\n", am, Index, Name, this);
 
   } // end of DOSCOL constructor
@@ -2492,10 +2475,8 @@ bool DOSCOL::SetBuffer(PGLOBAL g, PVAL value, bool ok, bool check)
   } // endif's Value, Buf_Type
 
   // Allocate the buffer used in WriteColumn for numeric columns
-	if (!Buf && IsTypeNum(Buf_Type))
-		Buf = (char*)PlugSubAlloc(g, NULL, MY_MAX(64, Long + 1));
-	else // Text columns do not need additional buffer
-		Buf = (char*)Value->GetTo_Val();
+  if (!Buf && IsTypeNum(Buf_Type))
+    Buf = (char*)PlugSubAlloc(g, NULL, MY_MAX(32, Long + Dcm + 1));
 
   // Because Colblk's have been made from a copy of the original TDB in
   // case of Update, we must reset them to point to the original one.
@@ -2520,7 +2501,7 @@ void DOSCOL::ReadColumn(PGLOBAL g)
   double  dval;
   PTDBDOS tdbp = (PTDBDOS)To_Tdb;
 
-  if (trace(2))
+  if (trace > 1)
     htrc(
       "DOS ReadColumn: col %s R%d coluse=%.4X status=%.4X buf_type=%d\n",
          Name, tdbp->GetTdb_No(), ColUse, Status, Buf_Type);
@@ -2533,8 +2514,8 @@ void DOSCOL::ReadColumn(PGLOBAL g)
       if (rc == RC_EF)
         sprintf(g->Message, MSG(INV_DEF_READ), rc);
 
-			throw 11;
-		} // endif
+      longjmp(g->jumper[g->jump_level], 11);
+      } // endif
 
   p = tdbp->To_Line + Deplac;
   field = Long;
@@ -2589,8 +2570,8 @@ void DOSCOL::ReadColumn(PGLOBAL g)
       break;
     default:
       sprintf(g->Message, MSG(BAD_RECFM), tdbp->Ftype);
-			throw 34;
-	} // endswitch Ftype
+      longjmp(g->jumper[g->jump_level], 34);
+    } // endswitch Ftype
 
   // Set null when applicable
   if (Nullable)
@@ -2605,17 +2586,17 @@ void DOSCOL::ReadColumn(PGLOBAL g)
 /***********************************************************************/
 void DOSCOL::WriteColumn(PGLOBAL g)
   {
-  char   *p, fmt[32];
-  int     i, k, n, len, field;
+  char   *p, *p2, fmt[32];
+  int     i, k, len, field;
   PTDBDOS tdbp = (PTDBDOS)To_Tdb;
 
-  if (trace(2))
+  if (trace > 1)
     htrc("DOS WriteColumn: col %s R%d coluse=%.4X status=%.4X\n",
           Name, tdbp->GetTdb_No(), ColUse, Status);
 
   p = tdbp->To_Line + Deplac;
 
-  if (trace(2))
+  if (trace > 1)
     htrc("Lrecl=%d deplac=%d int=%d\n", tdbp->Lrecl, Deplac, Long);
 
   field = Long;
@@ -2632,7 +2613,7 @@ void DOSCOL::WriteColumn(PGLOBAL g)
 
     } // endif Ftype
 
-  if (trace(2))
+  if (trace > 1)
     htrc("Long=%d field=%d coltype=%d colval=%p\n",
           Long, field, Buf_Type, Value);
 
@@ -2681,8 +2662,8 @@ void DOSCOL::WriteColumn(PGLOBAL g)
         case TYPE_DOUBLE:
         case TYPE_DECIM:
           strcpy(fmt, (Ldz) ? "%0*.*lf" : "%*.*lf");
-					len = field + ((Nod && Dcm) ? 1 : 0);
-          snprintf(Buf, len, fmt, len, Dcm, Value->GetFloatValue());
+          sprintf(Buf, fmt, field + ((Nod && Dcm) ? 1 : 0),
+                  Dcm, Value->GetFloatValue());
           len = strlen(Buf);
 
           if (Nod && Dcm)
@@ -2698,40 +2679,38 @@ void DOSCOL::WriteColumn(PGLOBAL g)
           break;
         default:
           sprintf(g->Message, "Invalid field format for column %s", Name);
-					throw 31;
-			} // endswitch BufType
+          longjmp(g->jumper[g->jump_level], 31);
+        } // endswitch BufType
 
-			n = strlen(Buf);
+      p2 = Buf;
     } else                 // Standard CONNECT format
-      n = Value->ShowValue(Buf, field);
+      p2 = Value->ShowValue(Buf, field);
 
-    if (trace(1))
-      htrc("new length(%p)=%d\n", Buf, n);
+    if (trace)
+      htrc("new length(%p)=%d\n", p2, strlen(p2));
 
-    if ((len = n) > field) {
-			char *p = Value->GetCharString(Buf);
-
-      sprintf(g->Message, MSG(VALUE_TOO_LONG), p, Name, field);
-			throw 31;
-		} else if (Dsp)
+    if ((len = strlen(p2)) > field) {
+      sprintf(g->Message, MSG(VALUE_TOO_LONG), p2, Name, field);
+      longjmp(g->jumper[g->jump_level], 31);
+    } else if (Dsp)
       for (i = 0; i < len; i++)
-        if (Buf[i] == '.')
-          Buf[i] = Dsp; 
+        if (p2[i] == '.')
+          p2[i] = Dsp; 
 
-    if (trace(2))
-      htrc("buffer=%s\n", Buf);
+    if (trace > 1)
+      htrc("buffer=%s\n", p2);
 
     /*******************************************************************/
     /*  Updating must be done only when not in checking pass.          */
     /*******************************************************************/
     if (Status) {
       memset(p, ' ', field);
-      memcpy(p, Buf, len);
+      memcpy(p, p2, len);
 
-      if (trace(2))
+      if (trace > 1)
         htrc(" col write: '%.*s'\n", len, p);
 
-    } // endif Status
+      } // endif Use
 
   } else    // BIN compressed table
     /*******************************************************************/
@@ -2741,8 +2720,8 @@ void DOSCOL::WriteColumn(PGLOBAL g)
     if (Value->GetBinValue(p, Long, Status)) {
       sprintf(g->Message, MSG(BIN_F_TOO_LONG),
                           Name, Value->GetSize(), Long);
-      throw 31;
-    } // endif
+      longjmp(g->jumper[g->jump_level], 31);
+      } // endif
 
   } // end of WriteColumn
 
@@ -2884,4 +2863,13 @@ bool DOSCOL::AddDistinctValue(PGLOBAL g)
   return false;
   } // end of AddDistinctValue
 
+/***********************************************************************/
+/*  Make file output of a Dos column descriptor block.                 */
+/***********************************************************************/
+void DOSCOL::Print(PGLOBAL g, FILE *f, uint n)
+  {
+  COLBLK::Print(g, f, n);
+  } // end of Print
+
 /* ------------------------------------------------------------------- */
+
