@@ -340,7 +340,7 @@ err:
       for (j=0 ; j < share->base.keys ; j++)
         maria_flush_bulk_insert(info, j);
     }
-    info->errkey= i < share->base.keys ? (int) i : -1;
+    info->errkey= (int) i;
     /*
       We delete keys in the reverse order of insertion. This is the order that
       a rollback would do and is important for CLR_ENDs generated by
@@ -880,7 +880,7 @@ ChangeSet@1.2562, 2008-04-09 07:41:40+02:00, serg@janus.mylan +9 -0
       DBUG_ASSERT(info->ft1_to_ft2==0);
       if (alen == blen &&
           ha_compare_text(keyinfo->seg->charset, a, alen,
-                          b, blen, 0, 0) == 0)
+                          b, blen, 0) == 0)
       {
         /* Yup. converting */
         info->ft1_to_ft2=(DYNAMIC_ARRAY *)
