@@ -191,8 +191,9 @@ int _create_index_by_sort(MI_SORT_PARAM *info,my_bool no_messages,
       while ((maxbuffer= (uint) (records/(keys-1)+1)) != maxbuffer_org);
     }
 
-    if ((sort_keys=(uchar **)my_malloc(keys*(sort_length+sizeof(char*))+
-				       HA_FT_MAXBYTELEN, MYF(0))))
+    if ((sort_keys= ((uchar **)
+                     my_malloc((size_t) (keys*(sort_length+sizeof(char*))+
+                                         HA_FT_MAXBYTELEN), MYF(0)))))
     {
       if (my_init_dynamic_array(&buffpek, sizeof(BUFFPEK), maxbuffer,
                                 MY_MIN(maxbuffer/2, 1000), MYF(0)))
@@ -344,192 +345,197 @@ static ha_rows find_all_keys(MI_SORT_PARAM *info, ha_rows keys,
   DBUG_RETURN((*maxbuffer)*(keys-1)+idx);
 } /* find_all_keys */
 
+static my_bool thr_find_all_keys_exec(MI_SORT_PARAM *sort_param)
+{
+  ulonglong memavl, old_memavl, sortbuff_size;
+  ha_keys UNINIT_VAR(keys), idx;
+  uint sort_length;
+  uint maxbuffer;
+  uchar **sort_keys= NULL;
+  int error= 0;
+  DBUG_ENTER("thr_find_all_keys");
+  DBUG_PRINT("enter", ("master: %d", sort_param->master));
+
+  if (sort_param->sort_info->got_error)
+    DBUG_RETURN(TRUE);
+
+  set_sort_param_read_write(sort_param);
+
+  my_b_clear(&sort_param->tempfile);
+  my_b_clear(&sort_param->tempfile_for_exceptions);
+  bzero((char*) &sort_param->buffpek, sizeof(sort_param->buffpek));
+  bzero((char*) &sort_param->unique,  sizeof(sort_param->unique));
+
+  sortbuff_size= sort_param->sortbuff_size;
+  memavl=       MY_MAX(sortbuff_size, MIN_SORT_BUFFER);
+  idx=          (ha_keys) sort_param->sort_info->max_records;
+  sort_length=  sort_param->key_length;
+  maxbuffer=    1;
+
+  while (memavl >= MIN_SORT_BUFFER)
+  {
+    if ((my_off_t) (idx+1)*(sort_length+sizeof(char*)) <=
+        (my_off_t) memavl)
+      keys= idx+1;
+    else if ((sort_param->sort_info->param->testflag &
+              (T_FORCE_SORT_MEMORY | T_CREATE_MISSING_KEYS)) ==
+             T_FORCE_SORT_MEMORY)
+    {
+      /*
+        Use all of the given sort buffer for key data.
+        Allocate 1000 buffers at a start for new data. More buffers
+        will be allocated when needed.
+      */
+      keys= memavl / (sort_length+sizeof(char*));
+      maxbuffer= (uint) MY_MIN((ulonglong) 1000, (idx / keys)+1);
+    }
+    else
+    {
+      uint maxbuffer_org;
+      do
+      {
+        maxbuffer_org= maxbuffer;
+        if (memavl < sizeof(BUFFPEK)*maxbuffer ||
+            (keys=(memavl-sizeof(BUFFPEK)*maxbuffer)/
+             (sort_length+sizeof(char*))) <= 1 ||
+            keys < (uint) maxbuffer)
+        {
+          mi_check_print_error(sort_param->sort_info->param,
+                               "myisam_sort_buffer_size is too small. Current myisam_sort_buffer_size: %llu  rows: %llu  sort_length: %u",
+                               sortbuff_size, (ulonglong) idx, sort_length);
+          DBUG_RETURN(TRUE);
+        }
+      }
+      while ((maxbuffer= (uint) (idx/(keys-1)+1)) != maxbuffer_org);
+    }
+    if ((sort_keys= (uchar**) my_malloc((size_t)(keys * (sort_length + sizeof(char*)) +
+                   ((sort_param->keyinfo->flag & HA_FULLTEXT) ?
+                    HA_FT_MAXBYTELEN : 0)), MYF(0))))
+    {
+      if (my_init_dynamic_array(&sort_param->buffpek, sizeof(BUFFPEK),
+                                maxbuffer, MY_MIN(maxbuffer / 2, 1000), MYF(0)))
+      {
+        my_free(sort_keys);
+        sort_keys= NULL;          /* Safety against double free on error. */
+      }
+      else
+        break;
+    }
+    old_memavl= memavl;
+    if ((memavl= memavl / 4 * 3) < MIN_SORT_BUFFER &&
+        old_memavl > MIN_SORT_BUFFER)
+      memavl= MIN_SORT_BUFFER;
+  }
+  if (memavl < MIN_SORT_BUFFER)
+  {
+    /* purecov: begin inspected */
+    mi_check_print_error(sort_param->sort_info->param,
+                         "myisam_sort_buffer_size is too small. Current myisam_sort_buffer_size: %llu  rows: %llu  sort_length: %u",
+                         sortbuff_size, (ulonglong) idx, sort_length);
+    my_errno= ENOMEM;
+    goto err;
+  /* purecov: end inspected */
+  }
+
+  if (sort_param->sort_info->param->testflag & T_VERBOSE)
+    my_fprintf(stdout,
+               "Key %d - Allocating buffer for %llu keys\n",
+               sort_param->key + 1, (ulonglong) keys);
+  sort_param->sort_keys= sort_keys;
+
+  idx= error= 0;
+  sort_keys[0]= (uchar*) (sort_keys+keys);
+
+  DBUG_PRINT("info", ("reading keys"));
+  while (!(error= sort_param->sort_info->got_error) &&
+         !(error= (*sort_param->key_read)(sort_param, sort_keys[idx])))
+  {
+    if (sort_param->real_key_length > sort_param->key_length)
+    {
+      if (write_key(sort_param, sort_keys[idx],
+                    &sort_param->tempfile_for_exceptions))
+        goto err;
+      continue;
+    }
+
+    if (++idx == keys)
+    {
+      if (sort_param->write_keys(sort_param, sort_keys, idx - 1,
+                                 (BUFFPEK*) alloc_dynamic(&sort_param->buffpek),
+                                 &sort_param->tempfile))
+        goto err;
+      sort_keys[0]= (uchar*) (sort_keys+keys);
+      memcpy(sort_keys[0], sort_keys[idx - 1], (size_t) sort_param->key_length);
+      idx= 1;
+    }
+    sort_keys[idx]= sort_keys[idx - 1] + sort_param->key_length;
+  }
+
+  if (error > 0)
+    goto err;
+
+  if (sort_param->buffpek.elements)
+  {
+    if (sort_param->write_keys(sort_param, sort_keys, idx,
+                               (BUFFPEK*) alloc_dynamic(&sort_param->buffpek),
+                               &sort_param->tempfile))
+      goto err;
+    sort_param->keys= (sort_param->buffpek.elements - 1) * (keys - 1) + idx;
+  }
+  else
+    sort_param->keys= idx;
+
+  DBUG_RETURN(FALSE);
+
+err:
+  DBUG_PRINT("error", ("got some error"));
+  sort_param->sort_info->got_error= 1; /* no need to protect with a mutex */
+  my_free(sort_keys);
+  sort_param->sort_keys= 0;
+  delete_dynamic(& sort_param->buffpek);
+  close_cached_file(&sort_param->tempfile);
+  close_cached_file(&sort_param->tempfile_for_exceptions);
+
+  DBUG_RETURN(TRUE);
+}
+
 /* Search after all keys and place them in a temp. file */
 
 pthread_handler_t thr_find_all_keys(void *arg)
 {
   MI_SORT_PARAM *sort_param= (MI_SORT_PARAM*) arg;
-  int error;
-  ulonglong memavl, old_memavl, sortbuff_size;
-  ha_keys UNINIT_VAR(keys), idx;
-  uint sort_length;
-  uint maxbuffer;
-  uchar **sort_keys=0;
+  my_bool error= FALSE;
+  /* If my_thread_init fails */
+  if (my_thread_init() || thr_find_all_keys_exec(sort_param))
+    error= TRUE;
 
-  error=1;
+  /*
+     Thread must clean up after itself.
+  */
+  free_root(&sort_param->wordroot, MYF(0));
+  /*
+    Detach from the share if the writer is involved. Avoid others to
+    be blocked. This includes a flush of the write buffer. This will
+    also indicate EOF to the readers.
+    That means that a writer always gets here first and readers -
+    only when they see EOF. But if a reader finishes prematurely
+    because of an error it may reach this earlier - don't allow it
+    to detach the writer thread.
+  */
+  if (sort_param->master && sort_param->sort_info->info->rec_cache.share)
+    remove_io_thread(&sort_param->sort_info->info->rec_cache);
 
-  if (my_thread_init())
-    goto err;
+  /* Readers detach from the share if any. Avoid others to be blocked. */
+  if (sort_param->read_cache.share)
+    remove_io_thread(&sort_param->read_cache);
 
-  { /* Add extra block since DBUG_ENTER declare variables */
-    DBUG_ENTER("thr_find_all_keys");
-    DBUG_PRINT("enter", ("master: %d", sort_param->master));
-    if (sort_param->sort_info->got_error)
-      goto err;
+  mysql_mutex_lock(&sort_param->sort_info->mutex);
+  if (error)
+    sort_param->sort_info->got_error= 1;
 
-    set_sort_param_read_write(sort_param);
-
-    my_b_clear(&sort_param->tempfile);
-    my_b_clear(&sort_param->tempfile_for_exceptions);
-    bzero((char*) &sort_param->buffpek, sizeof(sort_param->buffpek));
-    bzero((char*) &sort_param->unique,  sizeof(sort_param->unique));
-    sort_keys= (uchar **) NULL;
-
-    sortbuff_size= sort_param->sortbuff_size;
-    memavl=       MY_MAX(sortbuff_size, MIN_SORT_BUFFER);
-    idx=          (ha_keys) sort_param->sort_info->max_records;
-    sort_length=  sort_param->key_length;
-    maxbuffer=    1;
-
-    while (memavl >= MIN_SORT_BUFFER)
-    {
-      if ((my_off_t) (idx+1)*(sort_length+sizeof(char*)) <=
-          (my_off_t) memavl)
-        keys= idx+1;
-      else if ((sort_param->sort_info->param->testflag &
-                (T_FORCE_SORT_MEMORY | T_CREATE_MISSING_KEYS)) ==
-               T_FORCE_SORT_MEMORY)
-      {
-        /*
-          Use all of the given sort buffer for key data.
-          Allocate 1000 buffers at a start for new data. More buffers
-          will be allocated when needed.
-        */
-        keys= memavl / (sort_length+sizeof(char*));
-        maxbuffer= (uint) MY_MIN((ulonglong) 1000, (idx / keys)+1);
-      }
-      else
-      {
-        uint maxbuffer_org;
-        do
-        {
-          maxbuffer_org= maxbuffer;
-          if (memavl < sizeof(BUFFPEK)*maxbuffer ||
-              (keys=(memavl-sizeof(BUFFPEK)*maxbuffer)/
-               (sort_length+sizeof(char*))) <= 1 ||
-              keys < (uint) maxbuffer)
-          {
-            mi_check_print_error(sort_param->sort_info->param,
-                                 "myisam_sort_buffer_size is too small. Current myisam_sort_buffer_size: %llu  rows: %llu  sort_length: %u",
-                                 sortbuff_size, (ulonglong) idx, sort_length);
-            goto err;
-          }
-        }
-        while ((maxbuffer= (uint) (idx/(keys-1)+1)) != maxbuffer_org);
-      }
-      if ((sort_keys= (uchar**)
-           my_malloc(keys*(sort_length+sizeof(char*))+
-                     ((sort_param->keyinfo->flag & HA_FULLTEXT) ?
-                      HA_FT_MAXBYTELEN : 0), MYF(0))))
-      {
-        if (my_init_dynamic_array(&sort_param->buffpek, sizeof(BUFFPEK),
-                                  maxbuffer, MY_MIN(maxbuffer/2, 1000), MYF(0)))
-        {
-          my_free(sort_keys);
-          sort_keys= (uchar **) NULL; /* for err: label */
-        }
-        else
-          break;
-      }
-      old_memavl= memavl;
-      if ((memavl= memavl / 4 * 3) < MIN_SORT_BUFFER &&
-          old_memavl > MIN_SORT_BUFFER)
-        memavl= MIN_SORT_BUFFER;
-    }
-    if (memavl < MIN_SORT_BUFFER)
-    {
-      /* purecov: begin inspected */
-      mi_check_print_error(sort_param->sort_info->param,
-                           "myisam_sort_buffer_size is too small. Current myisam_sort_buffer_size: %llu  rows: %llu  sort_length: %u",
-                           sortbuff_size, (ulonglong) idx, sort_length);
-      my_errno= ENOMEM;
-      goto err;
-    /* purecov: end inspected */
-    }
-
-    if (sort_param->sort_info->param->testflag & T_VERBOSE)
-      my_fprintf(stdout,
-                 "Key %d - Allocating buffer for %llu keys\n",
-                 sort_param->key + 1, (ulonglong) keys);
-    sort_param->sort_keys= sort_keys;
-
-    idx= error= 0;
-    sort_keys[0]= (uchar*) (sort_keys+keys);
-
-    DBUG_PRINT("info", ("reading keys"));
-    while (!(error= sort_param->sort_info->got_error) &&
-           !(error= (*sort_param->key_read)(sort_param, sort_keys[idx])))
-    {
-      if (sort_param->real_key_length > sort_param->key_length)
-      {
-        if (write_key(sort_param, sort_keys[idx],
-                      &sort_param->tempfile_for_exceptions))
-          goto err;
-        continue;
-      }
-
-      if (++idx == keys)
-      {
-        if (sort_param->write_keys(sort_param, sort_keys, idx - 1,
-                                   (BUFFPEK*) alloc_dynamic(&sort_param->buffpek),
-                                   &sort_param->tempfile))
-          goto err;
-        sort_keys[0]= (uchar*) (sort_keys+keys);
-        memcpy(sort_keys[0], sort_keys[idx - 1],
-               (size_t) sort_param->key_length);
-        idx= 1;
-      }
-      sort_keys[idx]= sort_keys[idx - 1] + sort_param->key_length;
-    }
-    if (error > 0)
-      goto err;
-    if (sort_param->buffpek.elements)
-    {
-      if (sort_param->write_keys(sort_param, sort_keys, idx,
-                                 (BUFFPEK*) alloc_dynamic(&sort_param->buffpek),
-                                 &sort_param->tempfile))
-        goto err;
-      sort_param->keys= (sort_param->buffpek.elements - 1) * (keys - 1) + idx;
-    }
-    else
-      sort_param->keys= idx;
-
-    goto ok;
-
-err:
-    DBUG_PRINT("error", ("got some error"));
-    sort_param->sort_info->got_error= 1; /* no need to protect with a mutex */
-    my_free(sort_keys);
-    sort_param->sort_keys= 0;
-    delete_dynamic(& sort_param->buffpek);
-    close_cached_file(&sort_param->tempfile);
-    close_cached_file(&sort_param->tempfile_for_exceptions);
-
-ok:
-    free_root(&sort_param->wordroot, MYF(0));
-    /*
-      Detach from the share if the writer is involved. Avoid others to
-      be blocked. This includes a flush of the write buffer. This will
-      also indicate EOF to the readers.
-      That means that a writer always gets here first and readers -
-      only when they see EOF. But if a reader finishes prematurely
-      because of an error it may reach this earlier - don't allow it
-      to detach the writer thread.
-    */
-    if (sort_param->master && sort_param->sort_info->info->rec_cache.share)
-      remove_io_thread(&sort_param->sort_info->info->rec_cache);
-
-    /* Readers detach from the share if any. Avoid others to be blocked. */
-    if (sort_param->read_cache.share)
-      remove_io_thread(&sort_param->read_cache);
-
-    mysql_mutex_lock(&sort_param->sort_info->mutex);
-    if (!--sort_param->sort_info->threads_running)
-      mysql_cond_signal(&sort_param->sort_info->cond);
-    mysql_mutex_unlock(&sort_param->sort_info->mutex);
-    DBUG_PRINT("exit", ("======== ending thread ========"));
-    DBUG_LEAVE;
-  }
+  if (!--sort_param->sort_info->threads_running)
+    mysql_cond_signal(&sort_param->sort_info->cond);
+  mysql_mutex_unlock(&sort_param->sort_info->mutex);
   my_thread_end();
   return NULL;
 }
@@ -603,7 +609,7 @@ int thr_write_keys(MI_SORT_PARAM *sort_param)
         length=param->sort_buffer_length;
         while (length >= MIN_SORT_BUFFER)
         {
-          if ((mergebuf= my_malloc(length, MYF(0))))
+          if ((mergebuf= my_malloc((size_t) length, MYF(0))))
               break;
           length=length*3/4;
         }
@@ -695,8 +701,8 @@ static int write_keys(MI_SORT_PARAM *info, register uchar **sort_keys,
   if (!buffpek)
     DBUG_RETURN(1);                             /* Out of memory */
 
-  my_qsort2((uchar*) sort_keys,count,sizeof(uchar*),(qsort2_cmp) info->key_cmp,
-            info);
+  my_qsort2((uchar*) sort_keys,(size_t) count, sizeof(uchar*),
+            (qsort2_cmp) info->key_cmp, info);
   if (!my_b_inited(tempfile) &&
       open_cached_file(tempfile, my_tmpdir(info->tmpdir), "ST",
                        DISK_BUFFER_SIZE, info->sort_info->param->myf_rw))
@@ -741,8 +747,8 @@ static int write_keys_varlen(MI_SORT_PARAM *info,
   if (!buffpek)
     DBUG_RETURN(1);                             /* Out of memory */
 
-  my_qsort2((uchar*) sort_keys,count,sizeof(uchar*),(qsort2_cmp) info->key_cmp,
-            info);
+  my_qsort2((uchar*) sort_keys, (size_t) count, sizeof(uchar*),
+            (qsort2_cmp) info->key_cmp, info);
   if (!my_b_inited(tempfile) &&
       open_cached_file(tempfile, my_tmpdir(info->tmpdir), "ST",
                        DISK_BUFFER_SIZE, info->sort_info->param->myf_rw))
@@ -863,12 +869,14 @@ static my_off_t read_to_buffer(IO_CACHE *fromfile, BUFFPEK *buffpek,
                                 uint sort_length)
 {
   register ha_keys count;
-  my_off_t length;
+  size_t length;
 
-  if ((count= (ha_keys) MY_MIN((ha_rows) buffpek->max_keys,buffpek->count)))
+  if ((count= (ha_keys) MY_MIN((ha_rows) buffpek->max_keys,
+                               (ha_rows) buffpek->count)))
   {
     if (my_b_pread(fromfile, (uchar*) buffpek->base,
-                   (length= sort_length * count), buffpek->file_pos))
+                   (length= (size_t) (sort_length * count)),
+                   buffpek->file_pos))
       return(HA_OFFSET_ERROR);
     buffpek->key=buffpek->base;
     buffpek->file_pos+= length;               /* New filepos */
@@ -933,7 +941,7 @@ static int write_merge_key(MI_SORT_PARAM *info __attribute__((unused)),
                            IO_CACHE *to_file, uchar *key,
                            uint sort_length, ha_keys count)
 {
-  return my_b_write(to_file, key, ((size_t) sort_length) * count);
+  return my_b_write(to_file, key, (size_t) (sort_length * count));
 }
 
 /*
@@ -993,7 +1001,7 @@ merge_buffers(MI_SORT_PARAM *info, ha_keys keys, IO_CACHE *from_file,
       if (to_file)
       {
         if (info->write_key(info,to_file,(uchar*) buffpek->key,
-                            (uint) sort_length,1))
+                            sort_length, 1))
         {
           error=1; goto err; /* purecov: inspected */
         }
@@ -1016,7 +1024,7 @@ merge_buffers(MI_SORT_PARAM *info, ha_keys keys, IO_CACHE *from_file,
         if (!(read_length= info->read_to_buffer(from_file,buffpek,sort_length)))
         {
           uchar *base= buffpek->base;
-          uint max_keys=buffpek->max_keys;
+          ha_keys max_keys=buffpek->max_keys;
 
           queue_remove_top(&queue);
 
