@@ -36,6 +36,7 @@ Copyright (c) 2006, 2015, Percona and/or its affiliates. All rights reserved.
 
 #ident "Copyright (c) 2006, 2015, Percona and/or its affiliates. All rights reserved."
 
+#include <my_global.h>
 #include "ft/serialize/block_table.h"
 #include "ft/ft.h"
 #include "ft/ft-cachetable-wrappers.h"
@@ -50,17 +51,16 @@ Copyright (c) 2006, 2015, Percona and/or its affiliates. All rights reserved.
 #include <toku_assert.h>
 #include <portability/toku_atomic.h>
 
-toku_instr_key *ft_ref_lock_mutex_key;
-
-void toku_reset_root_xid_that_created(FT ft, TXNID new_root_xid_that_created) {
-    // Reset the root_xid_that_created field to the given value.
+void 
+toku_reset_root_xid_that_created(FT ft, TXNID new_root_xid_that_created) {
+    // Reset the root_xid_that_created field to the given value.  
     // This redefines which xid created the dictionary.
 
     // hold lock around setting and clearing of dirty bit
     // (see cooperative use of dirty bit in ft_begin_checkpoint())
     toku_ft_lock(ft);
     ft->h->root_xid_that_created = new_root_xid_that_created;
-    ft->h->set_dirty();
+    ft->h->dirty = 1;
     toku_ft_unlock(ft);
 }
 
@@ -101,11 +101,15 @@ toku_ft_free (FT ft) {
     toku_free(ft);
 }
 
-void toku_ft_init_reflock(FT ft) {
-    toku_mutex_init(*ft_ref_lock_mutex_key, &ft->ft_ref_lock, nullptr);
+void
+toku_ft_init_reflock(FT ft) {
+    toku_mutex_init(&ft->ft_ref_lock, NULL);
 }
 
-void toku_ft_destroy_reflock(FT ft) { toku_mutex_destroy(&ft->ft_ref_lock); }
+void
+toku_ft_destroy_reflock(FT ft) {
+    toku_mutex_destroy(&ft->ft_ref_lock);
+}
 
 void
 toku_ft_grab_reflock(FT ft) {
@@ -146,7 +150,7 @@ static void ft_begin_checkpoint (LSN checkpoint_lsn, void *header_v) {
     assert(ft->h->type == FT_CURRENT);
     assert(ft->checkpoint_header == NULL);
     ft_copy_for_checkpoint_unlocked(ft, checkpoint_lsn);
-    ft->h->clear_dirty();             // this is only place this bit is cleared        (in currentheader)
+    ft->h->dirty = 0;             // this is only place this bit is cleared        (in currentheader)
     ft->blocktable.note_start_checkpoint_unlocked();
     toku_ft_unlock (ft);
 }
@@ -185,7 +189,7 @@ static void ft_checkpoint (CACHEFILE cf, int fd, void *header_v) {
     FT_HEADER ch = ft->checkpoint_header;
     assert(ch);
     assert(ch->type == FT_CHECKPOINT_INPROGRESS);
-    if (ch->dirty()) {            // this is only place this bit is tested (in checkpoint_header)
+    if (ch->dirty) {            // this is only place this bit is tested (in checkpoint_header)
         TOKULOGGER logger = toku_cachefile_logger(cf);
         if (logger) {
             toku_logger_fsync_if_lsn_not_fsynced(logger, ch->checkpoint_lsn);
@@ -200,7 +204,7 @@ static void ft_checkpoint (CACHEFILE cf, int fd, void *header_v) {
                                                              
         // write translation and header to disk (or at least to OS internal buffer)
         toku_serialize_ft_to(fd, ch, &ft->blocktable, ft->cf);
-        ch->clear_dirty();                      // this is only place this bit is cleared (in checkpoint_header)
+        ch->dirty = 0;                      // this is only place this bit is cleared (in checkpoint_header)
         
         // fsync the cachefile
         toku_cachefile_fsync(cf);
@@ -254,7 +258,7 @@ static void ft_close(CACHEFILE cachefile, int fd, void *header_v, bool oplsn_val
                 toku_log_fclose(
                     logger,
                     &lsn,
-                    ft->h->dirty(),
+                    ft->h->dirty,
                     bs,
                     toku_cachefile_filenum(cachefile));  // flush the log on
                                                          // close (if new header
@@ -265,7 +269,7 @@ static void ft_close(CACHEFILE cachefile, int fd, void *header_v, bool oplsn_val
             }
         }
     }
-    if (ft->h->dirty()) {               // this is the only place this bit is tested (in currentheader)
+    if (ft->h->dirty) {               // this is the only place this bit is tested (in currentheader)
         bool do_checkpoint = true;
         if (logger && logger->rollback_cachefile == cachefile) {
             do_checkpoint = false;
@@ -274,7 +278,7 @@ static void ft_close(CACHEFILE cachefile, int fd, void *header_v, bool oplsn_val
             ft_begin_checkpoint(lsn, header_v);
             ft_checkpoint(cachefile, fd, ft);
             ft_end_checkpoint(cachefile, fd, header_v);
-            assert(!ft->h->dirty()); // dirty bit should be cleared by begin_checkpoint and never set again (because we're closing the dictionary)
+            assert(!ft->h->dirty); // dirty bit should be cleared by begin_checkpoint and never set again (because we're closing the dictionary)
         }
     }
 }
@@ -370,7 +374,7 @@ ft_header_create(FT_OPTIONS options, BLOCKNUM root_blocknum, TXNID root_xid_that
     uint64_t now = (uint64_t) time(NULL);
     struct ft_header h = {
         .type = FT_CURRENT,
-        .dirty_ = 0,
+        .dirty = 0,
         .checkpoint_count = 0,
         .checkpoint_lsn = ZERO_LSN,
         .layout_version = FT_LAYOUT_VERSION,
@@ -432,8 +436,7 @@ int toku_read_ft_and_store_in_cachefile (FT_HANDLE ft_handle, CACHEFILE cf, LSN
     }
 
     int fd = toku_cachefile_get_fd(cf);
-    const char *fn = toku_cachefile_fname_in_env(cf);
-    int r = toku_deserialize_ft_from(fd, fn, max_acceptable_lsn, &ft);
+    int r = toku_deserialize_ft_from(fd, max_acceptable_lsn, &ft);
     if (r == TOKUDB_BAD_CHECKSUM) {
         fprintf(stderr, "Checksum failure while reading header in file %s.\n", toku_cachefile_fname_in_env(cf));
         assert(false);  // make absolutely sure we crash before doing anything else
@@ -521,7 +524,7 @@ toku_ft_note_hot_begin(FT_HANDLE ft_handle) {
     toku_ft_lock(ft);
     ft->h->time_of_last_optimize_begin = now;
     ft->h->count_of_optimize_in_progress++;
-    ft->h->set_dirty();
+    ft->h->dirty = 1;
     toku_ft_unlock(ft);
 }
 
@@ -545,7 +548,7 @@ toku_ft_note_hot_complete(FT_HANDLE ft_handle, bool success, MSN msn_at_start_of
         if (ft->h->count_of_optimize_in_progress == ft->h->count_of_optimize_in_progress_read_from_disk)
             ft->h->count_of_optimize_in_progress = 0;
     }
-    ft->h->set_dirty();
+    ft->h->dirty = 1;
     toku_ft_unlock(ft);
 }
 
@@ -958,7 +961,7 @@ void toku_ft_remove_reference(
 void toku_ft_set_nodesize(FT ft, unsigned int nodesize) {
     toku_ft_lock(ft);
     ft->h->nodesize = nodesize;
-    ft->h->set_dirty();
+    ft->h->dirty = 1;
     toku_ft_unlock(ft);
 }
 
@@ -971,7 +974,7 @@ void toku_ft_get_nodesize(FT ft, unsigned int *nodesize) {
 void toku_ft_set_basementnodesize(FT ft, unsigned int basementnodesize) {
     toku_ft_lock(ft);
     ft->h->basementnodesize = basementnodesize;
-    ft->h->set_dirty();
+    ft->h->dirty = 1;
     toku_ft_unlock(ft);
 }
 
@@ -984,7 +987,7 @@ void toku_ft_get_basementnodesize(FT ft, unsigned int *basementnodesize) {
 void toku_ft_set_compression_method(FT ft, enum toku_compression_method method) {
     toku_ft_lock(ft);
     ft->h->compression_method = method;
-    ft->h->set_dirty();
+    ft->h->dirty = 1;
     toku_ft_unlock(ft);
 }
 
@@ -997,7 +1000,7 @@ void toku_ft_get_compression_method(FT ft, enum toku_compression_method *methodp
 void toku_ft_set_fanout(FT ft, unsigned int fanout) {
     toku_ft_lock(ft);
     ft->h->fanout = fanout;
-    ft->h->set_dirty();
+    ft->h->dirty = 1;
     toku_ft_unlock(ft);
 }
 
