@@ -1,6 +1,6 @@
 /*****************************************************************************
 Copyright (C) 2013, 2015, Google Inc. All Rights Reserved.
-Copyright (c) 2014, 2020, MariaDB Corporation.
+Copyright (c) 2014, 2017, MariaDB Corporation. All Rights Reserved.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -12,7 +12,7 @@ FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
 this program; if not, write to the Free Software Foundation, Inc.,
-51 Franklin St, Fifth Floor, Boston, MA 02110-1335 USA
+51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 
 *****************************************************************************/
 /**************************************************//**
@@ -25,18 +25,13 @@ Modified           Jan Lindström jan.lindstrom@mariadb.com
 
 #include "fil0fil.h"
 #include "fil0crypt.h"
-#include "mach0data.h"
-#include "page0zip.h"
-#include "buf0buf.h"
-#include "buf0checksum.h"
-
-#ifndef UNIV_INNOCHECKSUM
-
 #include "srv0srv.h"
 #include "srv0start.h"
+#include "mach0data.h"
 #include "log0recv.h"
 #include "mtr0mtr.h"
 #include "mtr0log.h"
+#include "page0zip.h"
 #include "ut0ut.h"
 #include "btr0scrub.h"
 #include "fsp0fsp.h"
@@ -112,20 +107,13 @@ UNIV_INTERN mysql_pfs_key_t fil_crypt_data_mutex_key;
 extern my_bool srv_background_scrub_data_uncompressed;
 extern my_bool srv_background_scrub_data_compressed;
 
-/***********************************************************************
-Check if a key needs rotation given a key_state
-@param[in]	crypt_data		Encryption information
-@param[in]	key_version		Current key version
-@param[in]	latest_key_version	Latest key version
-@param[in]	rotate_key_age		when to rotate
-@return true if key needs rotation, false if not */
 static bool
 fil_crypt_needs_rotation(
-	const fil_space_crypt_t*	crypt_data,
-	uint				key_version,
-	uint				latest_key_version,
-	uint				rotate_key_age)
-	MY_ATTRIBUTE((warn_unused_result));
+	fil_encryption_t        encrypt_mode,           /*!< in: Encryption
+							mode */
+	uint			key_version,		/*!< in: Key version */
+	uint			latest_key_version,	/*!< in: Latest key version */
+	uint			rotate_key_age);	/*!< in: When to rotate */
 
 /*********************************************************************
 Init space crypt */
@@ -187,17 +175,11 @@ fil_crypt_get_latest_key_version(
 
 	if (crypt_data->is_key_found()) {
 
-		if (fil_crypt_needs_rotation(
-				crypt_data,
+		if (fil_crypt_needs_rotation(crypt_data->encryption,
 				crypt_data->min_key_version,
 				key_version,
 				srv_fil_crypt_rotate_key_age)) {
-			/* Below event seen as NULL-pointer at startup
-			when new database was created and we create a
-			checkpoint. Only seen when debugging. */
-			if (fil_crypt_threads_inited) {
-				os_event_set(fil_crypt_threads_event);
-			}
+			os_event_set(fil_crypt_threads_event);
 		}
 	}
 
@@ -497,15 +479,14 @@ Parse a MLOG_FILE_WRITE_CRYPT_DATA log entry
 @param[in]	block		buffer block
 @return position on log buffer */
 UNIV_INTERN
-byte*
+const byte*
 fil_parse_write_crypt_data(
-	byte*			ptr,
+	const byte*		ptr,
 	const byte*		end_ptr,
-	const buf_block_t*	block,
-	dberr_t*		err)
+	const buf_block_t*	block)
 {
 	/* check that redo log entry is complete */
-	uint entry_size =
+        size_t entry_size =
 		4 + // size of space_id
 		2 + // size of offset
 		1 + // size of type
@@ -514,8 +495,6 @@ fil_parse_write_crypt_data(
 		4 +  // size of key_id
 		1; // fil_encryption_t
 
-	*err = DB_SUCCESS;
-
 	if (ptr + entry_size > end_ptr) {
 		return NULL;
 	}
@@ -526,15 +505,13 @@ fil_parse_write_crypt_data(
 	ptr += 2;
 	uint type = mach_read_from_1(ptr);
 	ptr += 1;
-	uint len = mach_read_from_1(ptr);
+	size_t len = mach_read_from_1(ptr);
 	ptr += 1;
 
-	if ((type != CRYPT_SCHEME_1 && type != CRYPT_SCHEME_UNENCRYPTED)
-	    || len != CRYPT_SCHEME_1_IV_LEN) {
-		*err = DB_CORRUPTION;
-		return NULL;
-	}
+	ut_a(type == CRYPT_SCHEME_UNENCRYPTED ||
+	     type == CRYPT_SCHEME_1); // only supported
 
+	ut_a(len == CRYPT_SCHEME_1_IV_LEN); // only supported
 	uint min_key_version = mach_read_from_4(ptr);
 	ptr += 4;
 
@@ -553,21 +530,15 @@ fil_parse_write_crypt_data(
 	crypt_data->page0_offset = offset;
 	crypt_data->min_key_version = min_key_version;
 	crypt_data->encryption = encryption;
-	crypt_data->type = type;
 	memcpy(crypt_data->iv, ptr, len);
 	ptr += len;
 
 	/* update fil_space memory cache with crypt_data */
-	if (fil_space_t* space = fil_space_acquire_silent(space_id)) {
+	fil_space_t* space = fil_space_acquire_silent(space_id);
+
+	if (space) {
 		crypt_data = fil_space_set_crypt_data(space, crypt_data);
 		fil_space_release(space);
-		/* Check is used key found from encryption plugin */
-		if (crypt_data->should_encrypt()
-		    && !crypt_data->is_key_found()) {
-			*err = DB_DECRYPTION_FAILED;
-		}
-	} else {
-		fil_space_destroy_crypt_data(&crypt_data);
 	}
 
 	return ptr;
@@ -665,7 +636,7 @@ fil_encrypt_buf(
 	// store the post-encryption checksum after the key-version
 	mach_write_to_4(dst_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION + 4, checksum);
 
-	ut_ad(fil_space_verify_crypt_checksum(dst_frame, zip_size));
+	ut_ad(fil_space_verify_crypt_checksum(dst_frame, zip_size, NULL, offset));
 
 	srv_stats.pages_encrypted.inc();
 
@@ -704,10 +675,65 @@ fil_space_encrypt(
 	}
 
 	fil_space_crypt_t* crypt_data = space->crypt_data;
-	ut_ad(space->n_pending_ios > 0);
+	ut_ad(space->n_pending_ops);
 	ulint zip_size = fsp_flags_get_zip_size(space->flags);
-	return fil_encrypt_buf(crypt_data, space->id, offset, lsn,
-			       src_frame, zip_size, dst_frame);
+	byte* tmp = fil_encrypt_buf(crypt_data, space->id, offset, lsn, src_frame, zip_size, dst_frame);
+
+#ifdef UNIV_DEBUG
+	if (tmp) {
+		/* Verify that encrypted buffer is not corrupted */
+		byte* tmp_mem = (byte *)malloc(UNIV_PAGE_SIZE);
+		dberr_t err = DB_SUCCESS;
+		byte* src = src_frame;
+		bool page_compressed_encrypted = (mach_read_from_2(tmp+FIL_PAGE_TYPE) == FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED);
+		byte* comp_mem = NULL;
+		byte* uncomp_mem = NULL;
+		ulint size = (zip_size) ? zip_size : UNIV_PAGE_SIZE;
+
+		if (page_compressed_encrypted) {
+			comp_mem = (byte *)malloc(UNIV_PAGE_SIZE);
+			uncomp_mem = (byte *)malloc(UNIV_PAGE_SIZE);
+			memcpy(comp_mem, src_frame, UNIV_PAGE_SIZE);
+			fil_decompress_page(uncomp_mem, comp_mem, UNIV_PAGE_SIZE, NULL);
+			src = uncomp_mem;
+		}
+
+		bool corrupted1 = buf_page_is_corrupted(true, src, zip_size, space);
+		bool ok = fil_space_decrypt(crypt_data, tmp_mem, size, tmp, &err);
+
+		/* Need to decompress the page if it was also compressed */
+		if (page_compressed_encrypted) {
+			memcpy(comp_mem, tmp_mem, UNIV_PAGE_SIZE);
+			fil_decompress_page(tmp_mem, comp_mem, UNIV_PAGE_SIZE, NULL);
+		}
+
+		bool corrupted = buf_page_is_corrupted(true, tmp_mem, zip_size, space);
+		bool different = memcmp(src, tmp_mem, size);
+
+		if (!ok || corrupted || corrupted1 || err != DB_SUCCESS || different) {
+			fprintf(stderr, "ok %d corrupted %d corrupted1 %d err %d different %d\n", ok , corrupted, corrupted1, err, different);
+			fprintf(stderr, "src_frame\n");
+			buf_page_print(src_frame, zip_size, BUF_PAGE_PRINT_NO_CRASH);
+			fprintf(stderr, "encrypted_frame\n");
+			buf_page_print(tmp, zip_size, BUF_PAGE_PRINT_NO_CRASH);
+			fprintf(stderr, "decrypted_frame\n");
+			buf_page_print(tmp_mem, zip_size, 0);
+		}
+
+		free(tmp_mem);
+
+		if (comp_mem) {
+			free(comp_mem);
+		}
+
+		if (uncomp_mem) {
+			free(uncomp_mem);
+		}
+	}
+
+#endif /* UNIV_DEBUG */
+
+	return tmp;
 }
 
 /******************************************************************
@@ -740,6 +766,25 @@ fil_space_decrypt(
 		return false;
 	}
 
+	if (crypt_data == NULL) {
+		if (!(space == 0 && offset == 0) && key_version != 0) {
+			/* FIL_PAGE_FILE_FLUSH_LSN field i.e.
+			FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION
+			should be only defined for the
+			first page in a system tablespace
+			data file (ibdata*, not *.ibd), if not
+			clear it. */
+
+			DBUG_PRINT("ib_crypt",
+				("Page on space %lu offset %lu has key_version %u"
+				" when it shoud be undefined.",
+				space, offset, key_version));
+
+			mach_write_to_4(src_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, 0);
+		}
+		return false;
+	}
+
 	ut_a(crypt_data != NULL && crypt_data->is_encrypted());
 
 	/* read space & lsn */
@@ -791,6 +836,9 @@ fil_space_decrypt(
 		memcpy(tmp_frame + page_size - FIL_PAGE_DATA_END,
 		       src_frame + page_size - FIL_PAGE_DATA_END,
 		       FIL_PAGE_DATA_END);
+
+		// clear key-version & crypt-checksum from dst
+		memset(tmp_frame + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION, 0, 8);
 	}
 
 	srv_stats.pages_decrypted.inc();
@@ -822,7 +870,7 @@ fil_space_decrypt(
 	*decrypted = false;
 
 	ut_ad(space->crypt_data != NULL && space->crypt_data->is_encrypted());
-	ut_ad(space->n_pending_ios > 0);
+	ut_ad(space->n_pending_ops > 0);
 
 	bool encrypted = fil_space_decrypt(
 				space->crypt_data,
@@ -849,7 +897,7 @@ fil_space_decrypt(
 Calculate post encryption checksum
 @param[in]	zip_size	zip_size or 0
 @param[in]	dst_frame	Block where checksum is calculated
-@return page checksum
+@return page checksum or BUF_NO_CHECKSUM_MAGIC
 not needed. */
 UNIV_INTERN
 ulint
@@ -858,18 +906,165 @@ fil_crypt_calculate_checksum(
 	const byte*	dst_frame)
 {
 	ib_uint32_t checksum = 0;
+	srv_checksum_algorithm_t algorithm =
+			static_cast<srv_checksum_algorithm_t>(srv_checksum_algorithm);
 
-	/* For encrypted tables we use only crc32 and strict_crc32 */
 	if (zip_size == 0) {
-		checksum = buf_calc_page_crc32(dst_frame);
+		switch (algorithm) {
+		case SRV_CHECKSUM_ALGORITHM_CRC32:
+		case SRV_CHECKSUM_ALGORITHM_STRICT_CRC32:
+			checksum = buf_calc_page_crc32(dst_frame);
+			break;
+		case SRV_CHECKSUM_ALGORITHM_INNODB:
+		case SRV_CHECKSUM_ALGORITHM_STRICT_INNODB:
+			checksum = (ib_uint32_t) buf_calc_page_new_checksum(
+				dst_frame);
+			break;
+		case SRV_CHECKSUM_ALGORITHM_NONE:
+		case SRV_CHECKSUM_ALGORITHM_STRICT_NONE:
+			checksum = BUF_NO_CHECKSUM_MAGIC;
+			break;
+			/* no default so the compiler will emit a warning
+			* if new enum is added and not handled here */
+		}
 	} else {
 		checksum = page_zip_calc_checksum(dst_frame, zip_size,
-				                  SRV_CHECKSUM_ALGORITHM_CRC32);
+				                          algorithm);
 	}
 
 	return checksum;
 }
 
+/*********************************************************************
+Verify that post encryption checksum match calculated checksum.
+This function should be called only if tablespace contains crypt_data
+metadata (this is strong indication that tablespace is encrypted).
+Function also verifies that traditional checksum does not match
+calculated checksum as if it does page could be valid unencrypted,
+encrypted, or corrupted.
+
+@param[in]	page		Page to verify
+@param[in]	zip_size	zip size
+@param[in]	space		Tablespace
+@param[in]	pageno		Page no
+@return true if page is encrypted AND OK, false otherwise */
+UNIV_INTERN
+bool
+fil_space_verify_crypt_checksum(
+	byte* 			page,
+	ulint			zip_size,
+	const fil_space_t*	space,
+	ulint			pageno)
+{
+	uint key_version = mach_read_from_4(page+ FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
+
+	/* If page is not encrypted, return false */
+	if (key_version == 0) {
+		return false;
+	}
+
+	srv_checksum_algorithm_t algorithm =
+			static_cast<srv_checksum_algorithm_t>(srv_checksum_algorithm);
+	/* If no checksum is used, can't continue checking. */
+	if (algorithm == SRV_CHECKSUM_ALGORITHM_NONE) {
+		return(true);
+	}
+
+	/* Read stored post encryption checksum. */
+	ib_uint32_t checksum = mach_read_from_4(
+		page + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION + 4);
+
+	/* Declare empty pages non-corrupted */
+	if (checksum == 0
+	    && *reinterpret_cast<const ib_uint64_t*>(page + FIL_PAGE_LSN) == 0
+	    && buf_page_is_zeroes(page, zip_size)) {
+		return(true);
+	}
+
+	/* Compressed and encrypted pages do not have checksum. Assume not
+	corrupted. Page verification happens after decompression in
+	buf_page_io_complete() using buf_page_is_corrupted(). */
+	if (mach_read_from_2(page+FIL_PAGE_TYPE) == FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED) {
+		return (true);
+	}
+
+	/* Compressed pages use different checksum method. We first store
+	the post encryption checksum on checksum location and after function
+	restore the original. */
+	if (zip_size) {
+		ib_uint32_t old = static_cast<ib_uint32_t>(mach_read_from_4(
+				page + FIL_PAGE_SPACE_OR_CHKSUM));
+
+		mach_write_to_4(page + FIL_PAGE_SPACE_OR_CHKSUM, checksum);
+
+		bool valid = page_zip_verify_checksum(page, zip_size);
+
+		mach_write_to_4(page + FIL_PAGE_SPACE_OR_CHKSUM, old);
+
+		return (valid);
+	}
+
+	/* If stored checksum matches one of the calculated checksums
+	page is not corrupted. */
+
+	ib_uint32_t cchecksum1 = buf_calc_page_crc32(page);
+	ib_uint32_t cchecksum2 = (ib_uint32_t) buf_calc_page_new_checksum(
+				page);
+	bool encrypted = (checksum == cchecksum1 || checksum == cchecksum2
+		|| checksum == BUF_NO_CHECKSUM_MAGIC);
+
+	/* MySQL 5.6 and MariaDB 10.0 and 10.1 will write an LSN to the
+	first page of each system tablespace file at
+	FIL_PAGE_FILE_FLUSH_LSN offset. On other pages and in other files,
+	the field might have been uninitialized until MySQL 5.5. In MySQL 5.7
+	(and MariaDB Server 10.2.2) WL#7990 stopped writing the field for other
+	than page 0 of the system tablespace.
+
+	Starting from MariaDB 10.1 the field has been repurposed for
+	encryption key_version.
+
+	Starting with MySQL 5.7 (and MariaDB Server 10.2), the
+	field has been repurposed for SPATIAL INDEX pages for
+	FIL_RTREE_SPLIT_SEQ_NUM.
+
+	Note that FIL_PAGE_FILE_FLUSH_LSN is not included in the InnoDB page
+	checksum.
+
+	Thus, FIL_PAGE_FILE_FLUSH_LSN could contain any value. While the
+	field would usually be 0 for pages that are not encrypted, we cannot
+	assume that a nonzero value means that the page is encrypted.
+	Therefore we must validate the page both as encrypted and unencrypted
+	when FIL_PAGE_FILE_FLUSH_LSN does not contain 0.
+	*/
+
+	ulint checksum1 = mach_read_from_4(
+		page + FIL_PAGE_SPACE_OR_CHKSUM);
+
+	ulint checksum2 = mach_read_from_4(
+		page + UNIV_PAGE_SIZE - FIL_PAGE_END_LSN_OLD_CHKSUM);
+
+
+	bool valid = (buf_page_is_checksum_valid_crc32(page,checksum1,checksum2)
+		|| buf_page_is_checksum_valid_none(page,checksum1,checksum2)
+		|| buf_page_is_checksum_valid_innodb(page,checksum1, checksum2));
+
+	if (encrypted && valid) {
+		/* If page is encrypted and traditional checksums match,
+		page could be still encrypted, or not encrypted and valid or
+		corrupted. */
+		ib_logf(IB_LOG_LEVEL_ERROR,
+			" Page %lu in space %s (%lu) maybe corrupted."
+			" Post encryption checksum %u stored [%lu:%lu] key_version %u",
+			pageno,
+			space ? space->name : "N/A",
+			mach_read_from_4(page + FIL_PAGE_ARCH_LOG_NO_OR_SPACE_ID),
+			checksum, checksum1, checksum2, key_version);
+		encrypted = false;
+	}
+
+	return(encrypted);
+}
+
 /***********************************************************************/
 
 /** A copy of global key state */
@@ -907,17 +1102,17 @@ fil_crypt_get_key_state(
 
 /***********************************************************************
 Check if a key needs rotation given a key_state
-@param[in]	crypt_data		Encryption information
+@param[in]	encrypt_mode		Encryption mode
 @param[in]	key_version		Current key version
 @param[in]	latest_key_version	Latest key version
 @param[in]	rotate_key_age		when to rotate
 @return true if key needs rotation, false if not */
 static bool
 fil_crypt_needs_rotation(
-	const fil_space_crypt_t*	crypt_data,
-	uint				key_version,
-	uint				latest_key_version,
-	uint				rotate_key_age)
+	fil_encryption_t	encrypt_mode,
+	uint			key_version,
+	uint			latest_key_version,
+	uint			rotate_key_age)
 {
 	if (key_version == ENCRYPTION_KEY_VERSION_INVALID) {
 		return false;
@@ -930,20 +1125,13 @@ fil_crypt_needs_rotation(
 	}
 
 	if (latest_key_version == 0 && key_version != 0) {
-		if (crypt_data->encryption == FIL_ENCRYPTION_DEFAULT) {
+		if (encrypt_mode == FIL_ENCRYPTION_DEFAULT) {
 			/* this is rotation encrypted => unencrypted */
 			return true;
 		}
 		return false;
 	}
 
-	if (crypt_data->encryption == FIL_ENCRYPTION_DEFAULT
-	    && crypt_data->type == CRYPT_SCHEME_1
-	    && srv_encrypt_tables == 0 ) {
-		/* This is rotation encrypted => unencrypted */
-		return true;
-	}
-
 	/* this is rotation encrypted => encrypted,
 	* only reencrypt if key is sufficiently old */
 	if (key_version + rotate_key_age < latest_key_version) {
@@ -953,43 +1141,6 @@ fil_crypt_needs_rotation(
 	return false;
 }
 
-/** Read page 0 and possible crypt data from there.
-@param[in,out]	space		Tablespace */
-static inline
-void
-fil_crypt_read_crypt_data(fil_space_t* space)
-{
-	if (space->crypt_data || space->size
-	    || !fil_space_get_size(space->id)) {
-		/* The encryption metadata has already been read, or
-		the tablespace is not encrypted and the file has been
-		opened already, or the file cannot be accessed,
-		likely due to a concurrent TRUNCATE or
-		RENAME or DROP (possibly as part of ALTER TABLE).
-		FIXME: The file can become unaccessible any time
-		after this check! We should really remove this
-		function and instead make crypt_data an integral
-		part of fil_space_t. */
-		return;
-	}
-
-	mtr_t	mtr;
-	mtr_start(&mtr);
-	ulint zip_size = fsp_flags_get_zip_size(space->flags);
-	ulint offset = fsp_header_get_crypt_offset(zip_size);
-	if (buf_block_t* block = buf_page_get(space->id, zip_size, 0,
-					      RW_S_LATCH, &mtr)) {
-		mutex_enter(&fil_system->mutex);
-		if (!space->crypt_data) {
-			space->crypt_data = fil_space_read_crypt_data(
-				space->id, block->frame, offset);
-		}
-		mutex_exit(&fil_system->mutex);
-	}
-
-	mtr_commit(&mtr);
-}
-
 /***********************************************************************
 Start encrypting a space
 @param[in,out]		space		Tablespace
@@ -1000,7 +1151,6 @@ fil_crypt_start_encrypting_space(
 	fil_space_t*	space)
 {
 	bool recheck = false;
-
 	mutex_enter(&fil_crypt_threads_mutex);
 
 	fil_space_crypt_t *crypt_data = space->crypt_data;
@@ -1067,6 +1217,8 @@ fil_crypt_start_encrypting_space(
 		byte* frame = buf_block_get_frame(block);
 		crypt_data->type = CRYPT_SCHEME_1;
 		crypt_data->write_page0(frame, &mtr);
+
+
 		mtr_commit(&mtr);
 
 		/* record lsn of update */
@@ -1142,10 +1294,10 @@ struct rotate_thread_t {
 	bool should_shutdown() const {
 		switch (srv_shutdown_state) {
 		case SRV_SHUTDOWN_NONE:
-			return thread_no >= srv_n_fil_crypt_threads;
 		case SRV_SHUTDOWN_CLEANUP:
-			return true;
+			return thread_no >= srv_n_fil_crypt_threads;
 		case SRV_SHUTDOWN_FLUSH_PHASE:
+			return true;
 		case SRV_SHUTDOWN_LAST_PHASE:
 		case SRV_SHUTDOWN_EXIT_THREADS:
 			break;
@@ -1230,10 +1382,9 @@ fil_crypt_space_needs_rotation(
 		}
 
 		bool need_key_rotation = fil_crypt_needs_rotation(
-			crypt_data,
+			crypt_data->encryption,
 			crypt_data->min_key_version,
-			key_state->key_version,
-			key_state->rotate_key_age);
+			key_state->key_version, key_state->rotate_key_age);
 
 		crypt_data->rotate_state.scrubbing.is_active =
 			btr_scrub_start_space(space->id, &state->scrub_data);
@@ -1341,21 +1492,20 @@ fil_crypt_realloc_iops(
 
 	if (10 * state->cnt_waited > state->batch) {
 		/* if we waited more than 10% re-estimate max_iops */
-		ulint avg_wait_time_us =
+		uint avg_wait_time_us =
 			state->sum_waited_us / state->cnt_waited;
 
+		DBUG_PRINT("ib_crypt",
+			("thr_no: %u - update estimated_max_iops from %u to %u.",
+			state->thread_no,
+			state->estimated_max_iops,
+			1000000 / avg_wait_time_us));
+
 		if (avg_wait_time_us == 0) {
 			avg_wait_time_us = 1; // prevent division by zero
 		}
 
-		DBUG_PRINT("ib_crypt",
-			("thr_no: %u - update estimated_max_iops from %u to "
-			 ULINTPF ".",
-			state->thread_no,
-			state->estimated_max_iops,
-			1000000 / avg_wait_time_us));
-
-		state->estimated_max_iops = uint(1000000 / avg_wait_time_us);
+		state->estimated_max_iops = 1000000 / avg_wait_time_us;
 		state->cnt_waited = 0;
 		state->sum_waited_us = 0;
 	} else {
@@ -1363,7 +1513,7 @@ fil_crypt_realloc_iops(
 		DBUG_PRINT("ib_crypt",
 			("thr_no: %u only waited %lu%% skip re-estimate.",
 			state->thread_no,
-			(100 * state->cnt_waited) / (state->batch ? state->batch : 1)));
+			(100 * state->cnt_waited) / state->batch));
 	}
 
 	if (state->estimated_max_iops <= state->allocated_iops) {
@@ -1466,7 +1616,7 @@ fil_crypt_find_space_to_rotate(
 	/* we need iops to start rotating */
 	while (!state->should_shutdown() && !fil_crypt_alloc_iops(state)) {
 		os_event_reset(fil_crypt_threads_event);
-		os_event_wait_time(fil_crypt_threads_event, 100000);
+		os_event_wait_time(fil_crypt_threads_event, 1000000);
 	}
 
 	if (state->should_shutdown()) {
@@ -1495,8 +1645,6 @@ fil_crypt_find_space_to_rotate(
 	}
 
 	while (!state->should_shutdown() && state->space) {
-		fil_crypt_read_crypt_data(state->space);
-
 		if (fil_crypt_space_needs_rotation(state, key_state, recheck)) {
 			ut_ad(key_state->key_id);
 			/* init state->min_key_version_found before
@@ -1591,27 +1739,47 @@ fil_crypt_find_page_to_rotate(
 
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
-	mutex_enter(&crypt_data->mutex);
-	ut_ad(key_state->key_id == crypt_data->key_id);
+	/* Space might already be dropped */
+	if (crypt_data) {
+		mutex_enter(&crypt_data->mutex);
+		ut_ad(key_state->key_id == crypt_data->key_id);
 
-	bool found = crypt_data->rotate_state.max_offset >=
-		crypt_data->rotate_state.next_offset;
+		if (crypt_data->rotate_state.next_offset <
+		    crypt_data->rotate_state.max_offset) {
 
-	if (found) {
-		state->offset = crypt_data->rotate_state.next_offset;
-		ulint remaining = crypt_data->rotate_state.max_offset -
-			crypt_data->rotate_state.next_offset;
+			state->offset = crypt_data->rotate_state.next_offset;
+			ulint remaining = crypt_data->rotate_state.max_offset -
+				crypt_data->rotate_state.next_offset;
 
-		if (batch <= remaining) {
-			state->batch = batch;
-		} else {
-			state->batch = remaining;
+			if (batch <= remaining) {
+				state->batch = batch;
+			} else {
+				state->batch = remaining;
+			}
+
+			crypt_data->rotate_state.next_offset += batch;
+			mutex_exit(&crypt_data->mutex);
+			return true;
 		}
+
+		mutex_exit(&crypt_data->mutex);
 	}
 
-	crypt_data->rotate_state.next_offset += batch;
-	mutex_exit(&crypt_data->mutex);
-	return found;
+	return false;
+}
+
+/***********************************************************************
+Check if a page is uninitialized (doesn't need to be rotated)
+@param[in]	frame		Page to check
+@param[in]	zip_size	zip_size or 0
+@return true if page is uninitialized, false if not. */
+static inline
+bool
+fil_crypt_is_page_uninitialized(
+	const byte	*frame,
+	uint		zip_size)
+{
+	return (buf_page_is_zeroes(frame, zip_size));
 }
 
 #define fil_crypt_get_page_throttle(state,offset,mtr,sleeptime_ms) \
@@ -1659,18 +1827,19 @@ fil_crypt_get_page_throttle_func(
 
 	state->crypt_stat.pages_read_from_disk++;
 
-	const ulonglong start = my_interval_timer();
+	ullint start = ut_time_us(NULL);
 	block = buf_page_get_gen(space->id, zip_size, offset,
 				 RW_X_LATCH,
 				 NULL, BUF_GET_POSSIBLY_FREED,
 				 file, line, mtr);
-	const ulonglong end = my_interval_timer();
+	ullint end = ut_time_us(NULL);
+
+	if (end < start) {
+		end = start; // safety...
+	}
 
 	state->cnt_waited++;
-
-	if (end > start) {
-		state->sum_waited_us += (end - start) / 1000;
-	}
+	state->sum_waited_us += (end - start);
 
 	/* average page load */
 	ulint add_sleeptime_ms = 0;
@@ -1773,7 +1942,6 @@ fil_crypt_rotate_page(
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
 	ut_ad(space->n_pending_ops > 0);
-	ut_ad(offset > 0);
 
 	/* In fil_crypt_thread where key rotation is done we have
 	acquired space and checked that this space is not yet
@@ -1783,7 +1951,7 @@ fil_crypt_rotate_page(
 		return;
 	}
 
-	if (space_id == TRX_SYS_SPACE && offset == TRX_SYS_PAGE_NO) {
+	if (space == TRX_SYS_SPACE && offset == TRX_SYS_PAGE_NO) {
 		/* don't encrypt this as it contains address to dblwr buffer */
 		return;
 	}
@@ -1799,44 +1967,45 @@ fil_crypt_rotate_page(
 		bool modified = false;
 		int needs_scrubbing = BTR_SCRUB_SKIP_PAGE;
 		lsn_t block_lsn = block->page.newest_modification;
-		byte* frame = buf_block_get_frame(block);
-		uint kv =  mach_read_from_4(frame+FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION);
+		uint kv =  block->page.key_version;
 
+		/* check if tablespace is closing after reading page */
 		if (space->is_stopping()) {
-			/* The tablespace is closing (in DROP TABLE or
-			TRUNCATE TABLE or similar): avoid further access */
-		} else if (!*reinterpret_cast<uint32_t*>(FIL_PAGE_OFFSET
-							 + frame)) {
-			/* It looks like this page was never
-			allocated. Because key rotation is accessing
-			pages in a pattern that is unlike the normal
-			B-tree and undo log access pattern, we cannot
-			invoke fseg_page_is_free() here, because that
-			could result in a deadlock. If we invoked
-			fseg_page_is_free() and released the
-			tablespace latch before acquiring block->lock,
-			then the fseg_page_is_free() information
-			could be stale already. */
-			ut_ad(kv == 0);
-			ut_ad(page_get_space_id(frame) == 0);
-		} else if (fil_crypt_needs_rotation(
-				crypt_data,
-				kv,
-				key_state->key_version,
-				key_state->rotate_key_age)) {
+			byte* frame = buf_block_get_frame(block);
 
-			modified = true;
+			if (kv == 0 &&
+				fil_crypt_is_page_uninitialized(frame, zip_size)) {
+				;
+			} else if (fil_crypt_needs_rotation(
+					crypt_data->encryption,
+					kv, key_state->key_version,
+					key_state->rotate_key_age)) {
 
-			/* force rotation by dummy updating page */
-			mlog_write_ulint(frame + FIL_PAGE_SPACE_ID,
-					 space_id, MLOG_4BYTES, &mtr);
+				/* page can be "fresh" i.e never written in case
+				* kv == 0 or it should have a key version at least
+				* as big as the space minimum key version*/
+				ut_a(kv == 0 || kv >= crypt_data->min_key_version);
 
-			/* statistics */
-			state->crypt_stat.pages_modified++;
-		} else {
-			if (crypt_data->is_encrypted()) {
-				if (kv < state->min_key_version_found) {
-					state->min_key_version_found = kv;
+				modified = true;
+
+				/* force rotation by dummy updating page */
+				mlog_write_ulint(frame +
+					FIL_PAGE_ARCH_LOG_NO_OR_SPACE_ID,
+					space_id, MLOG_4BYTES, &mtr);
+
+				/* update block */
+				block->page.key_version = key_state->key_version;
+
+				/* statistics */
+				state->crypt_stat.pages_modified++;
+			} else {
+				if (crypt_data->is_encrypted()) {
+					ut_a(kv >= crypt_data->min_key_version ||
+						(kv == 0 && key_state->key_version == 0));
+
+					if (kv < state->min_key_version_found) {
+						state->min_key_version_found = kv;
+					}
 				}
 			}
 
@@ -1917,11 +2086,6 @@ fil_crypt_rotate_page(
 				state->end_lsn = block_lsn;
 			}
 		}
-	} else {
-		/* If block read failed mtr memo and log should be empty. */
-		ut_ad(dyn_array_get_data_size(&mtr.memo) == 0);
-		ut_ad(dyn_array_get_data_size(&mtr.log) == 0);
-		mtr_commit(&mtr);
 	}
 
 	if (sleeptime_ms) {
@@ -1961,12 +2125,6 @@ fil_crypt_rotate_pages(
 			continue;
 		}
 
-		/* If space is marked as stopping, stop rotating
-		pages. */
-		if (state->space->is_stopping()) {
-			break;
-		}
-
 		fil_crypt_rotate_page(key_state, state);
 	}
 }
@@ -1992,7 +2150,7 @@ fil_crypt_flush_space(
 		bool success = false;
 		ulint n_pages = 0;
 		ulint sum_pages = 0;
-		const ulonglong start = my_interval_timer();
+		ullint start = ut_time_us(NULL);
 
 		do {
 			success = buf_flush_list(ULINT_MAX, end_lsn, &n_pages);
@@ -2000,11 +2158,11 @@ fil_crypt_flush_space(
 			sum_pages += n_pages;
 		} while (!success && !space->is_stopping());
 
-		const ulonglong end = my_interval_timer();
+		ullint end = ut_time_us(NULL);
 
 		if (sum_pages && end > start) {
 			state->cnt_waited += sum_pages;
-			state->sum_waited_us += (end - start) / 1000;
+			state->sum_waited_us += (end - start);
 
 			/* statistics */
 			state->crypt_stat.pages_flushed += sum_pages;
@@ -2015,22 +2173,20 @@ fil_crypt_flush_space(
 		crypt_data->type = CRYPT_SCHEME_UNENCRYPTED;
 	}
 
-	if (!space->is_stopping()) {
-		/* update page 0 */
-		mtr_t mtr;
-		mtr_start(&mtr);
+	/* update page 0 */
+	mtr_t mtr;
+	mtr_start(&mtr);
 
-		const uint zip_size = fsp_flags_get_zip_size(state->space->flags);
+	const uint zip_size = fsp_flags_get_zip_size(state->space->flags);
 
-		buf_block_t* block = buf_page_get_gen(space->id, zip_size, 0,
+	buf_block_t* block = buf_page_get_gen(space->id, zip_size, 0,
 				RW_X_LATCH, NULL, BUF_GET,
 				__FILE__, __LINE__, &mtr);
-		byte* frame = buf_block_get_frame(block);
+	byte* frame = buf_block_get_frame(block);
 
-		crypt_data->write_page0(frame, &mtr);
+	crypt_data->write_page0(frame, &mtr);
 
-		mtr_commit(&mtr);
-	}
+	mtr_commit(&mtr);
 }
 
 /***********************************************************************
@@ -2095,8 +2251,7 @@ fil_crypt_complete_rotate_space(
 		mutex_exit(&crypt_data->mutex);
 
 		/* all threads must call btr_scrub_complete_space wo/ mutex held */
-		if (state->scrub_data.scrubbing) {
-			btr_scrub_complete_space(&state->scrub_data);
+		if (btr_scrub_complete_space(&state->scrub_data) == true) {
 			if (should_flush) {
 				/* only last thread updates last_scrub_completed */
 				ut_ad(crypt_data);
@@ -2114,11 +2269,6 @@ fil_crypt_complete_rotate_space(
 			crypt_data->rotate_state.flushing = false;
 			mutex_exit(&crypt_data->mutex);
 		}
-	} else {
-		mutex_enter(&crypt_data->mutex);
-		ut_a(crypt_data->rotate_state.active_threads > 0);
-		crypt_data->rotate_state.active_threads--;
-		mutex_exit(&crypt_data->mutex);
 	}
 }
 
@@ -2193,19 +2343,15 @@ DECLARE_THREAD(fil_crypt_thread)(
 			fil_crypt_start_rotate_space(&new_state, &thr);
 
 			/* iterate all pages (cooperativly with other threads) */
-			while (!thr.should_shutdown() &&
+			while (!thr.should_shutdown() && thr.space &&
 			       fil_crypt_find_page_to_rotate(&new_state, &thr)) {
 
-				if (!thr.space->is_stopping()) {
-					/* rotate a (set) of pages */
-					fil_crypt_rotate_pages(&new_state, &thr);
-				}
+				/* rotate a (set) of pages */
+				fil_crypt_rotate_pages(&new_state, &thr);
 
 				/* If space is marked as stopping, release
 				space and stop rotation. */
 				if (thr.space->is_stopping()) {
-					fil_crypt_complete_rotate_space(
-						&new_state, &thr);
 					fil_space_release(thr.space);
 					thr.space = NULL;
 					break;
@@ -2272,7 +2418,7 @@ fil_crypt_set_thread_cnt(
 			os_thread_create(fil_crypt_thread, NULL, &rotation_thread_id);
 
 			ib_logf(IB_LOG_LEVEL_INFO,
-				"Creating #%d encryption thread id %lu total threads %u.",
+				"Creating #%d thread id %lu total threads %u.",
 				i+1, os_thread_pf(rotation_thread_id), new_cnt);
 		}
 	} else if (new_cnt < srv_n_fil_crypt_threads) {
@@ -2284,13 +2430,7 @@ fil_crypt_set_thread_cnt(
 
 	while(srv_n_fil_crypt_threads_started != srv_n_fil_crypt_threads) {
 		os_event_reset(fil_crypt_event);
-		os_event_wait_time(fil_crypt_event, 100000);
-	}
-
-	/* Send a message to encryption threads that there could be
-	something to do. */
-	if (srv_n_fil_crypt_threads) {
-		os_event_set(fil_crypt_threads_event);
+		os_event_wait_time(fil_crypt_event, 1000000);
 	}
 }
 
@@ -2376,8 +2516,7 @@ void
 fil_space_crypt_close_tablespace(
 	const fil_space_t*	space)
 {
-	if (!srv_encrypt_tables || !space->crypt_data
-	    || srv_n_fil_crypt_threads == 0) {
+	if (!srv_encrypt_tables || !space->crypt_data) {
 		return;
 	}
 
@@ -2413,9 +2552,8 @@ fil_space_crypt_close_tablespace(
 
 		if (now >= last + 30) {
 			ib_logf(IB_LOG_LEVEL_WARN,
-				"Waited %ld seconds to drop space: %s (" ULINTPF
-				") active threads %u flushing=%d.",
-				(long)(now - start), space->name, space->id, cnt, flushing);
+				"Waited %ld seconds to drop space: %s(" ULINTPF ").",
+				now - start, space->name, space->id);
 			last = now;
 		}
 	}
@@ -2436,11 +2574,10 @@ fil_space_crypt_get_status(
 	memset(status, 0, sizeof(*status));
 
 	ut_ad(space->n_pending_ops > 0);
-	fil_crypt_read_crypt_data(const_cast<fil_space_t*>(space));
-	status->space = ULINT_UNDEFINED;
+	fil_space_crypt_t* crypt_data = space->crypt_data;
+	status->space = space->id;
 
-	if (fil_space_crypt_t* crypt_data = space->crypt_data) {
-		status->space = space->id;
+	if (crypt_data != NULL) {
 		mutex_enter(&crypt_data->mutex);
 		status->scheme = crypt_data->type;
 		status->keyserver_requests = crypt_data->keyserver_requests;
@@ -2519,77 +2656,3 @@ fil_space_get_scrub_status(
 		mutex_exit(&crypt_data->mutex);
 	}
 }
-
-#endif /* !UNIV_INNOCHECKSUM */
-
-/*********************************************************************
-Verify that post encryption checksum match calculated checksum.
-This function should be called only if tablespace contains crypt_data
-metadata (this is strong indication that tablespace is encrypted).
-Function also verifies that traditional checksum does not match
-calculated checksum as if it does page could be valid unencrypted,
-encrypted, or corrupted.
-
-@param[in]	page		Page to verify
-@param[in]	zip_size	zip size
-@return whether the encrypted page is OK */
-UNIV_INTERN
-bool fil_space_verify_crypt_checksum(const byte* page, ulint zip_size)
-{
-	ut_ad(mach_read_from_4(page + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION));
-
-	/* Compressed and encrypted pages do not have checksum. Assume not
-	corrupted. Page verification happens after decompression in
-	buf_page_io_complete() using buf_page_is_corrupted(). */
-	if (mach_read_from_2(page + FIL_PAGE_TYPE)
-	    == FIL_PAGE_PAGE_COMPRESSED_ENCRYPTED) {
-		return true;
-	}
-
-	/* Read stored post encryption checksum. */
-	const ib_uint32_t checksum = mach_read_from_4(
-		page + FIL_PAGE_FILE_FLUSH_LSN_OR_KEY_VERSION + 4);
-
-	/* If stored checksum matches one of the calculated checksums
-	page is not corrupted. */
-
-	switch (srv_checksum_algorithm_t(srv_checksum_algorithm)) {
-	case SRV_CHECKSUM_ALGORITHM_STRICT_CRC32:
-		if (zip_size) {
-			return checksum == page_zip_calc_checksum(
-				page, zip_size, SRV_CHECKSUM_ALGORITHM_CRC32);
-		}
-
-		return checksum == buf_calc_page_crc32(page);
-	case SRV_CHECKSUM_ALGORITHM_STRICT_NONE:
-		/* Starting with MariaDB 10.1.25, 10.2.7, 10.3.1,
-		due to MDEV-12114, fil_crypt_calculate_checksum()
-		is only using CRC32 for the encrypted pages.
-		Due to this, we must treat "strict_none" as "none". */
-	case SRV_CHECKSUM_ALGORITHM_NONE:
-		return true;
-	case SRV_CHECKSUM_ALGORITHM_STRICT_INNODB:
-		/* Starting with MariaDB 10.1.25, 10.2.7, 10.3.1,
-		due to MDEV-12114, fil_crypt_calculate_checksum()
-		is only using CRC32 for the encrypted pages.
-		Due to this, we must treat "strict_innodb" as "innodb". */
-	case SRV_CHECKSUM_ALGORITHM_INNODB:
-	case SRV_CHECKSUM_ALGORITHM_CRC32:
-		if (checksum == BUF_NO_CHECKSUM_MAGIC) {
-			return true;
-		}
-		if (zip_size) {
-			return checksum == page_zip_calc_checksum(
-				page, zip_size, SRV_CHECKSUM_ALGORITHM_CRC32)
-				|| checksum == page_zip_calc_checksum(
-					page, zip_size,
-					SRV_CHECKSUM_ALGORITHM_INNODB);
-		}
-
-		return checksum == buf_calc_page_crc32(page)
-			|| checksum == buf_calc_page_new_checksum(page);
-	}
-
-	ut_ad(!"unhandled innodb_checksum_algorithm");
-	return false;
-}
