@@ -25,7 +25,8 @@ Modified           Jan Lindström jan.lindstrom@mariadb.com
 *******************************************************/
 #include "m_string.h"
 #include "log0crypt.h"
-#include <mysql/service_my_crypt.h>
+#include <my_crypt.h>
+#include <my_crypt.h>
 
 #include "log0log.h"
 #include "srv0start.h" // for srv_start_lsn
@@ -33,6 +34,8 @@ Modified           Jan Lindström jan.lindstrom@mariadb.com
 
 #include "ha_prototypes.h" // IB_LOG_
 
+#include "my_crypt.h"
+
 /* Used for debugging */
 // #define DEBUG_CRYPT 1
 #define UNENCRYPTED_KEY_VER	0
