@@ -1,9 +1,9 @@
 /*****************************************************************************
 
-Copyright (c) 1996, 2017, Oracle and/or its affiliates. All rights reserved.
+Copyright (c) 1996, 2016, Oracle and/or its affiliates. All rights reserved.
 Copyright (c) 2008, Google Inc.
 Copyright (c) 2009, Percona Inc.
-Copyright (c) 2013, 2019, MariaDB Corporation.
+Copyright (c) 2013, 2017, MariaDB Corporation
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -73,7 +73,8 @@ Created 2/16/1996 Heikki Tuuri
 #include "btr0defragment.h"
 #include "ut0timer.h"
 #include "btr0scrub.h"
-#include "mysql/service_wsrep.h" /* wsrep_recovery */
+
+#include <mysql/service_wsrep.h>
 
 #ifndef UNIV_HOTBACKUP
 # include "trx0rseg.h"
@@ -120,9 +121,6 @@ UNIV_INTERN ibool	srv_have_fullfsync = FALSE;
 /** TRUE if a raw partition is in use */
 UNIV_INTERN ibool	srv_start_raw_disk_in_use = FALSE;
 
-/** UNDO tablespaces starts with space id. */
-ulint	srv_undo_space_id_start;
-
 /** TRUE if the server is being started, before rolling back any
 incomplete transactions */
 UNIV_INTERN ibool	srv_startup_is_before_trx_rollback_phase = FALSE;
@@ -131,11 +129,7 @@ UNIV_INTERN ibool	srv_is_being_started = FALSE;
 /** TRUE if the server was successfully started */
 UNIV_INTERN ibool	srv_was_started = FALSE;
 /** TRUE if innobase_start_or_create_for_mysql() has been called */
-static ibool		srv_start_has_been_called;
-
-/** Whether any undo log records can be generated */
-UNIV_INTERN bool	srv_undo_sources;
-
+static ibool		srv_start_has_been_called = FALSE;
 #ifdef UNIV_DEBUG
 /** InnoDB system tablespace to set during recovery */
 UNIV_INTERN uint	srv_sys_space_size_debug;
@@ -145,8 +139,8 @@ UNIV_INTERN uint	srv_sys_space_size_debug;
 SRV_SHUTDOWN_CLEANUP and then to SRV_SHUTDOWN_LAST_PHASE, and so on */
 UNIV_INTERN enum srv_shutdown_state	srv_shutdown_state = SRV_SHUTDOWN_NONE;
 
-/** Files comprising the system tablespace. Also used by Mariabackup. */
-UNIV_INTERN pfs_os_file_t	files[1000];
+/** Files comprising the system tablespace */
+static os_file_t	files[1000];
 
 /** io_handler_thread parameters for thread identification */
 static ulint		n[SRV_MAX_N_IO_THREADS];
@@ -209,39 +203,6 @@ UNIV_INTERN mysql_pfs_key_t	srv_purge_thread_key;
 UNIV_INTERN mysql_pfs_key_t	srv_log_tracking_thread_key;
 #endif /* UNIV_PFS_THREAD */
 
-/** Innobase start-up aborted. Perform cleanup actions.
-@param[in]	create_new_db	TRUE if new db is  being created
-@param[in]	file		File name
-@param[in]	line		Line number
-@param[in]	err		Reason for aborting InnoDB startup
-@return DB_SUCCESS or error code. */
-static
-dberr_t
-srv_init_abort(
-	bool		create_new_db,
-	const char*	file,
-	ulint		line,
-	dberr_t		err)
-{
-	if (create_new_db) {
-		ib_logf(IB_LOG_LEVEL_ERROR,
-			"Database creation was aborted"
-			" at %s [" ULINTPF "]"
-			" with error %s. You may need"
-			" to delete the ibdata1 file before trying to start"
-			" up again.",
-			file, line, ut_strerr(err));
-	} else {
-		ib_logf(IB_LOG_LEVEL_ERROR,
-			"Plugin initialization aborted"
-			" at %s [" ULINTPF "]"
-			" with error %s.",
-			file, line, ut_strerr(err));
-	}
-
-	return(err);
-}
-
 /*********************************************************************//**
 Convert a numeric string that optionally ends in G or M or K, to a number
 containing megabytes.
@@ -623,7 +584,7 @@ static MY_ATTRIBUTE((nonnull, warn_unused_result))
 dberr_t
 create_log_file(
 /*============*/
-	pfs_os_file_t*	file,	/*!< out: file handle */
+	os_file_t*	file,	/*!< out: file handle */
 	const char*	name)	/*!< in: log file name */
 {
 	ibool		ret;
@@ -767,7 +728,7 @@ create_log_files(
 
 	/* Create a log checkpoint. */
 	mutex_enter(&log_sys->mutex);
-	ut_d(recv_no_log_write = srv_apply_log_only);
+	ut_d(recv_no_log_write = FALSE);
 	recv_reset_logs(
 #ifdef UNIV_LOG_ARCHIVE
 		UT_LIST_GET_FIRST(log_sys->log_groups)->archived_file_no,
@@ -841,7 +802,7 @@ static MY_ATTRIBUTE((nonnull, warn_unused_result))
 dberr_t
 open_log_file(
 /*==========*/
-	pfs_os_file_t*	file,	/*!< out: file handle */
+	os_file_t*	file,	/*!< out: file handle */
 	const char*	name,	/*!< in: log file name */
 	os_offset_t*	size)	/*!< out: file size */
 {
@@ -862,35 +823,35 @@ open_log_file(
 	return(DB_SUCCESS);
 }
 
-
-/** Creates or opens database data files and closes them.
-@param[out]	create_new_db		true = create new database
-@param[out]	min_arch_log_no		min of archived log numbers in
-					data files
-@param[out]	max_arch_log_no		max of archived log numbers in
-					data files
-@param[out]	flushed_lsn		flushed lsn in fist datafile
-@param[out]	sum_of_new_sizes	sum of sizes of the new files
-					added
+/*********************************************************************//**
+Creates or opens database data files and closes them.
 @return	DB_SUCCESS or error code */
-MY_ATTRIBUTE((nonnull, warn_unused_result))
+static MY_ATTRIBUTE((nonnull, warn_unused_result))
 dberr_t
 open_or_create_data_files(
-	bool*		create_new_db,
+/*======================*/
+	ibool*		create_new_db,	/*!< out: TRUE if new database should be
+					created */
 #ifdef UNIV_LOG_ARCHIVE
-	lsn_t*		min_arch_log_no,
-	lsn_t*		max_arch_log_no,
+	lsn_t*		min_arch_log_no,/*!< out: min of archived log
+					numbers in data files */
+	lsn_t*		max_arch_log_no,/*!< out: max of archived log
+					numbers in data files */
 #endif /* UNIV_LOG_ARCHIVE */
-	lsn_t*		flushed_lsn,
-	ulint*		sum_of_new_sizes)
+	lsn_t*		min_flushed_lsn,/*!< out: min of flushed lsn
+					values in data files */
+	lsn_t*		max_flushed_lsn,/*!< out: max of flushed lsn
+					values in data files */
+	ulint*		sum_of_new_sizes)/*!< out: sum of sizes of the
+					new files added */
 {
 	ibool		ret;
 	ulint		i;
-	bool		one_opened	= false;
-	bool		one_created	= false;
+	ibool		one_opened	= FALSE;
+	ibool		one_created	= FALSE;
 	os_offset_t	size;
 	ulint		flags;
-	ulint		space = 0;
+	ulint		space;
 	ulint		rounded_size_pages;
 	char		name[10000];
 	fil_space_crypt_t*    crypt_data=NULL;
@@ -906,7 +867,7 @@ open_or_create_data_files(
 
 	*sum_of_new_sizes = 0;
 
-	*create_new_db = false;
+	*create_new_db = FALSE;
 
 	srv_normalize_path_for_win(srv_data_home);
 
@@ -958,7 +919,7 @@ open_or_create_data_files(
 				   && os_file_get_last_error(false)
 				   != OS_FILE_ALREADY_EXISTS
 #ifdef UNIV_AIX
-				   /* AIX 5.1 after security patch ML7 may have
+			    	   /* AIX 5.1 after security patch ML7 may have
 			           errno set to 0 here, which causes our
 				   function to return 100; work around that
 				   AIX problem */
@@ -994,10 +955,9 @@ open_or_create_data_files(
 			}
 
 			const char*	check_msg;
-
 			check_msg = fil_read_first_page(
 				files[i], FALSE, &flags, &space,
-				flushed_lsn, NULL);
+				min_flushed_lsn, max_flushed_lsn, NULL);
 
 			/* If first page is valid, don't overwrite DB.
 			It prevents overwriting DB when mysql_install_db
@@ -1028,7 +988,6 @@ open_or_create_data_files(
 					name);
 				return(DB_ERROR);
 			}
-
 			if (srv_data_file_is_raw_partition[i] == SRV_OLD_RAW) {
 				ut_a(!srv_read_only_mode);
 				files[i] = os_file_create(
@@ -1048,6 +1007,7 @@ open_or_create_data_files(
 			}
 
 			if (!ret) {
+
 				os_file_get_last_error(true);
 
 				ib_logf(IB_LOG_LEVEL_ERROR,
@@ -1057,6 +1017,7 @@ open_or_create_data_files(
 			}
 
 			if (srv_data_file_is_raw_partition[i] == SRV_OLD_RAW) {
+
 				goto skip_size_check;
 			}
 
@@ -1083,15 +1044,16 @@ size_check:
 						"auto-extending "
 						"data file %s is "
 						"of a different size "
-						ULINTPF " pages (rounded "
+						"%lu pages (rounded "
 						"down to MB) than specified "
 						"in the .cnf file: "
-						"initial " ULINTPF " pages, "
-						"max " ULINTPF " (relevant if "
+						"initial %lu pages, "
+						"max %lu (relevant if "
 						"non-zero) pages!",
 						name,
-						rounded_size_pages,
-						srv_data_file_sizes[i],
+						(ulong) rounded_size_pages,
+						(ulong) srv_data_file_sizes[i],
+						(ulong)
 						srv_last_file_size_max);
 
 					return(DB_ERROR);
@@ -1104,12 +1066,12 @@ size_check:
 
 				ib_logf(IB_LOG_LEVEL_ERROR,
 					"Data file %s is of a different "
-					"size " ULINTPF " pages (rounded down to MB) "
+					"size %lu pages (rounded down to MB) "
 					"than specified in the .cnf file "
-					ULINTPF " pages!",
+					"%lu pages!",
 					name,
-					rounded_size_pages,
-					srv_data_file_sizes[i]);
+					(ulong) rounded_size_pages,
+					(ulong) srv_data_file_sizes[i]);
 
 				return(DB_ERROR);
 			}
@@ -1118,17 +1080,15 @@ skip_size_check:
 			/* This is the earliest location where we can load
 			the double write buffer. */
 			if (i == 0) {
-				/* XtraBackup never loads corrupted pages from
-				the doublewrite buffer */
 				buf_dblwr_init_or_load_pages(
-					files[i], srv_data_file_names[i], !IS_XTRABACKUP());
+					files[i], srv_data_file_names[i], true);
 			}
 
 			bool retry = true;
 check_first_page:
 			check_msg = fil_read_first_page(
 				files[i], one_opened, &flags, &space,
-				flushed_lsn, &crypt_data);
+				min_flushed_lsn, max_flushed_lsn, &crypt_data);
 
 			if (check_msg) {
 
@@ -1165,9 +1125,9 @@ check_first_page:
 			       != fsp_flags_get_page_size(flags)) {
 
 				ib_logf(IB_LOG_LEVEL_ERROR,
-					"Data file \"%s\" uses page size " ULINTPF " ,"
+					"Data file \"%s\" uses page size %lu,"
 					"but the start-up parameter "
-					"is --innodb-page-size=" ULINTPF " .",
+					"is --innodb-page-size=%lu",
 					name,
 					fsp_flags_get_page_size(flags),
 					UNIV_PAGE_SIZE);
@@ -1198,9 +1158,9 @@ check_first_page:
 			}
 
 			ib_logf(IB_LOG_LEVEL_INFO,
-				"Setting file %s size to " ULINTPF " MB",
+				"Setting file %s size to %lu MB",
 				name,
-				(srv_data_file_sizes[i]
+				(ulong) (srv_data_file_sizes[i]
 					 >> (20 - UNIV_PAGE_SIZE_SHIFT)));
 
 			ret = os_file_set_size(
@@ -1227,6 +1187,11 @@ check_first_page:
 		ut_a(ret);
 
 		if (i == 0) {
+			if (!crypt_data) {
+				crypt_data = fil_space_create_crypt_data(FIL_ENCRYPTION_DEFAULT,
+					FIL_DEFAULT_ENCRYPTION_KEY);
+			}
+
 			flags = FSP_FLAGS_PAGE_SSIZE();
 
 			fil_space_create(name, 0, flags, FIL_TABLESPACE,
@@ -1254,7 +1219,7 @@ srv_undo_tablespace_create(
 	const char*	name,		/*!< in: tablespace name */
 	ulint		size)		/*!< in: tablespace size in pages */
 {
-	pfs_os_file_t	fh;
+	os_file_t	fh;
 	ibool		ret;
 	dberr_t		err = DB_SUCCESS;
 
@@ -1332,7 +1297,7 @@ srv_undo_tablespace_open(
 	const char*	name,		/*!< in: tablespace name */
 	ulint		space)		/*!< in: tablespace id */
 {
-	pfs_os_file_t	fh;
+	os_file_t	fh;
 	dberr_t		err	= DB_ERROR;
 	ibool		ret;
 	ulint		flags;
@@ -1363,30 +1328,11 @@ srv_undo_tablespace_open(
 		size = os_file_get_size(fh);
 		ut_a(size != (os_offset_t) -1);
 
-		/* Load the tablespace into InnoDB's internal
-		data structures. */
-
-		const char* check_msg;
-		fil_space_crypt_t* crypt_data = NULL;
-
-		/* Set the compressed page size to 0 (non-compressed) */
-		flags = FSP_FLAGS_PAGE_SSIZE();
-
-		/* Read first page to find out does the crypt_info
-		exists on undo tablespace. */
-		check_msg = fil_read_first_page(
-				fh, FALSE, &flags, &space,
-				NULL, &crypt_data, false);
-
 		ret = os_file_close(fh);
 		ut_a(ret);
 
-		if (check_msg) {
-			ib_logf(IB_LOG_LEVEL_ERROR,
-				"%s in data file %s",
-				check_msg, name);
-			return (err);
-		}
+		/* Load the tablespace into InnoDB's internal
+		data structures. */
 
 		/* We set the biggest space id to the undo tablespace
 		because InnoDB hasn't opened any other tablespace apart
@@ -1394,8 +1340,10 @@ srv_undo_tablespace_open(
 
 		fil_set_max_space_id_if_bigger(space);
 
+		/* Set the compressed page size to 0 (non-compressed) */
+		flags = FSP_FLAGS_PAGE_SSIZE();
 		fil_space_create(name, space, flags, FIL_TABLESPACE,
-				crypt_data,
+				NULL /* no encryption */,
 				true /* create */);
 
 		ut_a(fil_validate());
@@ -1417,15 +1365,12 @@ srv_undo_tablespace_open(
 /********************************************************************
 Opens the configured number of undo tablespaces.
 @return	DB_SUCCESS or error code */
+static
 dberr_t
 srv_undo_tablespaces_init(
 /*======================*/
 	ibool		create_new_db,		/*!< in: TRUE if new db being
 						created */
-	ibool		backup_mode,		/*!< in: TRUE disables reading
-						the system tablespace (used in
-						XtraBackup), FALSE is passed on
-						recovery. */
 	const ulint	n_conf_tablespaces,	/*!< in: configured undo
 						tablespaces */
 	ulint*		n_opened)		/*!< out: number of UNDO
@@ -1454,23 +1399,13 @@ srv_undo_tablespaces_init(
 
 	for (i = 0; create_new_db && i < n_conf_tablespaces; ++i) {
 		char	name[OS_FILE_MAX_PATH];
-		ulint	space_id  = i + 1;
-
-		DBUG_EXECUTE_IF("innodb_undo_upgrade",
-				space_id = i + 3;);
 
 		ut_snprintf(
 			name, sizeof(name),
 			"%s%cundo%03lu",
-			srv_undo_dir, SRV_PATH_SEPARATOR, space_id);
-
-		if (i == 0) {
-			srv_undo_space_id_start = space_id;
-			prev_space_id = srv_undo_space_id_start - 1;
-		}
-
-		undo_tablespace_ids[i] = space_id;
+			srv_undo_dir, SRV_PATH_SEPARATOR, i + 1);
 
+		/* Undo space ids start from 1. */
 		err = srv_undo_tablespace_create(
 			name, SRV_UNDO_TABLESPACE_SIZE_IN_PAGES);
 
@@ -1489,29 +1424,17 @@ srv_undo_tablespaces_init(
 	we build the undo_tablespace_ids ourselves since they don't
 	already exist. */
 
-	if (!create_new_db && !backup_mode) {
+	if (!create_new_db) {
 		n_undo_tablespaces = trx_rseg_get_n_undo_tablespaces(
 			undo_tablespace_ids);
-
-		if (n_undo_tablespaces != 0) {
-			srv_undo_space_id_start = undo_tablespace_ids[0];
-			prev_space_id = srv_undo_space_id_start - 1;
-		}
-
 	} else {
 		n_undo_tablespaces = n_conf_tablespaces;
 
-		undo_tablespace_ids[n_conf_tablespaces] = ULINT_UNDEFINED;
-
-		if (backup_mode) {
-			ut_ad(!create_new_db);
-			for (i = 0; i < n_undo_tablespaces; i++) {
-				undo_tablespace_ids[i]
-					= i + srv_undo_space_id_start;
-			}
-
-			prev_space_id = srv_undo_space_id_start - 1;
+		for (i = 1; i <= n_undo_tablespaces; ++i) {
+			undo_tablespace_ids[i - 1] = i;
 		}
+
+		undo_tablespace_ids[i] = ULINT_UNDEFINED;
 	}
 
 	/* Open all the undo tablespaces that are currently in use. If we
@@ -1535,6 +1458,8 @@ srv_undo_tablespaces_init(
 		ut_a(undo_tablespace_ids[i] != 0);
 		ut_a(undo_tablespace_ids[i] != ULINT_UNDEFINED);
 
+		/* Undo space ids start from 1. */
+
 		err = srv_undo_tablespace_open(name, undo_tablespace_ids[i]);
 
 		if (err != DB_SUCCESS) {
@@ -1569,23 +1494,11 @@ srv_undo_tablespaces_init(
 			break;
 		}
 
-		/** Note the first undo tablespace id in case of
-		no active undo tablespace. */
-		if (n_undo_tablespaces == 0) {
-			srv_undo_space_id_start = i;
-		}
-
 		++n_undo_tablespaces;
 
 		++*n_opened;
 	}
 
-	/** Explictly specify the srv_undo_space_id_start
-	as zero when there are no undo tablespaces. */
-	if (n_undo_tablespaces == 0) {
-		srv_undo_space_id_start = 0;
-	}
-
 	/* If the user says that there are fewer than what we find we
 	tolerate that discrepancy but not the inverse. Because there could
 	be unused undo tablespaces for future use. */
@@ -1630,11 +1543,10 @@ srv_undo_tablespaces_init(
 		mtr_start(&mtr);
 
 		/* The undo log tablespace */
-		for (i = 0; i < n_undo_tablespaces; ++i) {
+		for (i = 1; i <= n_undo_tablespaces; ++i) {
 
 			fsp_header_init(
-				undo_tablespace_ids[i],
-				SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, &mtr);
+				i, SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, &mtr);
 		}
 
 		mtr_commit(&mtr);
@@ -1709,10 +1621,12 @@ are not found and the user wants.
 @return	DB_SUCCESS or error code */
 UNIV_INTERN
 dberr_t
-innobase_start_or_create_for_mysql()
+innobase_start_or_create_for_mysql(void)
+/*====================================*/
 {
-	bool		create_new_db;
-	lsn_t		flushed_lsn;
+	ibool		create_new_db;
+	lsn_t		min_flushed_lsn;
+	lsn_t		max_flushed_lsn;
 #ifdef UNIV_LOG_ARCHIVE
 	lsn_t		min_arch_log_no	= LSN_MAX;
 	lsn_t		max_arch_log_no	= LSN_MAX;
@@ -1746,10 +1660,6 @@ innobase_start_or_create_for_mysql()
 	/* This should be initialized early */
 	ut_init_timer();
 
-	if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO) {
-		srv_read_only_mode = 1;
-	}
-
 	high_level_read_only = srv_read_only_mode
 		|| srv_force_recovery > SRV_FORCE_NO_TRX_UNDO;
 
@@ -1924,32 +1834,7 @@ innobase_start_or_create_for_mysql()
 	srv_startup_is_before_trx_rollback_phase = TRUE;
 
 #ifdef __WIN__
-	switch (os_get_os_version()) {
-	case OS_WIN95:
-	case OS_WIN31:
-	case OS_WINNT:
-		srv_use_native_conditions = FALSE;
-		/* On Win 95, 98, ME, Win32 subsystem for Windows 3.1,
-		and NT use simulated aio. In NT Windows provides async i/o,
-		but when run in conjunction with InnoDB Hot Backup, it seemed
-		to corrupt the data files. */
-
-		srv_use_native_aio = FALSE;
-		break;
-
-	case OS_WIN2000:
-	case OS_WINXP:
-		/* On 2000 and XP, async IO is available, but no condition variables. */
-		srv_use_native_aio = TRUE;
-		srv_use_native_conditions = FALSE;
- 		break;
-
-	default:
-		/* Vista and later have both async IO and condition variables */
-		srv_use_native_aio = TRUE;
-		srv_use_native_conditions = TRUE;
-		break;
-	}
+  srv_use_native_aio = TRUE;
 
 #elif defined(LINUX_NATIVE_AIO)
 
@@ -2051,13 +1936,7 @@ innobase_start_or_create_for_mysql()
 
 	srv_boot();
 
-	if (ut_crc32_sse2_enabled) {
-		ib_logf(IB_LOG_LEVEL_INFO, "Using SSE crc32 instructions");
-	} else if (ut_crc32_power8_enabled) {
-		ib_logf(IB_LOG_LEVEL_INFO, "Using POWER8 crc32 instructions");
-	} else {
-		ib_logf(IB_LOG_LEVEL_INFO, "Using generic crc32 instructions");
-	}
+	ib_logf(IB_LOG_LEVEL_INFO, ut_crc32_implementation);
 
 	if (!srv_read_only_mode) {
 
@@ -2184,8 +2063,7 @@ innobase_start_or_create_for_mysql()
 	ib_logf(IB_LOG_LEVEL_INFO,
 		"Initializing buffer pool, size = %.1f%c", size, unit);
 
-	err = buf_pool_init(srv_buf_pool_size, static_cast<bool>(srv_numa_interleave),
-			    srv_buf_pool_instances);
+	err = buf_pool_init(srv_buf_pool_size, srv_buf_pool_instances);
 
 	if (err != DB_SUCCESS) {
 		ib_logf(IB_LOG_LEVEL_ERROR,
@@ -2293,7 +2171,7 @@ innobase_start_or_create_for_mysql()
 #ifdef UNIV_LOG_ARCHIVE
 					&min_arch_log_no, &max_arch_log_no,
 #endif /* UNIV_LOG_ARCHIVE */
-					&flushed_lsn,
+					&min_flushed_lsn, &max_flushed_lsn,
 					&sum_of_new_sizes);
 	if (err == DB_FAIL) {
 
@@ -2337,12 +2215,12 @@ innobase_start_or_create_for_mysql()
 		bool success = buf_flush_list(ULINT_MAX, LSN_MAX, NULL);
 		ut_a(success);
 
-		flushed_lsn = log_get_lsn();
+		min_flushed_lsn = max_flushed_lsn = log_get_lsn();
 
 		buf_flush_wait_batch_end(NULL, BUF_FLUSH_LIST);
 
 		err = create_log_files(create_new_db, logfilename, dirnamelen,
-				       flushed_lsn, logfile0);
+				       max_flushed_lsn, logfile0);
 
 		if (err != DB_SUCCESS) {
 			return(err);
@@ -2362,8 +2240,19 @@ innobase_start_or_create_for_mysql()
 
 			if (err == DB_NOT_FOUND) {
 				if (i == 0) {
+					if (max_flushed_lsn
+					    != min_flushed_lsn) {
+						ib_logf(IB_LOG_LEVEL_ERROR,
+							"Cannot create"
+							" log files because"
+							" data files are"
+							" corrupt or"
+							" not in sync"
+							" with each other");
+						return(DB_ERROR);
+					}
 
-					if (flushed_lsn < (lsn_t) 1000) {
+					if (max_flushed_lsn < (lsn_t) 1000) {
 						ib_logf(IB_LOG_LEVEL_ERROR,
 							"Cannot create"
 							" log files because"
@@ -2378,14 +2267,14 @@ innobase_start_or_create_for_mysql()
 
 					err = create_log_files(
 						create_new_db, logfilename,
-						dirnamelen, flushed_lsn,
+						dirnamelen, max_flushed_lsn,
 						logfile0);
 
 					if (err == DB_SUCCESS) {
 						err = create_log_files_rename(
 							logfilename,
 							dirnamelen,
-							flushed_lsn,
+							max_flushed_lsn,
 							logfile0);
 					}
 
@@ -2395,22 +2284,20 @@ innobase_start_or_create_for_mysql()
 
 					/* Suppress the message about
 					crash recovery. */
-					flushed_lsn = log_get_lsn();
+					max_flushed_lsn = min_flushed_lsn
+						= log_get_lsn();
 					goto files_checked;
-				} else if (i < 2 && !IS_XTRABACKUP()) {
-						/* must have at least 2 log files */
-						ib_logf(IB_LOG_LEVEL_ERROR,
-							"Only one log file found.");
-						return(err);
+				} else if (i < 2) {
+					/* must have at least 2 log files */
+					ib_logf(IB_LOG_LEVEL_ERROR,
+						"Only one log file found.");
+					return(err);
 				}
 
 				/* opened all files */
 				break;
 			}
 
-			if (stat_info.type != OS_FILE_TYPE_FILE) {
-				break;
-			}
 			if (!srv_file_check_mode(logfilename)) {
 				return(DB_ERROR);
 			}
@@ -2498,7 +2385,6 @@ files_checked:
 
 	err = srv_undo_tablespaces_init(
 		create_new_db,
-		FALSE,
 		srv_undo_tablespaces,
 		&srv_undo_tablespaces_open);
 
@@ -2528,23 +2414,9 @@ files_checked:
 		mtr_start(&mtr);
 
 		fsp_header_init(0, sum_of_new_sizes, &mtr);
-		compile_time_assert(TRX_SYS_SPACE == 0);
-		compile_time_assert(IBUF_SPACE_ID == 0);
-
-		ulint ibuf_root = btr_create(
-			DICT_CLUSTERED | DICT_UNIVERSAL | DICT_IBUF,
-			0, 0, DICT_IBUF_ID_MIN,
-			dict_ind_redundant, &mtr);
 
 		mtr_commit(&mtr);
 
-		if (ibuf_root == FIL_NULL) {
-			return(srv_init_abort(true, __FILE__, __LINE__,
-					      DB_ERROR));
-		}
-
-		ut_ad(ibuf_root == IBUF_TREE_ROOT_PAGE_NO);
-
 		/* To maintain backward compatibility we create only
 		the first rollback segment before the double write buffer.
 		All the remaining rollback segments will be created later,
@@ -2570,19 +2442,17 @@ files_checked:
 		bool success = buf_flush_list(ULINT_MAX, LSN_MAX, NULL);
 		ut_a(success);
 
-		flushed_lsn = log_get_lsn();
+		min_flushed_lsn = max_flushed_lsn = log_get_lsn();
 
 		buf_flush_wait_batch_end(NULL, BUF_FLUSH_LIST);
 
 		/* Stamp the LSN to the data files. */
-		err = fil_write_flushed_lsn(flushed_lsn);
+		fil_write_flushed_lsn_to_data_files(max_flushed_lsn, 0);
 
-		if (err != DB_SUCCESS) {
-			return(err);
-		}
+		fil_flush_file_spaces(FIL_TABLESPACE);
 
 		err = create_log_files_rename(logfilename, dirnamelen,
-					      flushed_lsn, logfile0);
+					      max_flushed_lsn, logfile0);
 
 		if (err != DB_SUCCESS) {
 			return(err);
@@ -2637,7 +2507,7 @@ files_checked:
 
 		err = recv_recovery_from_checkpoint_start(
 			LOG_CHECKPOINT, LSN_MAX,
-			flushed_lsn);
+			min_flushed_lsn, max_flushed_lsn);
 
 		if (err != DB_SUCCESS) {
 			return(err);
@@ -2661,11 +2531,6 @@ files_checked:
 			recv_group_scan_log_recs(). */
 
 			recv_apply_hashed_log_recs(true);
-
-			if (recv_sys->found_corrupt_log) {
-				return (DB_CORRUPTION);
-			}
-
 			DBUG_PRINT("ib_log", ("apply completed"));
 		}
 
@@ -2750,11 +2615,6 @@ files_checked:
 
 		recv_recovery_from_checkpoint_finish();
 
-		if (srv_apply_log_only) {
-			ut_ad(IS_XTRABACKUP());
-			goto skip_processes;
-		}
-
 		if (srv_force_recovery < SRV_FORCE_NO_IBUF_MERGE) {
 			/* The following call is necessary for the insert
 			buffer to work with multiple tablespaces. We must
@@ -2793,17 +2653,6 @@ files_checked:
 			dict_check_tablespaces_and_store_max_id(dict_check);
 		}
 
-		if (IS_XTRABACKUP()
-			&& !srv_backup_mode
-			&& srv_read_only_mode
-			&& srv_log_file_size_requested != srv_log_file_size) {
-
-			ib_logf(IB_LOG_LEVEL_WARN,
-				"Log files size mismatch, ignored in readonly mode");
-			srv_log_file_size_requested = srv_log_file_size;
-		}
-
-
 		if (!srv_force_recovery
 		    && !recv_sys->found_corrupt_log
 		    && (srv_log_file_size_requested != srv_log_file_size
@@ -2825,7 +2674,7 @@ files_checked:
 			DBUG_EXECUTE_IF("innodb_log_abort_1",
 					return(DB_ERROR););
 
-			flushed_lsn = log_get_lsn();
+			min_flushed_lsn = max_flushed_lsn = log_get_lsn();
 
 			ib_logf(IB_LOG_LEVEL_WARN,
 				"Resizing redo log from %u*%u to %u*%u pages"
@@ -2834,7 +2683,7 @@ files_checked:
 				(unsigned) srv_log_file_size,
 				(unsigned) srv_n_log_files,
 				(unsigned) srv_log_file_size_requested,
-				flushed_lsn);
+				max_flushed_lsn);
 
 			buf_flush_wait_batch_end(NULL, BUF_FLUSH_LIST);
 
@@ -2844,7 +2693,7 @@ files_checked:
 			we need to explicitly flush the log buffers. */
 			fil_flush(SRV_LOG_SPACE_FIRST_ID);
 
-			ut_ad(flushed_lsn == log_get_lsn());
+			ut_ad(max_flushed_lsn == log_get_lsn());
 
 			/* Prohibit redo log writes from any other
 			threads until creating a log checkpoint at the
@@ -2856,7 +2705,8 @@ files_checked:
 					return(DB_ERROR););
 
 			/* Stamp the LSN to the data files. */
-			err = fil_write_flushed_lsn(flushed_lsn);
+			fil_write_flushed_lsn_to_data_files(
+				max_flushed_lsn, 0);
 
 			DBUG_EXECUTE_IF("innodb_log_abort_4", err = DB_ERROR;);
 
@@ -2864,6 +2714,8 @@ files_checked:
 				return(err);
 			}
 
+			fil_flush_file_spaces(FIL_TABLESPACE);
+
 			/* Close and free the redo log files, so that
 			we can replace them. */
 			fil_close_log_files(true);
@@ -2880,23 +2732,28 @@ files_checked:
 			srv_log_file_size = srv_log_file_size_requested;
 
 			err = create_log_files(create_new_db, logfilename,
-					       dirnamelen, flushed_lsn,
+					       dirnamelen, max_flushed_lsn,
 					       logfile0);
 
 			if (err != DB_SUCCESS) {
 				return(err);
 			}
 
+			/* create_log_files() can increase system lsn that is
+			why FIL_PAGE_FILE_FLUSH_LSN have to be updated */
+			min_flushed_lsn = max_flushed_lsn = log_get_lsn();
+			fil_write_flushed_lsn_to_data_files(min_flushed_lsn, 0);
+			fil_flush_file_spaces(FIL_TABLESPACE);
+
 			err = create_log_files_rename(logfilename, dirnamelen,
 						      log_get_lsn(), logfile0);
-
 			if (err != DB_SUCCESS) {
 				return(err);
 			}
 		}
 
-		recv_recovery_rollback_active();
 		srv_startup_is_before_trx_rollback_phase = FALSE;
+		recv_recovery_rollback_active();
 
 		/* It is possible that file_format tag has never
 		been set. In this case we initialize it to minimum
@@ -2936,9 +2793,10 @@ files_checked:
 	/* fprintf(stderr, "Max allowed record size %lu\n",
 	page_get_free_space_of_empty() / 2); */
 
-	if (!buf_dblwr_create()) {
-		return(srv_init_abort(create_new_db, __FILE__, __LINE__,
-				      DB_ERROR));
+	if (buf_dblwr == NULL) {
+		/* Create the doublewrite buffer to a new tablespace */
+
+		buf_dblwr_create();
 	}
 
 	/* Here the double write buffer has already been created and so
@@ -2968,9 +2826,6 @@ files_checked:
 		/* Can only happen if server is read only. */
 		ut_a(srv_read_only_mode);
 		srv_undo_logs = ULONG_UNDEFINED;
-	} else if (srv_available_undo_logs < srv_undo_logs) {
-		/* Should due to out of file space. */
-		return (srv_init_abort(create_new_db, __FILE__, __LINE__, DB_ERROR));
 	}
 
 	if (!srv_read_only_mode) {
@@ -3028,16 +2883,6 @@ files_checked:
 			srv_master_thread,
 			NULL, thread_ids + (1 + SRV_MAX_N_IO_THREADS));
 		thread_started[1 + SRV_MAX_N_IO_THREADS] = true;
-
-		srv_undo_sources = true;
-		/* Create the dict stats gathering thread */
-		srv_dict_stats_thread_active = true;
-		dict_stats_thread_handle = os_thread_create(
-			dict_stats_thread, NULL, NULL);
-		dict_stats_thread_started = true;
-
-		/* Create the thread that will optimize the FTS sub-system. */
-		fts_optimize_init();
 	}
 
 	if (!srv_read_only_mode
@@ -3082,17 +2927,13 @@ files_checked:
 		}
 
 
-		buf_page_cleaner_is_active = true;
-		buf_flush_page_cleaner_thread_handle = os_thread_create(
-			buf_flush_page_cleaner_thread, NULL, NULL);
+		buf_flush_page_cleaner_thread_handle = os_thread_create(buf_flush_page_cleaner_thread, NULL, NULL);
 		buf_flush_page_cleaner_thread_started = true;
-
-		buf_lru_manager_is_active = true;
-		buf_flush_lru_manager_thread_handle = os_thread_create(
-			buf_flush_lru_manager_thread, NULL, NULL);
-		buf_flush_lru_manager_thread_started = true;
 	}
 
+	buf_flush_lru_manager_thread_handle = os_thread_create(buf_flush_lru_manager_thread, NULL, NULL);
+	buf_flush_lru_manager_thread_started = true;
+
 	if (!srv_file_per_table && srv_pass_corrupt_table) {
 		fprintf(stderr, "InnoDB: Warning:"
 			" The option innodb_file_per_table is disabled,"
@@ -3139,10 +2980,10 @@ files_checked:
 		if (!wsrep_recovery) {
 #endif /* WITH_WSREP */
 		/* Create the buffer pool dump/load thread */
-		srv_buf_dump_thread_active = true;
 		buf_dump_thread_handle=
 			os_thread_create(buf_dump_thread, NULL, NULL);
 
+		srv_buf_dump_thread_active = true;
 		buf_dump_thread_started = true;
 #ifdef WITH_WSREP
 		} else {
@@ -3151,19 +2992,28 @@ files_checked:
 				"wsrep recovery.");
 		}
 #endif /* WITH_WSREP */
+
+		/* Create the dict stats gathering thread */
+		dict_stats_thread_handle = os_thread_create(
+			dict_stats_thread, NULL, NULL);
+		srv_dict_stats_thread_active = true;
+		dict_stats_thread_started = true;
+
+		/* Create the thread that will optimize the FTS sub-system. */
+		fts_optimize_init();
+
 		/* Create thread(s) that handles key rotation */
 		fil_system_enter();
-		btr_scrub_init();
 		fil_crypt_threads_init();
 		fil_system_exit();
-
-		/* Initialize online defragmentation. */
-		btr_defragment_init();
-		btr_defragment_thread_active = true;
-		os_thread_create(btr_defragment_thread, NULL, NULL);
 	}
 
-skip_processes:
+	/* Init data for datafile scrub threads */
+	btr_scrub_init();
+
+	/* Initialize online defragmentation. */
+	btr_defragment_init();
+
 	srv_was_started = TRUE;
 
 	return(DB_SUCCESS);
@@ -3199,10 +3049,13 @@ srv_fts_close(void)
 }
 #endif
 
-/** Shut down InnoDB. */
+/****************************************************************//**
+Shuts down the InnoDB database.
+@return	DB_SUCCESS or error code */
 UNIV_INTERN
-void
-innodb_shutdown()
+dberr_t
+innobase_shutdown_for_mysql(void)
+/*=============================*/
 {
 	ulint	i;
 
@@ -3212,20 +3065,15 @@ innodb_shutdown()
 				"Shutting down an improperly started, "
 				"or created database!");
 		}
+
+		return(DB_SUCCESS);
 	}
 
-	if (srv_undo_sources) {
-		ut_ad(!srv_read_only_mode);
+	if (!srv_read_only_mode) {
 		/* Shutdown the FTS optimize sub system. */
 		fts_optimize_start_shutdown();
 
 		fts_optimize_end();
-		dict_stats_shutdown();
-		while (row_get_background_drop_list_len_low()) {
-			srv_wake_master_thread();
-			os_thread_yield();
-		}
-		srv_undo_sources = false;
 	}
 
 	/* 1. Flush the buffer pool to disk, write the current lsn to
@@ -3326,13 +3174,14 @@ innodb_shutdown()
 		srv_misc_tmpfile = 0;
 	}
 
-	if (!srv_read_only_mode && !srv_apply_log_only) {
+	if (!srv_read_only_mode) {
 		dict_stats_thread_deinit();
 		fil_crypt_threads_cleanup();
-		btr_scrub_cleanup();
-		btr_defragment_shutdown();
 	}
 
+	/* Cleanup data for datafile scrubbing */
+	btr_scrub_cleanup();
+
 #ifdef __WIN__
 	/* MDEV-361: ha_innodb.dll leaks handles on Windows
 	MDEV-7403: should not pass recv_writer_thread_handle to
@@ -3438,11 +3287,89 @@ innodb_shutdown()
 
 	srv_was_started = FALSE;
 	srv_start_has_been_called = FALSE;
-	/* reset io_tid_i, in case current process does second innodb start (xtrabackup might do that).*/
-	io_tid_i = 0;
+
+	return(DB_SUCCESS);
 }
 #endif /* !UNIV_HOTBACKUP */
 
+
+/********************************************************************
+Signal all per-table background threads to shutdown, and wait for them to do
+so. */
+UNIV_INTERN
+void
+srv_shutdown_table_bg_threads(void)
+/*===============================*/
+{
+	dict_table_t*	table;
+	dict_table_t*	first;
+	dict_table_t*	last = NULL;
+
+	mutex_enter(&dict_sys->mutex);
+
+	/* Signal all threads that they should stop. */
+	table = UT_LIST_GET_FIRST(dict_sys->table_LRU);
+	first = table;
+	while (table) {
+		dict_table_t*	next;
+		fts_t*		fts = table->fts;
+
+		if (fts != NULL) {
+			fts_start_shutdown(table, fts);
+		}
+
+		next = UT_LIST_GET_NEXT(table_LRU, table);
+
+		if (!next) {
+			last = table;
+		}
+
+		table = next;
+	}
+
+	/* We must release dict_sys->mutex here; if we hold on to it in the
+	loop below, we will deadlock if any of the background threads try to
+	acquire it (for example, the FTS thread by calling que_eval_sql).
+
+	Releasing it here and going through dict_sys->table_LRU without
+	holding it is safe because:
+
+	 a) MySQL only starts the shutdown procedure after all client
+	 threads have been disconnected and no new ones are accepted, so no
+	 new tables are added or old ones dropped.
+
+	 b) Despite its name, the list is not LRU, and the order stays
+	 fixed.
+
+	To safeguard against the above assumptions ever changing, we store
+	the first and last items in the list above, and then check that
+	they've stayed the same below. */
+
+	mutex_exit(&dict_sys->mutex);
+
+	/* Wait for the threads of each table to stop. This is not inside
+	the above loop, because by signaling all the threads first we can
+	overlap their shutting down delays. */
+	table = UT_LIST_GET_FIRST(dict_sys->table_LRU);
+	ut_a(first == table);
+	while (table) {
+		dict_table_t*	next;
+		fts_t*		fts = table->fts;
+
+		if (fts != NULL) {
+			fts_shutdown(table, fts);
+		}
+
+		next = UT_LIST_GET_NEXT(table_LRU, table);
+
+		if (table == last) {
+			ut_a(!next);
+		}
+
+		table = next;
+	}
+}
+
 /*****************************************************************//**
 Get the meta-data filename from the table name. */
 UNIV_INTERN
@@ -3476,8 +3403,9 @@ srv_get_meta_data_filename(
 	if (strncmp(suffix, ".cfg", suffix_len) == 0) {
 		strcpy(filename, path);
 	} else {
-		ut_ad(!strcmp(suffix, ".ibd"));
-		memcpy(filename, path, len - suffix_len);
+		ut_ad(strncmp(suffix, ".ibd", suffix_len) == 0);
+
+		strncpy(filename, path, len - suffix_len);
 		suffix = filename + (len - suffix_len);
 		strcpy(suffix, ".cfg");
 	}
